Package: livelink
Title: Create Shareable Links for 'webR' and 'Shinylive' Environments
Version: 0.1.0
Authors@R: c(
    person("James Joseph", "Balamuta", , "james.balamuta@gmail.com", role = c("aut", "cre", "cph")),
    person("reprex authors", role = "cph",
           comment = "stringify_expression() and trim_common_leading_ws(), adapted in R/process-input.R; see inst/COPYRIGHTS")
  )
Description: Creates shareable links for 'R' code in 'WebAssembly' (WASM)
    Read-Eval-Print Loop (REPL) environments like 'webR'
    <https://webr.r-wasm.org/> and for 'Shiny' applications using
    'Shinylive' <https://shinylive.io/>. Supports single scripts,
    multi-file projects, exercise and solution pairs, and batch
    processing. Includes encoding, decoding, and previewing of links
    for both 'R' and 'Python' environments.
License: AGPL (>= 3)
URL: https://r-pkg.thecoatlessprofessor.com/livelink/,
        https://github.com/coatless-rpkg/livelink
BugReports: https://github.com/coatless-rpkg/livelink/issues
Depends: R (>= 4.1.0)
Imports: base64enc, cli, clipr, jsonlite, lzstring, RcppMsgPack, stats,
        tools, utils
Suggests: knitr, quarto, testthat (>= 3.0.0), withr
Encoding: UTF-8
SystemRequirements: Quarto command line tools
        (https://github.com/quarto-dev/quarto-cli).
Config/Needs/website: pkgdown
Config/testthat/edition: 3
Collate: 'as-data-frame.R' 'build-webr-repl-link.R' 'classes.R'
        'decode-shinylive-link.R' 'utils.R' 'decode-webr-repl-link.R'
        'extract-link.R' 'format-methods.R' 'knit-print.R'
        'knitr-engine.R' 'options.R' 'process-input.R'
        'shinylive-link.R' 'validators.R' 'webr-repl-directory-links.R'
        'webr-repl-links.R'
VignetteBuilder: quarto
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-07-17 08:50:52 UTC; ronin
Author: James Joseph Balamuta [aut, cre, cph],
  reprex authors [cph] (stringify_expression() and
    trim_common_leading_ws(), adapted in R/process-input.R; see
    inst/COPYRIGHTS)
Maintainer: James Joseph Balamuta <james.balamuta@gmail.com>
Repository: CRAN
Date/Publication: 2026-07-24 10:20:02 UTC

# Generated by roxygen2: do not edit by hand

S3method(as.character,shinylive_decoded)
S3method(as.character,shinylive_decoded_batch)
S3method(as.character,shinylive_directory)
S3method(as.character,shinylive_link)
S3method(as.character,shinylive_preview)
S3method(as.character,shinylive_project)
S3method(as.character,webr_decoded)
S3method(as.character,webr_decoded_batch)
S3method(as.character,webr_directory)
S3method(as.character,webr_exercise)
S3method(as.character,webr_link)
S3method(as.character,webr_preview)
S3method(as.character,webr_project)
S3method(as.data.frame,shinylive_decoded_batch)
S3method(as.data.frame,shinylive_directory)
S3method(as.data.frame,webr_decoded_batch)
S3method(as.data.frame,webr_directory)
S3method(format,shinylive_decoded)
S3method(format,shinylive_decoded_batch)
S3method(format,shinylive_directory)
S3method(format,shinylive_link)
S3method(format,shinylive_preview)
S3method(format,shinylive_project)
S3method(format,webr_decoded)
S3method(format,webr_decoded_batch)
S3method(format,webr_directory)
S3method(format,webr_exercise)
S3method(format,webr_link)
S3method(format,webr_preview)
S3method(format,webr_project)
S3method(knitr::knit_print,shinylive_directory)
S3method(knitr::knit_print,shinylive_link)
S3method(knitr::knit_print,shinylive_project)
S3method(knitr::knit_print,webr_directory)
S3method(knitr::knit_print,webr_exercise)
S3method(knitr::knit_print,webr_link)
S3method(knitr::knit_print,webr_project)
S3method(print,shinylive_decoded)
S3method(print,shinylive_decoded_batch)
S3method(print,shinylive_directory)
S3method(print,shinylive_link)
S3method(print,shinylive_preview)
S3method(print,shinylive_project)
S3method(print,webr_decoded)
S3method(print,webr_decoded_batch)
S3method(print,webr_directory)
S3method(print,webr_exercise)
S3method(print,webr_link)
S3method(print,webr_preview)
S3method(print,webr_project)
S3method(repl_urls,default)
export(decode_shinylive_link)
export(decode_webr_link)
export(preview_shinylive_link)
export(preview_webr_link)
export(repl_urls)
export(set_webr_base_url)
export(shinylive_directory)
export(shinylive_project)
export(shinylive_py_link)
export(shinylive_r_link)
export(use_livelink_engine)
export(use_livelink_hook)
export(webr_repl_directory)
export(webr_repl_exercise)
export(webr_repl_link)
export(webr_repl_project)

<!-- README.md is generated from README.qmd. Please edit that file -->

# livelink <picture><source media="(prefers-color-scheme: dark)" srcset="man/figures/livelink-logo-dark-animated.svg"><img class="ll-logo" src="man/figures/livelink-logo-light-animated.svg" align="right" alt="Logo: a hexagon holding a stack of code lines that recede into the distance, with the livelink wordmark and an outbound arrow." width="150"/></picture>

<!-- badges: start -->

[![R-CMD-check](https://github.com/coatless-rpkg/livelink/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/coatless-rpkg/livelink/actions/workflows/R-CMD-check.yaml)
[![Lifecycle:
experimental](https://img.shields.io/badge/lifecycle-experimental-orange.svg)](https://lifecycle.r-lib.org/articles/stages.html#experimental)
<!-- badges: end -->

Create shareable links for R code in WebAssembly (WASM) environments:
webR for a full R REPL, and Shinylive for R and Python Shiny apps.

There is no server and nothing to upload. Your code is compressed and
encoded *into the URL itself*. Everything after the `#` is a URL
fragment, which browsers never send anywhere, so the recipient clicks
the link and lands in a running R session in their browser.

<picture><source media="(prefers-color-scheme: dark)" srcset="man/figures/hero-dark.svg"><img class="ll-hero" src="man/figures/hero-light.svg" alt="Three steps. One: you write R inside webr_repl_link, with no quotes and no escaping. Two: you get a URL whose fragment carries the code itself. Three: the recipient opens it and lands in a live webR session, with the code in the editor and the plot already drawn." width="100%"/></picture>

Full documentation is at
<https://r-pkg.thecoatlessprofessor.com/livelink/>.

## Installation

You can install `livelink` from CRAN with:

``` r
install.packages("livelink")
```

Or install the development version from GitHub with:

``` r
# install.packages("remotes")
remotes::install_github("coatless-rpkg/livelink")
```

### Requirements

- R ≥ 4.1.0
- An internet connection (the links run in a browser)
- A modern web browser with WebAssembly support

## Quick Start

The examples below create and decode links for R code, R Shiny apps, and
Python Shiny apps.

### webR links

Share R code that runs in the browser.

Pass the code straight in, with no quotes and no escaping, just braces:

``` r
library(livelink)

link <- webr_repl_link({
  data(mtcars)
  plot(mtcars$mpg, mtcars$wt)
})

print(link)
#> 
#> ── webR Link ──
#> 
#> <https://webr.r-wasm.org/latest/#code=eJyb2LwkLzE3dUVxclFmQYle0JKCxJKM7foZ%2Bbmp%2BuWpSfGlxalF%2BnDJktSKkpsaKYkliRq5JcmJRcWaXAU5%2BSVQjkpuQbqOApRdXqIJAPpaJJY%3D&mz>
#> 
#> File: 'script.R' → '/home/web_user/script.R'
#> Version: "latest"
#> Autorun: FALSE
```

Strings, file paths, and the clipboard work too, if the code is already
somewhere else:

``` r
webr_repl_link("plot(1:10)")     # a string
webr_repl_link("analysis.R")     # a file on disk
webr_repl_link()                 # whatever is on your clipboard
```

### Shinylive apps

Share a Shiny application the same way:

``` r
# Simple Shiny app
app_code <- '
library(shiny)

ui <- fluidPage(
  titlePanel("Hello Shinylive!"),
  sidebarLayout(
    sidebarPanel(
      sliderInput("n", "Number of points:", 10, 100, 50)
    ),
    mainPanel(
      plotOutput("plot")
    )
  )
)

server <- function(input, output) {
  output$plot <- renderPlot({
    plot(rnorm(input$n), main = paste("Random points:", input$n))
  })
}

shinyApp(ui = ui, server = server)
'

# Create app link
app_link <- shinylive_r_link(app_code, mode = "app")
print(app_link)
#> 
#> ── Shinylive R App ──
#> 
#> <https://shinylive.io/r/app/#code=NobwRAdghgtgpmAXGKAHVA6ASmANGAYwHsIAXOMpMAHQgBsBLAIwCcoWBPACgGcALBhA4BKWrQCuDAAQAeALRSAZnUkATAApQA5nC60pU0g1J04miHDp6wACUt0iUgMoChjAG5wAhNTDDc+lI8DKpwTOwAMlAcROKkehAGBsGh4SzmlglJSTyMoSwAkhCocdYQvrhSvgBy4jBMcCxSRIpSqESCpDyIFVIAjAAMlYNDUgCsA6KJSf6BBjBQghlWc0moDqQA8nEl8b7rRKS+U9knUlNTtDyNnk3ySuIQBEYkXIK7lbGku8JSIIFfXYAEgOpFkChYFHy6g2XH+0wMoK4LAgRBYMDexTiQIg-ikC0EUgAvG0oDxyNYsFAIKoiDA2h0yN1eu9sbizgBfKYcsQQfiCDgAQXQXEkxKkkkq1xYt3F0tuUzwYFIHFQCGQ5AAHqQwByALpAA>
#> 
#> Files (1):
#> 'app.R'
#> 
#> Engine: "R"
#> Mode: "app"
```

## Multi-file Projects

A project can hold several files. Write each one as R, in braces, and
mix in a string for anything that is not code.

``` r
library(livelink)

project <- webr_repl_project(list(
  "analysis.R" = {
    source("utils.R")
    create_plot(mtcars)
  },
  "utils.R" = {
    create_plot <- function(data) {
      plot(data$mpg, data$wt, main = "MPG vs Weight")
    }
  },
  "README.md" = "# Car Analysis\nAnalysis of the mtcars dataset."
), autorun_files = "analysis.R")

print(project)
#> 
#> ── webR Project ──
#> 
#> <https://webr.r-wasm.org/latest/#code=eJxdj01qAkEQRvdzimKSwAyYmQskC0kkKyEIIUspx9JpmO4euqv9QVzpDbyCIHoA0bWn8DaOraJYq6K%2Bot6rxXypUNIGFRZjK2zSWpbI%2BT7NtaR0SJ22s2TSh5hpxMc3q53JKAodi6KahnGQGUKmdllojiRnaGy8QsfaOHWYecjquuwJ2yfCLfPn%2Fx6Owcc79JzKWGgVdZExhkkAVXnSefAqy34NfDfkGkgUCj4hbP7%2BwMDCP4l%2BzpXg9GKxbjXq381GIrveY%2FfkcU%2B9SfICX2igfv0%2FuDWge8A5weVTD7fEyQlXE38U&mza>
#> 
#> Files (3):
#> 'analysis.R' → '/home/web_user/analysis.R' (autorun)
#> 'utils.R' → '/home/web_user/utils.R'
#> 'README.md' → '/home/web_user/README.md'
#> Version: "latest"
```

Write the `list()` inside the call. If you assign it to a variable
first, R runs the braces before livelink ever sees them.

## Educational content

Create paired exercise and solution links:

``` r
exercise <- "
# Exercise: Calculate summary statistics
# TODO: Calculate the mean and median of mtcars$mpg
mean_mpg <- # YOUR CODE HERE
median_mpg <- # YOUR CODE HERE

cat('Mean MPG:', mean_mpg, '\\n')
cat('Median MPG:', median_mpg, '\\n')
"

solution <- "
# Solution: Calculate summary statistics
mean_mpg <- mean(mtcars$mpg)
median_mpg <- median(mtcars$mpg)

cat('Mean MPG:', mean_mpg, '\\n')
cat('Median MPG:', median_mpg, '\\n')
"

exercise_links <- webr_repl_exercise(exercise, solution, "mpg_stats")

# Share exercise with students
student_link <- repl_urls(exercise_links$exercise)

# Keep solution for instructor
solution_link <- repl_urls(exercise_links$solution)
```

## Batch processing

Turn a whole directory into links in one call:

``` r
# Process all R files in a directory
links <- webr_repl_directory("./examples/",
                            autorun = TRUE,
                            panels = c("editor", "plot"))

# Process Shiny app directories
shiny_links <- shinylive_directory("./shiny_apps/", 
                                  engine = "r", 
                                  mode = "app")
```

## Link preview and decoding

Inspect a webR link in memory, without writing any files:

``` r
# Preview a link without downloading
existing_url <- "https://webr.r-wasm.org/latest/#code=eJyb2LwkLzE3dUVxclFmQYle0JKCxJKM7foZ%2Bbmp%2BuWpSfGlxalF%2BnDJlMSSxCNcBTn5JRqGVoYGmgBEFxiu"
preview <- preview_webr_link(existing_url)
print(preview)
#> 
#> ── webR Link Preview ──
#> 
#> URL:
#> <https://webr.r-wasm.org/latest/#code=eJyb2LwkLzE3dUVxclFmQYle0JKCxJKM7foZ%2Bbmp%2BuWpSfGlxalF%2BnDJlMSSxCNcBTn5JRqGVoYGmgBEFxiu>
#> 
#> Files: 1
#> Total size: 10 bytes
#> Version: "latest"
#> Encoding: "mz"
#> 
#> 'script.R' (10 bytes)
#> 
#> Use print(preview, show_content = TRUE) to see file contents
```

Once you trust a link, decode it to write the files to disk:

``` r
# Extract files to local directory
result <- decode_webr_link(existing_url, output_dir = "./extracted")
print(result)

# Batch decode multiple URLs
urls <- c(url1, url2, url3)
results <- decode_webr_link(urls, output_dir = "./all_extracted")
```

## Python Shiny support

Python Shiny apps work exactly like the R ones:

``` r
py_app <- '
from shiny import App, render, ui

app_ui = ui.page_fluid(
    ui.h2("Hello from Python Shiny!"),
    ui.input_slider("n", "N", 0, 100, 20),
    ui.output_text_verbatim("result"),
)

def server(input, output, session):
    @output
    @render.text
    def result():
        return f"n*2 is {input.n() * 2}"

app = App(app_ui, server)
'

py_link <- shinylive_py_link(py_app, mode = "app")
print(py_link)
#> 
#> ── Shinylive Python App ──
#> 
#> <https://shinylive.io/py/app/#code=NobwRAdghgtgpmAXGKAHVA6VBPMAaMAYwHsIAXOcpMAHQgDMAnYmAAgGcALASwm1e4xUxRmVYBBdHlaNKAEziNpAV2506aVAH1VrALytVWKAHM4W+gBtVcgBR1Wjw9wycATPbAAJOJcvFWJhZWAAVsMk5SVgBlHj4AQhowAEo8BycjXlRlMi12S24FRk8IJOkkgDky1gAGaQBGGrrWNxrU9McjYhzs3IoAD1yAN0UAIygyQU9ZdmVLMiT2iGT1CAV6DkUR4qyc6W6yXul2OHZ2blJkxA7WAAED3pvb2TXFDAGFiCdWdZlTubItiuN2+sjIykYX3oSQgACo3AJ2KwQLsyBgIEDWLCWgBfJKrTT6CToWyaHTcY5bRQrCD4MBkbCoBDID5gHEAXSAA>
#> 
#> Files (1):
#> 'app.py'
#> 
#> Engine: "Python"
#> Mode: "app"
```

## License

AGPL (\>= 3)

## Acknowledgements

Thanks to [George Stagg](https://github.com/georgestagg) for
[webR](https://docs.r-wasm.org/webr/latest/) and its browser REPL, and
to [Winston Chang](https://github.com/wch) for the
[Shinylive](https://github.com/posit-dev/shinylive) share-URL feature.
livelink writes to the share formats they built and opens in the
runtimes they maintain, so it is mostly a friendly wrapper around a good
idea they had. [Pyodide](https://github.com/pyodide/pyodide) is its own
remarkable project, and it is what runs the Python side of Shinylive.
Thanks, too, to [`peeky`](https://github.com/coatless-rpkg/peeky) for
stating the other half of this idea plainly enough that the feature
became obvious. And for turning a braced R expression into clean,
verbatim source, livelink borrows
[reprex](https://reprex.tidyverse.org/)’s
[`stringify_expression()`](https://github.com/tidyverse/reprex/blob/main/R/stringify_expression.R),
copyright the reprex authors and [MIT
licensed](https://github.com/tidyverse/reprex/blob/main/LICENSE.md).
livelink is one step, and webrarian is the next, already taking shape.

#' Format a livelink object as a character vector
#'
#' Returns an object's printed representation as a character vector, one element
#' per line, so it can be captured, logged, pasted into a report, or otherwise
#' reused. [print()] renders the very same content to the console; `format()`
#' hands it back to you instead.
#'
#' @param x A livelink object (a link, project, exercise, directory, decoded
#'   result, batch, or preview).
#' @param ... Passed to the object's [print()] method, so preview-specific
#'   options such as `show_content` work here too.
#'
#' @return A character vector of formatted lines.
#'
#' @examples
#' link <- webr_repl_link("plot(1:10)")
#' format(link)
#' writeLines(format(link))
#'
#' @name format.livelink
NULL

# Each method captures its print() output with cli::cli_fmt(), so a single source
# of truth (the print method) drives both console output and this character form.

#' @rdname format.livelink
#' @export
format.webr_link <- function(x, ...) cli::cli_fmt(print(x, ...))

#' @rdname format.livelink
#' @export
format.webr_project <- function(x, ...) cli::cli_fmt(print(x, ...))

#' @rdname format.livelink
#' @export
format.webr_exercise <- function(x, ...) cli::cli_fmt(print(x, ...))

#' @rdname format.livelink
#' @export
format.webr_directory <- function(x, ...) cli::cli_fmt(print(x, ...))

#' @rdname format.livelink
#' @export
format.webr_decoded <- function(x, ...) cli::cli_fmt(print(x, ...))

#' @rdname format.livelink
#' @export
format.webr_decoded_batch <- function(x, ...) cli::cli_fmt(print(x, ...))

#' @rdname format.livelink
#' @export
format.webr_preview <- function(x, ...) cli::cli_fmt(print(x, ...))

#' @rdname format.livelink
#' @export
format.shinylive_link <- function(x, ...) cli::cli_fmt(print(x, ...))

#' @rdname format.livelink
#' @export
format.shinylive_project <- function(x, ...) cli::cli_fmt(print(x, ...))

#' @rdname format.livelink
#' @export
format.shinylive_directory <- function(x, ...) cli::cli_fmt(print(x, ...))

#' @rdname format.livelink
#' @export
format.shinylive_decoded <- function(x, ...) cli::cli_fmt(print(x, ...))

#' @rdname format.livelink
#' @export
format.shinylive_decoded_batch <- function(x, ...) cli::cli_fmt(print(x, ...))

#' @rdname format.livelink
#' @export
format.shinylive_preview <- function(x, ...) cli::cli_fmt(print(x, ...))

#' Stringify R expressions to source code
#'
#' Adapted from `stringify_expression()` in the reprex package
#' (<https://github.com/tidyverse/reprex/blob/main/R/stringify_expression.R>),
#' Copyright (c) 2024 reprex authors, MIT licensed -- see inst/COPYRIGHTS. The
#' srcref reconstruction, trailing-comment rescue, and common-indentation trim
#' come from reprex; the braced-body deparse fallback, the wholeSrcref bound on
#' the tail scan, and the clamped-range `getSrcLines()` shim below are local
#' changes.
#'
#' @param x Expression to stringify
#' @return Character string of source code
#' @noRd
stringify_expression <- function(x) {
  if (is.null(x)) {
    return(NULL)
  }

  .srcref <- utils::getSrcref(x)

  if (is.null(.srcref)) {
    # No srcrefs: R was parsed with keep.source = FALSE (the default under
    # Rscript and R CMD check), so comments are already gone and only the
    # abstract syntax survives. Deparse the braced body statement by statement
    # so the wrapping `{` and `}` do not end up in the shared script.
    if (is.call(x) && identical(x[[1]], as.name("{"))) {
      body_exprs <- as.list(x)[-1]
      lines <- unlist(lapply(body_exprs, deparse))
      return(enc2utf8(lines))
    }

    return(enc2utf8(deparse(x)))
  }

  # Construct a new srcref with the first_line, first_byte, etc. from the
  # first expression and the last_line, last_byte, etc. from the last one.
  first_src <- .srcref[[1]]
  last_src <- .srcref[[length(.srcref)]]

  .srcfile <- attr(first_src, "srcfile")

  src <- srcref(
    .srcfile,
    c(
      first_src[[1]], first_src[[2]],
      last_src[[3]], last_src[[4]],
      first_src[[5]], last_src[[6]],
      first_src[[7]], last_src[[8]]
    )
  )

  lines <- enc2utf8(as.character(src, useSource = TRUE))

  # remove the first brace and line if the brace is the only thing on the line
  lines[[1L]] <- sub("^[{]", "", lines[[1L]])
  if (!nzchar(lines[[1L]])) {
    lines <- lines[-1L]
  }

  # identify the last source line affiliated with an expression
  n <- utils::getSrcLocation(last_src, which = "line", first = FALSE)

  # rescue trailing comment on (current) last surviving line
  last_source_line <- getSrcLines(.srcfile, n, n) # "raw"
  last_line <- lines[length(lines)] # srcref'd
  m <- regexpr(last_line, last_source_line, fixed = TRUE)
  rescue_me <- substring(last_source_line, m + attr(m, "match.length"))
  if (grepl("^\\s*#", rescue_me)) {
    lines[length(lines)] <- paste0(last_line, rescue_me)
  }

  # rescue trailing comment lines, but only those still INSIDE the block.
  # Scanning to Inf for a closing brace swallows unrelated source whenever the
  # block's own `}` is not on a line of its own: a one-line `{ plot(1:10) }`
  # nested in a call would take everything down to the next line starting with
  # `}`. The `{` call carries a wholeSrcref spanning itself; that is the bound.
  whole <- attr(x, "wholeSrcref")
  block_end <- if (!is.null(whole)) whole[[3L]] else n

  tail_lines <- getSrcLines(.srcfile, n + 1, block_end)
  closing_bracket_line <- max(grep("^\\s*[}]", tail_lines), 0)
  tail_lines <- utils::head(tail_lines, closing_bracket_line - 1)

  trim_common_leading_ws(c(lines, tail_lines))
}

#' Trim common leading whitespace from lines
#'
#' Adapted from the reprex package's helper of the same name (R/utils.R),
#' Copyright (c) 2024 reprex authors, MIT licensed -- see inst/COPYRIGHTS.
#' The guards for empty and whitespace-only input are local additions.
#'
#' @param lines Character vector of lines
#' @return Character vector with common leading whitespace removed
#' @noRd
trim_common_leading_ws <- function(lines) {
  if (length(lines) == 0) return(lines)

  # Find common leading whitespace
  non_empty <- lines[nzchar(trimws(lines, which = "right"))]
  if (length(non_empty) == 0) return(lines)

  leading_ws <- regexpr("^\\s*", non_empty)
  ws_lengths <- attr(leading_ws, "match.length")
  min_ws <- min(ws_lengths)

  if (min_ws > 0) {
    lines <- sub(paste0("^\\s{", min_ws, "}"), "", lines)
  }

  lines
}

#' Get source lines from srcfile
#'
#' Deliberately shadows base::getSrcLines(), which the reprex original calls,
#' with clamped-range semantics: out-of-range requests return character(0)
#' rather than erroring.
#'
#' @param srcfile Source file object
#' @param start Start line
#' @param end End line
#' @return Character vector of source lines
#' @noRd
getSrcLines <- function(srcfile, start, end = start) {
  if (start > end || start < 1) {
    return(character(0))
  }

  srcfile$lines[start:min(end, length(srcfile$lines))]
}

#' Read clipboard content
#' @return Character vector of clipboard content
#' @noRd
ingest_clipboard <- function() {
  if (!requireNamespace("clipr", quietly = TRUE)) {
    cli::cli_abort(c(
      "Package {.pkg clipr} is required for clipboard input",
      "i" = "Install with: {.code install.packages('clipr')}"
    ))
  }

  if (!clipr::clipr_available()) {
    cli::cli_abort(c(
      "Clipboard is not available",
      "i" = "This may happen in non-interactive sessions or some server environments"
    ))
  }

  content <- clipr::read_clip()
  if (length(content) == 0) {
    cli::cli_abort("Clipboard is empty")
  }

  content
}

#' Locate and categorize input type
#'
#' The locate-then-switch dispatch design, and the helper names locate_input()
#' and ingest_clipboard(), follow the reprex package's input handling
#' (R/utils-io.R, R/utils-clipboard.R, R/reprex_impl.R); the implementations
#' here are livelink's own.
#'
#' @param input Input provided by user
#' @param x_expr Expression provided by user
#' @return Character string indicating input type
#' @noRd
locate_input <- function(input = NULL, x_expr = NULL) {
  if (!is.null(x_expr)) {
    return("expr")
  }

  if (is.null(input)) {
    return("clipboard")
  }

  if (is.character(input)) {
    # Check if all elements look like file paths
    if (all(is_likely_file_path(input))) {
      # Verify files exist
      if (all(file.exists(input))) {
        return("path")
      } else {
        missing_files <- input[!file.exists(input)]
        # The quantity has to appear in the same string as the {?s} or cli
        # cannot pluralize and aborts on its own error message.
        cli::cli_abort(c(
          "{length(missing_files)} file{?s} not found",
          "x" = "Cannot find: {.file {missing_files}}",
          "i" = "To share this as code rather than a file path, pass it as a multi-line string."
        ))
      }
    } else {
      return("input")
    }
  }

  if (is.list(input)) {
    return("input")
  }

  cli::cli_abort(c(
    "Invalid input type",
    "x" = "Input must be a character vector (code or file paths), a named list, or an expression"
  ))
}

#' Check if string looks like a file path
#'
#' A file on disk is always a path. For anything else we have to guess, and the
#' guess must be conservative: R code routinely contains `/` (division) and
#' trailing dot-suffixes (`df$col.name`), so treating those as paths would
#' reject ordinary code. We therefore only call a non-existent string a path
#' when it carries no code syntax *and* has a file extension -- which keeps
#' mistyped filenames reporting "not found" instead of being silently encoded.
#'
#' @param x Character vector to check
#' @return Logical vector, one element per input
#' @noRd
is_likely_file_path <- function(x) {
  if (!is.character(x)) return(FALSE)

  vapply(x, function(path) {
    if (!nzchar(path) || grepl("\n", path, fixed = TRUE)) {
      return(FALSE)
    }

    if (file.exists(path)) {
      return(TRUE)
    }

    # Syntax that cannot appear in a filename anyone meant to write.
    if (grepl("[()<>={};\"'$]", path)) {
      return(FALSE)
    }

    grepl("\\.[A-Za-z0-9]{1,10}$", path)
  }, logical(1), USE.NAMES = FALSE)
}

#' Process input based on its type
#' @param input Input provided by user
#' @param x_expr Expression provided by user
#' @return Character string or named list of processed content
#' @noRd
process_input <- function(input = NULL, x_expr = NULL) {
  where <- locate_input(input, x_expr)

  switch(where,
         expr = {
           code <- stringify_expression(x_expr)
           if (is.null(code)) {
             cli::cli_abort("Failed to convert expression to source code")
           }
           paste(code, collapse = "\n")
         },
         clipboard = {
           content <- ingest_clipboard()
           paste(content, collapse = "\n")
         },
         path = {
           if (length(input) == 1) {
             # Single file - return content as string
             content <- readLines(input, warn = FALSE)
             paste(content, collapse = "\n")
           } else {
             # Multiple files - return named list
             files <- lapply(input, function(file) {
               content <- readLines(file, warn = FALSE)
               paste(content, collapse = "\n")
             })
             names(files) <- basename(input)
             files
           }
         },
         input = {
           if (length(input) == 1) {
             # Single string - return as is
             input
           } else {
             # Multiple strings - treat as separate files
             cli::cli_abort(c(
               "Multiple character strings not supported for single code input",
               "i" = "Use a single string, file paths, or an expression",
               "i" = "For multiple files, use the project functions"
             ))
           }
         }
  )
}

#' Is this captured expression a `{ ... }` block?
#'
#' Only a brace block means "treat my argument as literal source code". Any
#' other call -- `list(...)`, `paste0(...)`, `readLines(f)` -- is a value the
#' user wants evaluated, and deparsing it would ship the call itself as the
#' shared script.
#'
#' @param x_expr Result of [substitute()] on the input argument
#' @return TRUE if `x_expr` is a brace block
#' @noRd
is_brace_call <- function(x_expr) {
  is.call(x_expr) && identical(x_expr[[1]], as.name("{"))
}

#' Is this captured expression a literal `list(...)` call?
#'
#' Only a literal call written at the call site can be taken apart without
#' evaluating it. A symbol -- `webr_repl_project(project)` -- has to be forced
#' and handled the ordinary way.
#'
#' @param x_expr Result of [substitute()] on the input argument
#' @return TRUE if `x_expr` is a call to `list`
#' @noRd
is_literal_list_call <- function(x_expr) {
  is.call(x_expr) && identical(x_expr[[1]], as.name("list"))
}

#' Turn a literal `list(...)` call into a named list of file contents
#'
#' Each element is either a `{ ... }` block, taken as literal source, or an
#' ordinary value, evaluated in the caller's environment.
#'
#' The braces must NOT be evaluated. `list("a.R" = { x <- 1 })` would otherwise
#' run the block in the caller's frame and leave `x` behind; the whole point is
#' that this is code to ship, not code to run. Because we work from
#' [substitute()] and never force the promise, nothing in a brace is executed.
#'
#' @param x_expr The captured `list(...)` call
#' @param env Environment in which to evaluate the non-braced elements
#' @return Named list of file contents, or NULL if this is not the shape we want
#' @noRd
eval_project_list <- function(x_expr, env) {
  args <- as.list(x_expr)[-1]

  if (length(args) == 0 || is.null(names(args)) || !all(nzchar(names(args)))) {
    # Not project-shaped, so the caller falls through to ordinary evaluation --
    # which would force any braces. A brace is code to ship, never code to run,
    # so refuse a malformed list that carries one before anything executes.
    if (any(vapply(args, is_brace_call, logical(1)))) {
      cli::cli_abort(c(
        "Project files must be a named list",
        "i" = 'Name every element: {.code list("app.R" = {{ ... }})}'
      ))
    }
    return(NULL)
  }

  contents <- lapply(args, function(arg) {
    if (is_brace_call(arg)) {
      code <- stringify_expression(arg)
      if (is.null(code)) {
        cli::cli_abort("Failed to convert expression to source code")
      }
      paste(code, collapse = "\n")
    } else {
      eval(arg, env)
    }
  })

  files <- stats::setNames(contents, names(args))
  check_file_contents(files)

  files
}

#' Process input for a Shinylive app
#'
#' Shinylive apps are single-file or multi-file, so the input can be a lone code
#' string (or expression) as well as the project-shaped forms -- a named list of
#' files, or several file paths. Route the project-shaped ones accordingly;
#' [process_input()] deliberately rejects them.
#'
#' @param input Input provided by the user
#' @param x_expr Expression provided by the user
#' @param env Environment for evaluating list elements
#' @return A single code string, or a named list of file contents
#' @noRd
process_shinylive_input <- function(input = NULL, x_expr = NULL,
                                    env = parent.frame()) {
  # A literal list() may carry braced file contents, and must be taken apart
  # before anything in it is forced.
  if (is_literal_list_call(x_expr)) {
    files <- eval_project_list(x_expr, env)
    if (!is.null(files)) {
      return(files)
    }
  }

  is_multi_file <- is.list(input) || (is.character(input) && length(input) > 1)

  if (is_multi_file) {
    return(process_project_input(input = input))
  }

  # As above: a symbol is a value, not source to deparse.
  process_input(input = input,
                x_expr = if (is_brace_call(x_expr)) x_expr else NULL)
}

#' Process input for multi-file projects
#' @param input Input provided by user
#' @param x_expr Expression provided by user
#' @param env Environment for evaluating list elements
#' @return Named list of file contents
#' @noRd
process_project_input <- function(input = NULL, x_expr = NULL,
                                  env = parent.frame()) {
  # A literal list() written at the call site may name each file's contents as a
  # `{ ... }` block. Take it apart before the promise is forced -- forcing would
  # execute the blocks in the caller's frame.
  if (is_literal_list_call(x_expr)) {
    files <- eval_project_list(x_expr, env)
    if (!is.null(files)) {
      return(files)
    }
  }

  # Only a brace block is an "expression" here. `webr_repl_project(project)`
  # captures a symbol, which must be forced and handled like any other value.
  where <- locate_input(input, if (is_brace_call(x_expr)) x_expr else NULL)

  switch(where,
         expr = {
           cli::cli_abort(c(
             "Expressions not supported for multi-file projects",
             "i" = "Use a named list of files, or file paths",
             "i" = "Example: {.code list('main.R' = {{ plot(1:10) }})}"
           ))
         },
         clipboard = {
           cli::cli_abort(c(
             "Clipboard input not supported for multi-file projects",
             "i" = "Use a named list or file paths for multiple files"
           ))
         },
         path = {
           # Read all files and create named list
           files <- lapply(input, function(file) {
             content <- readLines(file, warn = FALSE)
             paste(content, collapse = "\n")
           })
           names(files) <- basename(input)
           files
         },
         input = {
           if (!is_named_list(input)) {
             cli::cli_abort(c(
               "Multi-file input must be a named list or vector of file paths",
               "i" = "Example: {.code list('main.R' = code1, 'utils.R' = code2)}"
             ))
           }
           check_file_contents(input)
           input
         }
  )
}

#' Check that every element of a project list is file contents
#'
#' Catches the one way the braced form goes wrong. `list()` is an ordinary call,
#' so its arguments evaluate unless livelink captures them -- which it can only do
#' when the list is written inside the call:
#'
#' ```
#' project <- list("main.R" = { plot(1:10) })   # the block RUNS, here and now
#' webr_repl_project(project)                   # and we receive its value
#'
#' webr_repl_project(list("main.R" = { plot(1:10) }))   # captured, never run
#' ```
#'
#' Assigning first leaves non-character values in the list, so say what happened
#' rather than serializing a function into a link.
#'
#' @param files Named list of file contents
#' @return Invisible TRUE, or aborts
#' @noRd
check_file_contents <- function(files) {
  bad <- !vapply(files, function(x) is.character(x) && length(x) == 1, logical(1))

  if (any(bad)) {
    cli::cli_abort(c(
      "Each file's contents must be a single string",
      "x" = "Not a string: {.file {names(files)[bad]}}",
      "i" = "Writing a file as {.code {{ ... }}} only works inside the call:",
      "*" = "{.code webr_repl_project(list('main.R' = {{ plot(1:10) }}))}",
      "!" = "Assigning the list to a variable first runs the block instead."
    ))
  }

  invisible(TRUE)
}

test_that("is_likely_file_path only accepts paths that actually exist", {
  tmp <- withr::local_tempdir()
  real <- file.path(tmp, "real.R")
  writeLines("plot(1:10)", real)

  expect_true(is_likely_file_path(real))

  # Code, not paths -- these must never be treated as files.
  expect_false(is_likely_file_path("x <- 1/2"))
  expect_false(is_likely_file_path("mean(x)/n"))
  expect_false(is_likely_file_path("df$col.name"))
  expect_false(is_likely_file_path("plot(1:10)"))

  # A plausible-but-missing filename stays a path, so the user gets a
  # "not found" error instead of silently sharing the literal string as code.
  expect_true(is_likely_file_path("nonexistent_file.R"))
})

# Regression: any single-line string containing "/" or a trailing ".ext" was
# classified as a file path, so ordinary code aborted instead of encoding.
test_that("code containing / or a dot-suffix is treated as code, not a path", {
  code <- c(
    "x <- 1/2",
    "mean(x)/n",
    "df$col.name",
    "ratio <- hits/total",
    "p <- ggplot(d) + facet_wrap(~a/b)"
  )

  for (snippet in code) {
    link <- webr_repl_link(snippet)

    expect_s3_class(link, "webr_link")
  }
})

test_that("a missing file path produces a readable error", {
  # The abort message must actually render -- it previously crashed cli with
  # "Cannot pluralize without a quantity".
  expect_error(
    locate_input(c("no_such_file.R", "also_missing.R")),
    "not found"
  )
  expect_error(locate_input("no_such_file.R"), "not found")
})

test_that("locate_input classifies each input type", {
  tmp <- withr::local_tempdir()
  f <- file.path(tmp, "script.R")
  writeLines("plot(1:10)", f)

  expect_equal(locate_input(f), "path")
  expect_equal(locate_input("plot(1:10)"), "input")
  expect_equal(locate_input(list("a.R" = "x <- 1")), "input")
  expect_equal(locate_input(x_expr = quote({ plot(1:10) })), "expr")
  expect_equal(locate_input(NULL), "clipboard")
})

test_that("process_input reads a single file path", {
  tmp <- withr::local_tempdir()
  f <- file.path(tmp, "script.R")
  writeLines(c("x <- 1", "y <- 2"), f)

  expect_equal(process_input(f), "x <- 1\ny <- 2")
})

test_that("process_input reads multiple file paths into a named list", {
  tmp <- withr::local_tempdir()
  a <- file.path(tmp, "a.R")
  b <- file.path(tmp, "b.R")
  writeLines("x <- 1", a)
  writeLines("y <- 2", b)

  res <- process_input(c(a, b))

  expect_named(res, c("a.R", "b.R"))
  expect_equal(res$`a.R`, "x <- 1")
})

test_that("expression input is converted to source code", {
  link <- webr_repl_link({
    x <- 1
    plot(x)
  })

  expect_s3_class(link, "webr_link")

  code <- preview_webr_link(as.character(link))$files_data[[1]]$text

  expect_match(code, "x <- 1")
  expect_match(code, "plot(x)", fixed = TRUE)
  # The wrapping braces must not survive into the shared source.
  expect_false(startsWith(trimws(code), "{"))
})

test_that("expression input preserves comments when source refs are available", {
  # Comments only survive parsing when keep.source = TRUE, so parse explicitly
  # rather than depending on the session default.
  expr <- parse(
    text = "{\n  # a leading comment\n  x <- 1 # a trailing comment\n}",
    keep.source = TRUE
  )[[1]]

  code <- paste(stringify_expression(expr), collapse = "\n")

  expect_match(code, "# a leading comment")
  expect_match(code, "# a trailing comment")
  expect_false(startsWith(trimws(code), "{"))
  # Common leading indentation is trimmed.
  expect_false(grepl("^\\s+#", strsplit(code, "\n")[[1]][1]))
})

# Regression: the trailing-comment rescue scanned forward to Inf looking for a
# line starting with `}`. A one-line brace nested in a call -- whose own `}` is
# not on a line of its own -- therefore swallowed every source line down to the
# next such line, shipping unrelated code into the link.
test_that("a one-line brace does not swallow the lines after it", {
  expr <- parse(
    text = paste(
      'f(list(',
      '  "main.R" = { plot(1:10) },',
      '  "other"  = "not code"',
      '))',
      sep = "\n"
    ),
    keep.source = TRUE
  )[[1]]

  brace <- expr[[2]][[2]]
  code <- paste(stringify_expression(brace), collapse = "\n")

  expect_equal(code, "plot(1:10)")
  expect_false(grepl("not code", code, fixed = TRUE))
  expect_false(grepl("main.R", code, fixed = TRUE))
})

# Regression: with keep.source = FALSE (the default under Rscript and R CMD
# check) there are no srcrefs, and the deparse fallback shipped the wrapping
# braces into the shared script.
test_that("expression input strips braces even without source refs", {
  expr <- parse(text = "{\n  x <- 1\n  plot(x)\n}", keep.source = FALSE)[[1]]

  code <- paste(stringify_expression(expr), collapse = "\n")

  expect_false(startsWith(trimws(code), "{"))
  expect_false(grepl("}", code, fixed = TRUE))
  expect_match(code, "x <- 1")
  expect_match(code, "plot(x)", fixed = TRUE)
})

# Regression: only a `{ ... }` block means "this argument is source code".
# is.call() was true for ANY call, so a computed value like paste0(...) or
# readLines(f) got deparsed and the call itself was shared as the script.
test_that("a computed value is evaluated, not deparsed as source", {
  link <- webr_repl_link(paste0("plot(", "1:10", ")"))
  code <- preview_webr_link(as.character(link))$files_data[[1]]$text

  expect_equal(code, "plot(1:10)")
  expect_false(grepl("paste0", code, fixed = TRUE))
})

test_that("a computed value is evaluated for Shinylive too", {
  link <- shinylive_r_link(paste0("library(", "shiny", ")"))
  code <- preview_shinylive_link(as.character(link))$files_data[[1]]$content

  expect_equal(code, "library(shiny)")
  expect_false(grepl("paste0", code, fixed = TRUE))
})

# Regression: named-list input was documented for the Shinylive link functions
# but process_input() aborted on any multi-element input, so it never worked.
test_that("shinylive link functions accept a named list of files", {
  files <- list(
    "app.R" = "library(shiny)",
    "utils.R" = "helper <- function() 42"
  )

  link <- shinylive_r_link(files)

  expect_s3_class(link, "shinylive_link")
  expect_setequal(link$files, c("app.R", "utils.R"))

  preview <- preview_shinylive_link(as.character(link))
  expect_equal(preview$total_files, 2)
})

test_that("shinylive_py_link accepts a named list of files", {
  link <- shinylive_py_link(list(
    "app.py" = "from shiny import App",
    "utils.py" = "def helper(): return 42"
  ))

  expect_setequal(link$files, c("app.py", "utils.py"))
})

test_that("shinylive link functions accept several file paths", {
  tmp <- withr::local_tempdir()
  app <- file.path(tmp, "app.R")
  helper <- file.path(tmp, "helper.R")
  writeLines("library(shiny)", app)
  writeLines("# helper", helper)

  link <- shinylive_r_link(c(app, helper))

  expect_setequal(link$files, c("app.R", "helper.R"))
})

test_that("process_project_input rejects unsupported input", {
  expect_error(
    process_project_input(x_expr = quote({ plot(1:10) })),
    "not supported"
  )
  expect_error(process_project_input(list("x <- 1")), "named list")
})

test_that("process_input rejects multiple bare code strings", {
  expect_error(process_input(c("x <- 1", "y <- 2")), "Multiple")
})

# Regression: shinylive_r_link()/shinylive_py_link() documented `NULL: read from
# clipboard` but their input argument had no default, so calling them with no
# arguments failed with `argument "files" is missing, with no default`.
test_that("link functions read from the clipboard when called with no input", {
  local_mocked_bindings(ingest_clipboard = function() "library(shiny)")

  expect_s3_class(webr_repl_link(), "webr_link")
  expect_s3_class(shinylive_r_link(), "shinylive_link")
  expect_s3_class(shinylive_py_link(), "shinylive_link")
})

test_that("clipboard content becomes the app body", {
  local_mocked_bindings(ingest_clipboard = function() "marker <- 1")

  link <- shinylive_r_link()

  expect_equal(link$files, "app.R")

  contents <- preview_shinylive_link(as.character(link))$files_data[[1]]$content
  expect_equal(contents, "marker <- 1")
})

#' Get the base WebR URL from options or default
#' @param version WebR version to use ("latest" or specific version like "v0.5.4")
#' @return Character string with the base URL
#' @noRd
get_webr_base_url <- function(version = "latest") {
  # Check if user has set a custom base URL
  custom_url <- getOption("livelink.base_url", NULL)

  if (!is.null(custom_url)) {
    return(custom_url)
  }

  # Build default URL with version
  paste0("https://webr.r-wasm.org/", version, "/")
}

#' Set global base URL for webR links
#'
#' @description
#' Overrides the base URL used when building webR REPL links, for instance to
#' point at a self-hosted or pinned webR deployment. The value is stored in the
#' `livelink.base_url` option and applies to webR links only; Shinylive links are
#' unaffected. Once set, a custom base URL takes precedence over the `version`
#' argument passed to the link builders.
#'
#' @param base_url Custom base URL to use for all webR links
#' @return Invisibly returns the `base_url` value (or `NULL` if resetting
#'   to default).
#' @export
#' @examples
#' # Remember the current setting so it can be restored afterwards
#' old <- getOption("livelink.base_url")
#'
#' # Set custom base URL
#' set_webr_base_url("https://my-custom-webr.com/")
#'
#' # Reset to default (removes custom setting)
#' set_webr_base_url(NULL)
#'
#' # Restore the previous setting
#' set_webr_base_url(old)
set_webr_base_url <- function(base_url = NULL) {
  if (is.null(base_url)) {
    # Remove custom setting
    options(livelink.base_url = NULL)
    cli::cli_inform("webR base URL reset to default")
  } else {
    check_single_string(base_url, "base_url")
    options(livelink.base_url = base_url)
    cli::cli_inform("webR base URL set to: {.url {base_url}}")
  }
  invisible(base_url)
}

#' Create webR REPL sharelinks from a directory of R files
#'
#' Batch processes all R files in a directory. By default each file becomes its
#' own webR sharelink; with `single_link = TRUE` the whole directory is bundled
#' into one link instead, exactly as [webr_repl_project()] would. Useful for
#' converting collections of scripts, examples, or course materials.
#'
#' @param directory_path Character string specifying the path to the directory containing R files
#' @param autorun Logical. Whether to enable autorun for all generated links (default: FALSE).
#'   With `single_link = TRUE`, this runs every R file in the bundle on arrival.
#' @param single_link Logical. If `FALSE` (default), each matched file becomes its
#'   own link and the result is a `webr_directory`. If `TRUE`, all matched files
#'   are packed into one link and the result is a single `webr_project`.
#' @param pattern Regular expression matched against file names in
#'   `directory_path`. Defaults to `"\\.R$"`, i.e. files ending in `.R`.
#' @param base_path Base directory path for files in webR (default: "/home/web_user/")
#' @param panels Character vector or string specifying which webR interface panels to show.
#'   Valid panels: "plot", "files", "terminal", "editor". Can be c("plot", "files") or "plot-files".
#'   If NULL (default), shows all panels.
#' @param version webR version to use ("latest" or specific version >= "v0.5.4")
#' @param base_url webR application URL. If NULL, uses global option or builds from version
#'
#' @return
#' By default a `webr_directory` object, whose `urls` element is a named character
#' vector mapping each filename to its webR sharelink. With `single_link = TRUE`, a
#' single `webr_project` object bundling every matched file into one link.
#'
#' @seealso [webr_repl_project()], which bundles a named list or a vector of file
#'   paths into one link.
#'
#' @examples
#' # A directory of R scripts
#' examples <- tempfile()
#' dir.create(examples)
#' writeLines("plot(1:10)", file.path(examples, "plot.R"))
#' writeLines("hist(rnorm(100))", file.path(examples, "hist.R"))
#'
#' links <- webr_repl_directory(examples, autorun = TRUE)
#' print(links)
#'
#' # Bundle the whole directory into one link instead
#' webr_repl_directory(examples, single_link = TRUE, panels = c("editor", "plot"))
#'
#' # Show only the editor and terminal panels
#' webr_repl_directory(examples, panels = c("editor", "terminal"))
#'
#' # Match a subset of files
#' webr_repl_directory(examples, pattern = "^plot")
#'
#' # The URLs, named by file
#' repl_urls(links)
#'
#' @export
webr_repl_directory <- function(directory_path,
                                autorun = FALSE,
                                single_link = FALSE,
                                pattern = "\\.R$",
                                base_path = "/home/web_user/",
                                panels = NULL,
                                version = "latest",
                                base_url = NULL) {

  check_single_string(directory_path, "directory_path")
  ensure_directory_exists(directory_path, "directory_path")
  check_single_logical(autorun, "autorun")
  check_single_logical(single_link, "single_link")
  check_valid_path(base_path, "base_path")
  check_valid_mode(panels, "panels")
  check_valid_version(version, "version")

  if (is.null(base_url)) {
    base_url <- get_webr_base_url(version)
  } else {
    check_single_string(base_url, "base_url")
  }

  if (!grepl("/$", base_path)) {
    base_path <- paste0(base_path, "/")
  }

  r_files <- list.files(directory_path, pattern = pattern, full.names = TRUE)

  if (length(r_files) == 0) {
    cli::cli_warn(c(
      "No files found",
      "!" = "No files matching pattern {.val {pattern}} found in {.path {directory_path}}",
      "i" = "Try adjusting the {.arg pattern} argument"
    ))
    # Stay type-stable: callers should get a webr_directory whether or not the
    # directory turned up any files.
    return(new_webr_directory(character(0), base_path, panels, version, directory_path))
  }

  if (single_link) {
    return(directory_single_link(
      r_files, autorun, base_path, panels, version, base_url, directory_path
    ))
  }

  cli::cli_inform(c(
    "v" = "Found {length(r_files)} file{?s} matching pattern {.val {pattern}}",
    "i" = "Processing files in {.path {directory_path}}..."
  ))

  links <- vapply(r_files, function(file) {
    tryCatch({
      code_text <- paste(readLines(file, warn = FALSE), collapse = "\n")
      filename <- basename(file)

      link_obj <- webr_repl_link(code_text,
                                 filename = filename,
                                 path = paste0(base_path, filename),
                                 autorun = autorun,
                                 panels = panels,
                                 version = version,
                                 base_url = base_url)

      link_obj$url
    }, error = function(e) {
      cli::cli_warn(c(
        "Failed to process file {.file {basename(file)}}",
        "x" = "{conditionMessage(e)}"
      ))
      NA_character_
    })
  }, character(1), USE.NAMES = FALSE)

  valid_links <- links[!is.na(links)]
  names(valid_links) <- basename(r_files[!is.na(links)])

  failed_count <- sum(is.na(links))
  if (failed_count > 0) {
    cli::cli_warn(c(
      "Some files failed to process",
      "!" = "{failed_count} file{?s} could not be processed"
    ))
  }

  cli::cli_inform(c(
    "v" = "Successfully created {length(valid_links)} WebR link{?s}"
  ))

  new_webr_directory(valid_links, base_path, panels, version, directory_path)
}


#' Bundle every file in a directory into one webR project link
#'
#' The `single_link = TRUE` branch of [webr_repl_directory()]. Reads each matched
#' file, names it by its basename, and hands the lot to [build_webr_project()].
#'
#' @param r_files Character vector of matched file paths
#' @param autorun Logical; `TRUE` runs every R file on arrival
#' @param base_path Normalized base directory path
#' @param panels Panels to show, or NULL
#' @param version WebR version
#' @param base_url Resolved WebR base URL
#' @param directory_path The source directory, for messages
#' @return A `webr_project` object
#' @noRd
directory_single_link <- function(r_files, autorun, base_path, panels, version,
                                  base_url, directory_path) {
  contents <- lapply(r_files, function(file) {
    tryCatch(
      paste(readLines(file, warn = FALSE), collapse = "\n"),
      error = function(e) {
        cli::cli_warn(c(
          "Failed to read file {.file {basename(file)}}",
          "x" = "{conditionMessage(e)}"
        ))
        NA_character_
      }
    )
  })

  ok <- !vapply(contents, function(x) length(x) == 1 && is.na(x), logical(1))
  processed_files <- stats::setNames(contents[ok], basename(r_files[ok]))

  if (length(processed_files) == 0) {
    cli::cli_abort(c(
      "No files could be read",
      "x" = "Every file matched in {.path {directory_path}} failed to read"
    ))
  }

  cli::cli_inform(c(
    "v" = "Bundling {length(processed_files)} file{?s} into one link"
  ))

  autorun_files <- if (autorun) "all" else character(0)
  build_webr_project(processed_files, autorun_files, base_path, panels,
                     version, base_url)
}

make_shiny_app <- function(parent, name, engine = "r") {
  dir <- file.path(parent, name)
  dir.create(dir, recursive = TRUE)
  app_file <- if (engine == "python") "app.py" else "app.R"
  writeLines("# app", file.path(dir, app_file))
  dir
}

# Regression: sapply() returns list() when given zero subdirectories, and
# character(0)[list()] is an error -- so the friendly "no apps found" warning
# written for exactly this case was unreachable.
test_that("shinylive_directory warns rather than erroring on a directory with no apps", {
  empty <- withr::local_tempdir()

  expect_warning(result <- shinylive_directory(empty, engine = "r"), "No Shiny apps found")
  expect_s3_class(result, "shinylive_directory")
  expect_length(result$urls, 0)
})

test_that("shinylive_directory warns when subdirectories hold no app file", {
  parent <- withr::local_tempdir()
  dir.create(file.path(parent, "not_an_app"))
  writeLines("# nope", file.path(parent, "not_an_app", "helper.R"))

  expect_warning(result <- shinylive_directory(parent, engine = "r"), "No Shiny apps found")
  expect_length(result$urls, 0)
})

test_that("shinylive_directory builds a link per app subdirectory", {
  parent <- withr::local_tempdir()
  make_shiny_app(parent, "app_one")
  make_shiny_app(parent, "app_two")

  result <- suppressMessages(shinylive_directory(parent, engine = "r"))

  expect_s3_class(result, "shinylive_directory")
  expect_setequal(names(result$urls), c("app_one", "app_two"))
  expect_true(all(grepl("^https://shinylive\\.io/r/", result$urls)))
})

test_that("shinylive_directory finds Python apps", {
  parent <- withr::local_tempdir()
  make_shiny_app(parent, "py_app", engine = "python")

  result <- suppressMessages(shinylive_directory(parent, engine = "python"))

  expect_named(result$urls, "py_app")
  expect_match(result$urls[["py_app"]], "^https://shinylive\\.io/py/")
})

test_that("app files are named by their app-relative path in the link", {
  parent <- withr::local_tempdir()
  app <- make_shiny_app(parent, "app_one")
  dir.create(file.path(app, "data"))
  writeLines("x,y", file.path(app, "data", "values.csv"))

  result <- suppressMessages(shinylive_directory(parent, engine = "r"))
  files <- preview_shinylive_link(result$urls[["app_one"]])$files_data
  file_names <- vapply(files, function(f) f$name, character(1))

  expect_setequal(file_names, c("app.R", "data/values.csv"))
})

# Regression for CRAN's r-devel-windows failure: every Windows path contains
# backslash sequences (like \R, \U, \t) that are invalid TRE escapes, so an app
# path must never reach the regex engine as a pattern. Backslashes are legal in
# Unix file names, so a Windows-shaped directory name reproduces the Windows
# failure on every platform.
test_that("shinylive_directory survives regex-hostile directory names", {
  skip_on_os("windows") # backslash and colon are illegal in Windows file names

  parent <- withr::local_tempdir()
  hostile <- file.path(parent, "d:\\temp\\Rtmp+1")
  dir.create(hostile)
  make_shiny_app(hostile, "app_one")

  result <- suppressMessages(shinylive_directory(hostile, engine = "r"))

  expect_named(result$urls, "app_one")
  files <- preview_shinylive_link(result$urls[["app_one"]])$files_data
  expect_equal(vapply(files, function(f) f$name, character(1)), "app.R")
})

test_that("webr_repl_directory builds a link per R script", {
  dir <- withr::local_tempdir()
  writeLines("plot(1:10)", file.path(dir, "one.R"))
  writeLines("hist(rnorm(10))", file.path(dir, "two.R"))

  result <- suppressMessages(webr_repl_directory(dir))

  expect_s3_class(result, "webr_directory")
  expect_setequal(names(result$urls), c("one.R", "two.R"))
})

test_that("webr_repl_directory returns an empty object for a directory with no scripts", {
  empty <- withr::local_tempdir()

  expect_warning(result <- webr_repl_directory(empty), "No files found")
  # Must stay type-stable: an object, not character(0).
  expect_s3_class(result, "webr_directory")
  expect_length(result$urls, 0)
})

test_that("webr_repl_directory honors the pattern argument", {
  dir <- withr::local_tempdir()
  writeLines("plot(1:10)", file.path(dir, "keep.R"))
  writeLines("x <- 1", file.path(dir, "skip.R"))

  result <- suppressMessages(webr_repl_directory(dir, pattern = "^keep"))

  expect_named(result$urls, "keep.R")
})

test_that("webr_repl_directory(single_link = TRUE) bundles the whole directory", {
  dir <- withr::local_tempdir()
  writeLines("source('utils.R')", file.path(dir, "main.R"))
  writeLines("f <- function() 42", file.path(dir, "utils.R"))

  result <- suppressMessages(webr_repl_directory(dir, single_link = TRUE))

  # One link, not a directory of links.
  expect_s3_class(result, "webr_project")

  # The one link carries every matched file, named by basename.
  preview <- preview_webr_link(as.character(result))
  names <- vapply(preview$files_data, function(f) f$name, character(1))
  expect_setequal(names, c("main.R", "utils.R"))
})

test_that("single_link honors panels and autorun", {
  dir <- withr::local_tempdir()
  writeLines("plot(1:10)", file.path(dir, "one.R"))
  writeLines("hist(rnorm(10))", file.path(dir, "two.R"))

  result <- suppressMessages(webr_repl_directory(
    dir, single_link = TRUE, autorun = TRUE, panels = c("editor", "plot")
  ))
  url <- as.character(result)

  expect_match(url, "mode='editor-plot'", fixed = TRUE)
  # autorun bundles the a flag and marks every R file to run.
  expect_match(url, "&jza", fixed = TRUE)
  expect_setequal(preview_webr_link(url)$autorun_files, c("one.R", "two.R"))
})

test_that("single_link produces the same link as webr_repl_project on the same files", {
  dir <- withr::local_tempdir()
  writeLines("source('utils.R')", file.path(dir, "main.R"))
  writeLines("f <- function() 42", file.path(dir, "utils.R"))

  from_dir <- suppressMessages(webr_repl_directory(dir, single_link = TRUE))
  files <- list.files(dir, pattern = "\\.R$", full.names = TRUE)
  from_project <- webr_repl_project(files)

  expect_identical(as.character(from_dir), as.character(from_project))
})

#' Convert files to Shinylive JSON format
#' @param files Named list where names are filenames and values are content
#' @return List in Shinylive JSON format
#' @noRd
files_to_shinylive_json <- function(files) {
  lapply(names(files), function(filename) {
    list(
      name = filename,
      content = files[[filename]],
      type = "text"  # For now, only support text files
    )
  })
}

#' Proper LZ-string compression for Shinylive URLs
#' @param json_string JSON string to compress
#' @return LZ-string compressed and encoded string
#' @noRd
compress_for_shinylive <- function(json_string) {
  compressed <- lzstring::compressToEncodedURIComponent(json_string)

  # Defensive: lz-string's URI-safe alphabet does not emit "/", so this is a
  # no-op in practice; it guards against a path separator ever landing in the
  # URL fragment, and other R encoders for this format do the same.
  gsub("/", "-", compressed)
}

#' Build Shinylive URL with proper format
#' @param engine Engine ("r" or "python")
#' @param mode Mode ("editor" or "app")
#' @param encoded_data LZ-string compressed file data
#' @param header Whether to show header in app mode (default: TRUE)
#' @param base_url Base Shinylive URL
#' @return Complete Shinylive URL
#' @noRd
build_shinylive_url <- function(engine, mode, encoded_data, header = TRUE, base_url = NULL) {
  if (is.null(base_url)) {
    engine_short <- if (engine == "python") "py" else "r"
    base_url <- paste0("https://shinylive.io/", engine_short, "/", mode, "/")
  }

  # Add header parameter for app mode if needed
  header_param <- if (mode == "app" && !header) "h=0&" else ""

  paste0(base_url, "#", header_param, "code=", encoded_data)
}


#' Build a Shinylive link from already-processed input
#'
#' Shared body of [shinylive_r_link()] and [shinylive_py_link()]. A lone code
#' string becomes the app's entry-point file; anything else is already a named
#' list of files.
#'
#' @param processed_input Result of [process_input()]
#' @param engine Engine: `"r"` or `"python"`
#' @param mode Shinylive mode: `"editor"` or `"app"`
#' @param header Logical. Whether to show the header in app mode
#' @param base_url Custom Shinylive base URL, or NULL for the default
#' @return shinylive_link object
#' @noRd
build_shinylive_link <- function(processed_input, engine, mode, header, base_url) {
  check_valid_shinylive_mode(mode, "mode")
  check_single_logical(header, "header")
  if (!is.null(base_url)) {
    check_single_string(base_url, "base_url")
  }

  if (is.character(processed_input) && length(processed_input) == 1) {
    entry_point <- if (engine == "python") "app.py" else "app.R"
    files_list <- stats::setNames(list(processed_input), entry_point)
  } else {
    files_list <- processed_input
  }

  url <- encode_shinylive_url(files_list, engine, mode, header, base_url)

  new_shinylive_link(url, files_list, engine, mode)
}

#' Encode Shinylive files into a share URL
#'
#' The shared encoding tail for Shinylive links: reshape to the Shinylive JSON
#' form, serialize, LZ-string-compress, then build the URL. Shared by
#' [build_shinylive_link()] and [shinylive_project()].
#' @param files_list Named list of file contents
#' @param engine "r" or "python"
#' @param mode "editor" or "app"
#' @param header Logical; whether to keep the Shinylive header
#' @param base_url Base URL, or NULL for the default
#' @return A Shinylive share URL
#' @noRd
encode_shinylive_url <- function(files_list, engine, mode, header, base_url) {
  shinylive_files <- files_to_shinylive_json(files_list)
  json_data <- jsonlite::toJSON(shinylive_files, auto_unbox = TRUE)
  encoded_data <- compress_for_shinylive(json_data)
  build_shinylive_url(engine, mode, encoded_data, header = header, base_url)
}

#' Create a Shinylive sharelink for Python Shiny apps
#'
#' Generates a shareable URL for Python Shiny applications that can run in the browser
#' using Shinylive. Supports character strings, file paths, named lists, and clipboard input.
#'
#' @param input App input. Can be:
#'   - Character string: Python code for the app
#'   - File path: Path to app.py file
#'   - Vector of file paths: Multiple files for the app
#'   - Named list: `list("app.py" = code1, "utils.py" = code2)`
#'   - NULL: Read from clipboard
#' @param mode Shinylive display mode (default `"editor"`). `"editor"` shows an
#'   editable code panel beside the running app; `"app"` shows only the running app.
#' @param header Logical, whether to show the Shinylive header bar. It applies only
#'   when `mode = "app"` and is ignored in the default `"editor"` mode. Defaults to `TRUE`.
#' @param base_url Custom Shinylive base URL. If NULL (default), links point at https://shinylive.io.
#'
#' @return shinylive_link object containing the Shinylive URL and metadata
#'
#' @seealso [shinylive_project()] for multi-file apps; [decode_shinylive_link()]
#'   and [preview_shinylive_link()] to read a link back.
#'
#' @export
#' @examples
#' # String input
#' app_code <- "
#' from shiny import App, render, ui
#' app_ui = ui.page_fluid(ui.h2('Hello World'))
#' def server(input, output, session): pass
#' app = App(app_ui, server)
#' "
#' shinylive_py_link(app_code)
#'
#' # Multiple files as a named list
#' shinylive_py_link(list(
#'   "app.py" = app_code,
#'   "utils.py" = "def helper(): return 42"
#' ))
#'
#' # File path input
#' app_dir <- tempfile()
#' dir.create(app_dir)
#' app_path <- file.path(app_dir, "app.py")
#' writeLines("from shiny import App, ui", app_path)
#' shinylive_py_link(app_path)
#'
#' # Read the app from the clipboard
#' if (interactive()) {
#'   shinylive_py_link()
#' }
shinylive_py_link <- function(input = NULL, mode = "editor", header = TRUE, base_url = NULL) {
  build_shinylive_link(
    process_shinylive_input(input = input),
    engine = "python",
    mode = mode,
    header = header,
    base_url = base_url
  )
}

#' Create a Shinylive sharelink for R Shiny apps
#'
#' Generates a shareable URL for R Shiny applications that can run in the browser
#' using Shinylive. Supports expressions, character strings, file paths, named lists,
#' and clipboard input.
#'
#' @param input App input. Can be:
#'   - R expression (no quotes needed): `shinylive_r_link({ shinyApp(ui, server) })`
#'   - Character string: R code for the app
#'   - File path: Path to app.R file
#'   - Vector of file paths: Multiple files for the app
#'   - Named list: `list("app.R" = code1, "utils.R" = code2)`
#'   - NULL: Read from clipboard
#' @param mode Shinylive display mode (default `"editor"`). `"editor"` shows an
#'   editable code panel beside the running app; `"app"` shows only the running app.
#' @param header Logical, whether to show the Shinylive header bar. It applies only
#'   when `mode = "app"` and is ignored in the default `"editor"` mode. Defaults to `TRUE`.
#' @param base_url Custom Shinylive base URL. If NULL (default), links point at https://shinylive.io.
#'
#' @return shinylive_link object containing the Shinylive URL and metadata
#'
#' @section Comments in expression input:
#' Comments inside a `{ }` expression are recovered from R's source references, so
#' they survive interactively but are dropped inside a knitted 'Quarto' or
#' 'R Markdown' document. Pass a string or a file path, or use the `livelink` chunk
#' engine, if you need them preserved. See [webr_repl_link()] for the details.
#'
#' @seealso [shinylive_project()] for multi-file apps; [decode_shinylive_link()]
#'   and [preview_shinylive_link()] to read a link back; [livelink-knitr] to give
#'   a document chunk its own link;
#'   `vignette("webr-and-shinylive", package = "livelink")` for the guide.
#'
#' @export
#' @examples
#' # Expression input (no quotes needed!)
#' shinylive_r_link({
#'   ui <- fluidPage(titlePanel("Hello World"))
#'   server <- function(input, output) {}
#'   shinyApp(ui, server)
#' })
#'
#' # Multiple files as a named list
#' shinylive_r_link(list(
#'   "app.R" = "library(shiny)\nshinyApp(fluidPage(), function(i, o) {})",
#'   "utils.R" = "helper <- function() 42"
#' ))
#'
#' # File path input
#' app_dir <- tempfile()
#' dir.create(app_dir)
#' app_path <- file.path(app_dir, "app.R")
#' writeLines("library(shiny)", app_path)
#' shinylive_r_link(app_path)
#'
#' # Read the app from the clipboard
#' if (interactive()) {
#'   shinylive_r_link()
#' }
shinylive_r_link <- function(input = NULL, mode = "editor", header = TRUE, base_url = NULL) {
  # Capture the expression form without evaluating `input`.
  x_expr <- substitute(input)
  processed_input <- if (!missing(input) && is_brace_call(x_expr)) {
    process_input(x_expr = x_expr)
  } else {
    process_shinylive_input(input = input, x_expr = x_expr,
                            env = parent.frame())
  }

  build_shinylive_link(
    processed_input,
    engine = "r",
    mode = mode,
    header = header,
    base_url = base_url
  )
}

#' Create a Shinylive sharelink for multi-file projects
#'
#' Creates Shinylive projects for either R or Python from named lists or
#' file path vectors.
#'
#' @param input Input for multiple files. Can be:
#'   - Named list: `list("app.R" = code1, "utils.R" = code2)`
#'   - Vector of file paths: `c("app.R", "utils.R", "data.csv")`
#' @param engine Engine to use: "r" for R Shiny or "python" for Python Shiny
#' @param mode Shinylive display mode (default `"editor"`). `"editor"` shows an
#'   editable code panel beside the running app; `"app"` shows only the running app.
#' @param header Logical, whether to show the Shinylive header bar. It applies only
#'   when `mode = "app"` and is ignored in the default `"editor"` mode. Defaults to `TRUE`.
#' @param base_url Custom Shinylive base URL. If NULL (default), links point at https://shinylive.io.
#'
#' @return shinylive_project object containing the Shinylive URL and metadata
#'
#' @seealso [shinylive_r_link()] and [shinylive_py_link()] for single apps;
#'   [decode_shinylive_link()] and [preview_shinylive_link()] to read a link back.
#'
#' @export
#' @examples
#' # Named list input
#' files <- list(
#'   "app.R" = "library(shiny)\nshinyApp(fluidPage(), function(i, o) {})",
#'   "utils.R" = "# Utility functions"
#' )
#' shinylive_project(files, engine = "r", mode = "editor")
#'
#' # File paths input
#' project_dir <- tempfile()
#' dir.create(project_dir)
#' app <- file.path(project_dir, "app.R")
#' utils <- file.path(project_dir, "utils.R")
#' writeLines("library(shiny)", app)
#' writeLines("# utils", utils)
#' shinylive_project(c(app, utils), engine = "r")
shinylive_project <- function(input, engine, mode = "editor", header = TRUE, base_url = NULL) {

  # Captured, not forced: a literal list() may name each file's contents as a
  # `{ ... }` block.
  x_expr <- substitute(input)
  processed_files <- process_project_input(
    input = input, x_expr = x_expr, env = parent.frame()
  )

  check_valid_shinylive_engine(engine, "engine")
  check_valid_shinylive_mode(mode, "mode")
  check_single_logical(header, "header")

  if (!is.null(base_url)) {
    check_single_string(base_url, "base_url")
  }

  url <- encode_shinylive_url(processed_files, engine, mode, header, base_url)

  new_shinylive_project(url, processed_files, engine, mode)
}

#' Create Shinylive sharelinks from a directory of Shiny apps
#'
#' @description
#' Batch processes directories containing Shiny applications to create individual Shinylive links.
#' Each subdirectory is treated as a separate Shiny app project.
#'
#' Only text files with extensions .R, .py, .txt, .md, .csv, .json, .yaml, or
#' .yml are embedded in a link. Other files (for example images or binary data)
#' are skipped with a warning.
#'
#' @param directory_path Character string specifying the path to the directory containing Shiny app directories
#' @param engine Engine to use: "r" for R Shiny or "python" for Python Shiny
#' @param mode Shinylive display mode (default `"editor"`). `"editor"` shows an
#'   editable code panel beside the running app; `"app"` shows only the running app.
#' @param header Logical, whether to show the Shinylive header bar. It applies only
#'   when `mode = "app"` and is ignored in the default `"editor"` mode. Defaults to `TRUE`.
#' @param app_file Main app filename to look for (default: "app.R" for R, "app.py" for Python)
#' @param base_url Custom Shinylive base URL. If NULL (default), links point at https://shinylive.io.
#'
#' @return shinylive_directory object containing URLs and metadata for all found apps
#'
#' @examples
#' # Each app lives in its own subdirectory:
#' #   shiny_apps/
#' #     app1/app.R
#' #     app2/app.R
#' shiny_apps <- tempfile()
#' dir.create(file.path(shiny_apps, "app1"), recursive = TRUE)
#' dir.create(file.path(shiny_apps, "app2"), recursive = TRUE)
#' writeLines("library(shiny)", file.path(shiny_apps, "app1", "app.R"))
#' writeLines("library(shiny)", file.path(shiny_apps, "app2", "app.R"))
#'
#' links <- shinylive_directory(shiny_apps, engine = "r", mode = "editor")
#' print(links)
#'
#' # Extract just the URLs
#' repl_urls(links)
#'
#' @export
shinylive_directory <- function(directory_path,
                                engine,
                                mode = "editor",
                                header = TRUE,
                                app_file = NULL,
                                base_url = NULL) {
  # Validate inputs
  check_single_string(directory_path, "directory_path")
  ensure_directory_exists(directory_path, "directory_path")
  check_valid_shinylive_engine(engine, "engine")
  check_valid_shinylive_mode(mode, "mode")
  check_single_logical(header, "header")

  if (!is.null(base_url)) {
    check_single_string(base_url, "base_url")
  }

  # Determine default app file
  if (is.null(app_file)) {
    app_file <- if (engine == "python") "app.py" else "app.R"
  }

  # Find subdirectories that contain the main app file
  subdirs <- list.dirs(directory_path, recursive = FALSE, full.names = TRUE)

  # vapply, not sapply: sapply() returns list() for zero subdirectories and
  # character(0)[list()] is an error, which made the "no apps found" branch below
  # unreachable in exactly the case it was written for.
  app_dirs <- subdirs[vapply(subdirs, function(dir) {
    file.exists(file.path(dir, app_file))
  }, logical(1))]

  if (length(app_dirs) == 0) {
    cli::cli_warn(c(
      "No Shiny apps found",
      "!" = "No directories containing {.file {app_file}} found in {.path {directory_path}}",
      "i" = "Each app should be in its own subdirectory with {.file {app_file}} as the main file"
    ))
    return(new_shinylive_directory(character(0), engine, mode, directory_path))
  }

  cli::cli_inform(c(
    "v" = "Found {length(app_dirs)} Shiny app{?s} in {.path {directory_path}}",
    "i" = "Processing {engine} Shiny apps..."
  ))

  # Process each app directory
  links <- sapply(app_dirs, function(app_dir) {
    tryCatch({
      app_name <- basename(app_dir)

      # Read all files in the app directory
      all_files <- list.files(app_dir, recursive = TRUE, full.names = TRUE)

      files <- list()
      for (file_path in all_files) {
        # Skip hidden files and directories
        if (grepl("^\\.", basename(file_path)) || file.info(file_path)$isdir) {
          next
        }

        # fixed = TRUE: app_dir is a filesystem path, not a regex. Windows
        # paths contain backslash sequences (\R, \U, \t) that are invalid TRE
        # escapes -- R-devel on Windows rejects them as errors, and other
        # flavors silently fail to match, leaking absolute paths into file
        # names.
        relative_path <- sub(paste0(app_dir, "/"), "", file_path, fixed = TRUE)

        # Read file content
        if (grepl("\\.(R|py|txt|md|csv|json|yaml|yml)$", file_path, ignore.case = TRUE)) {
          content <- readLines(file_path, warn = FALSE)
          files[[relative_path]] <- paste(content, collapse = "\n")
        } else {
          cli::cli_warn("Skipping binary file: {.file {relative_path}} in {.file {app_name}}")
        }
      }

      if (length(files) == 0) {
        cli::cli_warn("No readable files found in {.file {app_name}}")
        return(NA_character_)
      }

      # Create Shinylive project
      project <- shinylive_project(files, engine, mode, header, base_url)
      project$url

    }, error = function(e) {
      cli::cli_warn(c(
        "Failed to process app {.file {basename(app_dir)}}",
        "x" = "{e$message}"
      ))
      NA_character_
    })
  }, USE.NAMES = FALSE)

  # Remove failed apps and set names
  valid_links <- links[!is.na(links)]
  names(valid_links) <- basename(app_dirs[!is.na(links)])

  failed_count <- sum(is.na(links))
  if (failed_count > 0) {
    cli::cli_warn(c(
      "Some apps failed to process",
      "!" = "{failed_count} app{?s} could not be processed"
    ))
  }

  cli::cli_inform(c(
    "v" = "Successfully created {length(valid_links)} Shinylive link{?s}"
  ))

  # Return shinylive_directory object
  new_shinylive_directory(valid_links, engine, mode, directory_path)
}

# The core invariant of this package: whatever we encode into a sharelink must
# come back out byte-for-byte.

decode_to_temp <- function(url, decoder) {
  dir <- withr::local_tempdir(.local_envir = parent.frame())
  suppressMessages(decoder(url, output_dir = dir, overwrite = TRUE))

  paths <- list.files(dir, recursive = TRUE, full.names = TRUE)
  contents <- lapply(paths, function(p) paste(readLines(p, warn = FALSE), collapse = "\n"))
  stats::setNames(contents, basename(paths))
}

test_that("webR single-file links round-trip", {
  code <- "x <- rnorm(100)\nhist(x)"

  files <- decode_to_temp(as.character(webr_repl_link(code)), decode_webr_link)

  expect_named(files, "script.R")
  expect_equal(files[["script.R"]], code)
})

test_that("webR multi-file projects round-trip", {
  project <- list(
    "main.R" = "source('utils.R')\nrun()",
    "utils.R" = "run <- function() 42",
    "README.md" = "# Title\n\nSome prose."
  )

  files <- decode_to_temp(as.character(webr_repl_project(project)), decode_webr_link)

  expect_setequal(names(files), names(project))
  for (nm in names(project)) {
    expect_equal(files[[nm]], project[[nm]])
  }
})

test_that("Shinylive R apps round-trip", {
  app <- "library(shiny)\nui <- fluidPage()\nserver <- function(input, output) {}\nshinyApp(ui, server)"

  files <- decode_to_temp(as.character(shinylive_r_link(app)), decode_shinylive_link)

  expect_named(files, "app.R")
  expect_equal(files[["app.R"]], app)
})

test_that("Shinylive Python apps round-trip", {
  app <- "from shiny import App, ui\napp_ui = ui.page_fluid()\napp = App(app_ui, None)"

  files <- decode_to_temp(as.character(shinylive_py_link(app)), decode_shinylive_link)

  expect_named(files, "app.py")
  expect_equal(files[["app.py"]], app)
})

test_that("unicode survives a round-trip", {
  code <- "# héllo 你好 — emoji: \U0001F600\nx <- \"naïve café\""

  files <- decode_to_temp(as.character(webr_repl_link(code)), decode_webr_link)

  expect_equal(files[["script.R"]], code)
})

# Regression: simple_hash() overflowed on long input, so decoding a link built
# from a realistically-sized script failed with `invalid format '%x'`.
test_that("a large script round-trips", {
  set.seed(42)
  # Deliberately incompressible, so the URL stays long after gzip.
  lines <- paste0(
    "v", 1:600, " <- \"",
    replicate(600, paste(sample(c(letters, LETTERS, 0:9), 40, TRUE), collapse = "")),
    "\""
  )
  code <- paste(lines, collapse = "\n")
  url <- as.character(webr_repl_link(code))

  expect_gt(nchar(url), 8000)

  files <- decode_to_temp(url, decode_webr_link)

  expect_equal(files[["script.R"]], code)
})

test_that("webR link metadata round-trips", {
  url <- as.character(
    webr_repl_link("plot(1:10)", autorun = TRUE, version = "v0.5.4")
  )

  preview <- preview_webr_link(url)

  expect_equal(preview$version, "v0.5.4")
  expect_equal(preview$autorun_files, "script.R")
  expect_match(preview$flags, "a")
})

test_that("panels round-trip into the URL", {
  url <- as.character(webr_repl_link("plot(1:10)", panels = c("plot", "editor")))

  expect_match(url, "mode='plot-editor'", fixed = TRUE)
  expect_equal(preview_webr_link(url)$mode, c("plot", "editor"))
})

# Tests for shinylive link functions

test_that("shinylive_r_link creates valid shinylive_link object", {
  app_code <- "library(shiny)\nshinyApp(ui = fluidPage(), server = function(input, output) {})"
  result <- shinylive_r_link(app_code)

  expect_s3_class(result, "shinylive_link")
  expect_true(grepl("shinylive.io/r/", result$url))
  expect_true(grepl("#code=", result$url))
  expect_equal(result$engine, "r")
  expect_equal(result$mode, "editor")
  expect_true("app.R" %in% result$files)
})

test_that("shinylive_r_link respects mode parameter", {
  app_code <- "shinyApp(ui = fluidPage(), server = function(input, output) {})"

  editor_result <- shinylive_r_link(app_code, mode = "editor")
  expect_true(grepl("/editor/", editor_result$url))
  expect_equal(editor_result$mode, "editor")

  app_result <- shinylive_r_link(app_code, mode = "app")
  expect_true(grepl("/app/", app_result$url))
  expect_equal(app_result$mode, "app")
})

test_that("shinylive_r_link validates mode parameter", {
  expect_error(
    shinylive_r_link("code", mode = "invalid"),
    "must be either"
  )
})

test_that("shinylive_py_link creates valid shinylive_link object", {
  py_code <- "from shiny import App, ui\napp = App(ui.page_fluid(), None)"
  result <- shinylive_py_link(py_code)

  expect_s3_class(result, "shinylive_link")
  expect_true(grepl("shinylive.io/py/", result$url))
  expect_equal(result$engine, "python")
  expect_true("app.py" %in% result$files)
})

test_that("shinylive_py_link respects mode parameter", {
  py_code <- "from shiny import App, ui\napp = App(ui.page_fluid(), None)"

  editor_result <- shinylive_py_link(py_code, mode = "editor")
  expect_true(grepl("/editor/", editor_result$url))

  app_result <- shinylive_py_link(py_code, mode = "app")
  expect_true(grepl("/app/", app_result$url))
})

test_that("shinylive_project creates valid shinylive_project object", {
  skip_on_cran()
  temp_dir <- tempfile("shinylive_proj")
  dir.create(temp_dir)
  writeLines("library(shiny)\nshinyApp(fluidPage(), function(input, output) {})", file.path(temp_dir, "app.R"))
  writeLines("helper <- function() 1", file.path(temp_dir, "utils.R"))

  files <- c(file.path(temp_dir, "app.R"), file.path(temp_dir, "utils.R"))
  result <- shinylive_project(files, engine = "r")

  expect_s3_class(result, "shinylive_project")
  expect_true(grepl("shinylive.io/r/", result$url))
  expect_equal(result$engine, "r")
  expect_equal(length(result$files), 2)

  unlink(temp_dir, recursive = TRUE)
})

test_that("shinylive_project validates engine parameter", {
  skip_on_cran()
  temp_dir <- tempfile("shinylive_validate")
  dir.create(temp_dir)
  writeLines("code", file.path(temp_dir, "app.R"))

  expect_error(
    shinylive_project(file.path(temp_dir, "app.R"), engine = "julia"),
    "must be either"
  )

  unlink(temp_dir, recursive = TRUE)
})

test_that("repl_urls extracts URL from shinylive_link", {
  result <- shinylive_r_link("shinyApp(fluidPage(), function(i, o) {})")
  url <- repl_urls(result)

  expect_type(url, "character")
  expect_equal(url, result$url)
})

test_that("as.character works for shinylive_link", {
  result <- shinylive_r_link("shinyApp(fluidPage(), function(i, o) {})")
  url <- as.character(result)

  expect_type(url, "character")
  expect_equal(url, result$url)
})

sample_objects <- function() {
  tmp <- withr::local_tempdir(.local_envir = parent.frame())
  writeLines("plot(1:10)", file.path(tmp, "one.R"))
  app <- file.path(tmp, "app_one")
  dir.create(app)
  writeLines("# app", file.path(app, "app.R"))

  webr_url <- as.character(webr_repl_link("plot(1:10)"))
  shiny_url <- as.character(shinylive_r_link("library(shiny)"))
  out <- withr::local_tempdir(.local_envir = parent.frame())

  list(
    webr_link       = webr_repl_link("plot(1:10)"),
    webr_project    = webr_repl_project(list("a.R" = "x <- 1")),
    webr_exercise   = webr_repl_exercise("# todo", "x <- 1", "ex"),
    webr_directory  = suppressMessages(webr_repl_directory(tmp)),
    webr_preview    = preview_webr_link(webr_url),
    webr_decoded    = suppressMessages(decode_webr_link(webr_url, output_dir = out, overwrite = TRUE)),
    shinylive_link      = shinylive_r_link("library(shiny)"),
    shinylive_project   = shinylive_project(list("app.R" = "library(shiny)"), engine = "r"),
    shinylive_directory = suppressMessages(shinylive_directory(tmp, engine = "r")),
    shinylive_preview   = preview_shinylive_link(shiny_url),
    shinylive_decoded   = suppressMessages(
      decode_shinylive_link(shiny_url, output_dir = out, overwrite = TRUE)
    )
  )
}

test_that("every class has a print method that returns its input invisibly", {
  objs <- sample_objects()

  for (nm in names(objs)) {
    x <- objs[[nm]]

    # print() must dispatch to a real method, not print.default.
    expect_true(
      !is.null(utils::getS3method("print", class(x)[1], optional = TRUE)),
      info = paste("no print method for class", nm)
    )

    expect_invisible(suppressMessages(print(x)))
    expect_identical(suppressMessages(print(x)), x, info = paste("print method for", nm))
  }
})

test_that("repl_urls works for every link-bearing class", {
  objs <- sample_objects()

  for (nm in names(objs)) {
    urls <- repl_urls(objs[[nm]])

    expect_type(urls, "character")
    expect_true(all(grepl("^https://", urls)), info = paste("repl_urls for", nm))
  }
})

test_that("as.character works for every class", {
  objs <- sample_objects()

  for (nm in names(objs)) {
    txt <- as.character(objs[[nm]])

    expect_type(txt, "character")
    expect_true(all(grepl("^https://", txt)), info = paste("as.character for", nm))
  }
})

test_that("format returns the print representation as a character vector", {
  objs <- sample_objects()

  for (nm in names(objs)) {
    out <- format(objs[[nm]])

    expect_type(out, "character")
    expect_gt(length(out), 0)
    # The URL should appear somewhere in the formatted lines.
    expect_true(any(grepl("https://", out)), info = paste("format for", nm))
  }
})

test_that("panel modes render consistently across print methods", {
  # Regression: print.webr_link used to show 'plot-editor' while print.webr_decoded
  # showed 'Plot, Editor'. Both now route through format_mode_for_display().
  link <- webr_repl_link("plot(1:10)", panels = c("editor", "plot"))
  interface <- grep("Interface", format(link), value = TRUE)

  expect_match(interface, "Editor")
  expect_match(interface, "Plot")
  expect_no_match(interface, "editor-plot")
})

test_that("as.data.frame turns a directory into filename + url rows", {
  dir <- withr::local_tempdir()
  writeLines("plot(1:10)",     file.path(dir, "one.R"))
  writeLines("hist(rnorm(5))", file.path(dir, "two.R"))

  df <- as.data.frame(suppressMessages(webr_repl_directory(dir)))

  expect_s3_class(df, "data.frame")
  expect_named(df, c("filename", "url"))
  expect_setequal(df$filename, c("one.R", "two.R"))
  expect_true(all(grepl("^https://", df$url)))
})

test_that("as.data.frame turns a decoded batch into one row per decode", {
  urls <- c(
    as.character(webr_repl_link("a <- 1", filename = "a.R")),
    as.character(webr_repl_link("b <- 2", filename = "b.R"))
  )
  batch <- suppressMessages(
    decode_webr_link(urls, output_dir = withr::local_tempdir())
  )

  df <- as.data.frame(batch)

  expect_s3_class(df, "data.frame")
  expect_true(all(c("url", "total_files", "total_size") %in% names(df)))
  expect_equal(nrow(df), 2)
})

test_that("as.data.frame stays type-stable on an empty directory", {
  empty <- withr::local_tempdir()
  df <- as.data.frame(suppressWarnings(webr_repl_directory(empty)))

  expect_s3_class(df, "data.frame")
  expect_named(df, c("filename", "url"))
  expect_equal(nrow(df), 0)
})

test_that("knit_print renders link classes as clickable markdown", {
  skip_if_not_installed("knitr")

  link <- webr_repl_link("plot(1:10)")
  out <- knitr::knit_print(link)

  expect_s3_class(out, "knit_asis")
  expect_match(paste(out, collapse = ""), "[Open in webR](https://", fixed = TRUE)

  # A directory becomes a bullet per named link.
  dir <- withr::local_tempdir()
  writeLines("plot(1:10)", file.path(dir, "one.R"))
  dir_out <- knitr::knit_print(suppressMessages(webr_repl_directory(dir)))
  expect_match(paste(dir_out, collapse = ""), "- [one.R](https://", fixed = TRUE)
})

test_that("repl_urls rejects unsupported objects", {
  expect_error(repl_urls(1:10), "Cannot extract URLs")
  expect_error(repl_urls("https://example.org"), "Cannot extract URLs")
})

test_that("print.webr_preview honors show_content", {
  url <- as.character(webr_repl_link("secret_marker <- 42"))
  preview <- preview_webr_link(url)

  expect_snapshot(print(preview))
  expect_snapshot(print(preview, show_content = TRUE))
})

test_that("print.shinylive_preview honors show_content", {
  url <- as.character(shinylive_r_link("secret_marker <- 42"))
  preview <- preview_shinylive_link(url)

  expect_snapshot(print(preview))
  expect_snapshot(print(preview, show_content = TRUE))
})

test_that("print methods render without leaking glue templates", {
  objs <- sample_objects()

  for (nm in names(objs)) {
    out <- paste(
      utils::capture.output(suppressMessages(print(objs[[nm]]))),
      collapse = "\n"
    )
    # An unevaluated {expr} in the output means cli globbed a template.
    expect_false(grepl("\\{[^}]*\\(", out), info = paste("glue template leaked in", nm))
  }
})

# Multi-file projects can name each file's contents as a braced expression rather
# than a string full of escaped newlines.

contents_of <- function(link) {
  preview <- preview_webr_link(as.character(link))
  stats::setNames(
    lapply(preview$files_data, function(f) f$text),
    vapply(preview$files_data, function(f) f$name, character(1))
  )
}

test_that("a project accepts braced expressions as file contents", {
  link <- webr_repl_project(list(
    "main.R"  = { source("utils.R"); summarise(mtcars) },
    "utils.R" = { summarise <- function(d) summary(d) }
  ))

  files <- contents_of(link)

  expect_setequal(names(files), c("main.R", "utils.R"))
  expect_match(files[["main.R"]], 'source("utils.R")', fixed = TRUE)
  expect_match(files[["main.R"]], "summarise(mtcars)", fixed = TRUE)
  expect_match(files[["utils.R"]], "summarise <- function(d) summary(d)",
               fixed = TRUE)
  # The braces themselves must not travel into the shared file.
  expect_false(startsWith(trimws(files[["main.R"]]), "{"))
})

test_that("expressions and strings can be mixed in one project", {
  link <- webr_repl_project(list(
    "main.R"    = { plot(1:10) },
    "README.md" = "# Analysis\n\nRun main.R to start."
  ))

  files <- contents_of(link)

  expect_equal(files[["main.R"]], "plot(1:10)")
  expect_equal(files[["README.md"]], "# Analysis\n\nRun main.R to start.")
})

# The braces must never be evaluated: `{ x <- 1 }` inside list() would otherwise
# run in the caller's environment and leak the binding.
test_that("braced file contents are not evaluated", {
  marker_env <- environment()

  webr_repl_project(list(
    "a.R" = { leaked_marker <- 99 }
  ))

  expect_false(exists("leaked_marker", envir = marker_env, inherits = FALSE))
})

test_that("values are still evaluated when they are not braces", {
  code <- "x <- 1"
  helper <- "y <- 2"

  link <- webr_repl_project(list("a.R" = code, "b.R" = helper))
  files <- contents_of(link)

  expect_equal(files[["a.R"]], "x <- 1")
  expect_equal(files[["b.R"]], "y <- 2")
})

test_that("a list held in a variable still works", {
  project <- list("a.R" = "x <- 1", "b.R" = "y <- 2")

  link <- webr_repl_project(project)

  expect_setequal(names(contents_of(link)), c("a.R", "b.R"))
})

test_that("file paths still work", {
  dir <- withr::local_tempdir()
  writeLines("x <- 1", file.path(dir, "a.R"))
  writeLines("y <- 2", file.path(dir, "b.R"))

  link <- webr_repl_project(c(file.path(dir, "a.R"), file.path(dir, "b.R")))

  expect_setequal(names(contents_of(link)), c("a.R", "b.R"))
})

test_that("autorun_files still matches a braced file", {
  link <- webr_repl_project(
    list("main.R" = { plot(1:10) }, "utils.R" = { f <- function() 42 }),
    autorun_files = "main.R"
  )

  expect_equal(preview_webr_link(as.character(link))$autorun_files, "main.R")
})

test_that("an unnamed list is still rejected", {
  expect_error(webr_repl_project(list({ plot(1:10) })), "named list")
})

# Rejecting a malformed list must not run the code it carries: the fall-through
# used to force the promise, executing the blocks in the caller's frame (and,
# with plot() in a block, leaving Rplots.pdf in the check directory on CRAN).
test_that("rejecting an unnamed literal list does not evaluate its blocks", {
  ran <- FALSE
  expect_error(webr_repl_project(list({ ran <- TRUE })), "named list")
  expect_false(ran)

  expect_error(shinylive_project(list({ ran <- TRUE }), engine = "r"), "named list")
  expect_false(ran)
})

# `list()` is an ordinary call, so assigning it to a variable first evaluates the
# blocks -- livelink never sees the source, and receives whatever they returned.
# Say so, rather than encoding a function into a link.
test_that("a list whose values are not strings is rejected with a useful error", {
  assigned <- list("a.R" = function() 42)

  expect_error(webr_repl_project(assigned), "must be a single string")
  expect_error(webr_repl_project(assigned), "inside the call")
})

test_that("the error names the offending file", {
  err <- tryCatch(
    webr_repl_project(list("ok.R" = "x <- 1", "bad.R" = 1:10)),
    error = function(e) conditionMessage(e)
  )

  expect_match(err, "bad.R", fixed = TRUE)
  expect_false(grepl("ok.R", err, fixed = TRUE))
})

# ---- Shinylive ------------------------------------------------------------

test_that("shinylive_project accepts braced expressions", {
  link <- shinylive_project(
    list(
      "app.R"   = { shinyApp(ui, server) },
      "utils.R" = { helper <- function() 42 }
    ),
    engine = "r"
  )

  expect_s3_class(link, "shinylive_project")
  expect_setequal(link$files, c("app.R", "utils.R"))

  files <- preview_shinylive_link(as.character(link))$files_data
  named <- stats::setNames(lapply(files, function(f) f$content),
                           vapply(files, function(f) f$name, character(1)))
  expect_equal(named[["utils.R"]], "helper <- function() 42")
})

test_that("shinylive_r_link accepts braced expressions in a named list", {
  link <- shinylive_r_link(list(
    "app.R"   = { shinyApp(ui, server) },
    "utils.R" = { helper <- function() 42 }
  ))

  expect_setequal(link$files, c("app.R", "utils.R"))
})

#' Format file size in human readable format
#' @param size_bytes Size in bytes
#' @return Formatted string
#' @noRd
format_file_size <- function(size_bytes) {
  if (is.na(size_bytes) || size_bytes == 0) {
    return("0 bytes")
  } else if (size_bytes < 1024) {
    paste(size_bytes, "bytes")
  } else if (size_bytes < 1024^2) {
    paste(round(size_bytes/1024, 1), "KB")
  } else {
    paste(round(size_bytes/1024^2, 1), "MB")
  }
}

#' Render a file list for display, truncating long ones
#'
#' The count is interpolated here rather than in the cli template, because cli
#' globs the template string it is handed -- an unevaluated `{length(x) - 3}`
#' would reach the user verbatim.
#'
#' @param files Character vector of filenames
#' @param max Number of names to show before truncating (default: 3)
#' @return A single string
#' @noRd
truncate_file_list <- function(files, max = 3) {
  if (length(files) > max) {
    paste0(
      paste(files[seq_len(max)], collapse = ", "),
      ", and ", length(files) - max, " more"
    )
  } else {
    paste(files, collapse = ", ")
  }
}

#' Calculate the decoded size of base64 data without decoding
#' @param base64_string Base64 encoded string
#' @return Integer size in bytes
#' @noRd
calculate_base64_size <- function(base64_string) {
  # Remove any whitespace that might be present
  base64_string <- gsub("\\s", "", base64_string)

  # Count padding characters ('=')
  padding_count <- sum(strsplit(base64_string, "")[[1]] == "=")

  # Base64 encoding: every 4 characters represent 3 bytes of original data
  # Formula: (length * 3 / 4) - padding_count
  decoded_size <- (nchar(base64_string) * 3) / 4 - padding_count

  return(as.integer(decoded_size))
}

#' Create a simple hash from a string using base R
#'
#' Polynomial rolling hash reduced modulo 2^32 on every step. The intermediate
#' value therefore never leaves the range where a double represents integers
#' exactly, which matters because R silently promotes an integer sum past
#' `.Machine$integer.max` to double -- and `sprintf("%x", <double>)` is an error,
#' not a fallback.
#'
#' @param x String to hash
#' @param length Length of hash to return (default: 8)
#' @return Character string hash of `length` hex digits
#' @noRd
simple_hash <- function(x, length = 8) {
  bytes <- utf8ToInt(enc2utf8(x))
  bytes <- bytes[!is.na(bytes)]

  modulus <- 2^32
  h <- 0
  for (b in bytes) {
    h <- (h * 31 + b) %% modulus
  }

  # Split into 16-bit halves so both fit in an integer before formatting.
  hex_string <- sprintf("%04x%04x", as.integer(h %/% 65536), as.integer(h %% 65536))

  if (nchar(hex_string) < length) {
    hex_string <- strrep(hex_string, ceiling(length / nchar(hex_string)))
  }

  substr(hex_string, 1, length)
}

test_that("simple_hash returns a fixed-width hex string for short input", {
  h <- simple_hash("hello world")

  expect_type(h, "character")
  expect_length(h, 1)
  expect_equal(nchar(h), 8)
  expect_match(h, "^[0-9a-f]+$")
})

test_that("simple_hash honors the length argument", {
  expect_equal(nchar(simple_hash("hello world", length = 4)), 4)
  expect_equal(nchar(simple_hash("hello world", length = 16)), 16)
})

test_that("simple_hash is deterministic", {
  expect_identical(simple_hash("abc"), simple_hash("abc"))
  expect_false(identical(simple_hash("abc"), simple_hash("abd")))
})

# Regression: the weighted sum overflowed .Machine$integer.max and was silently
# promoted to double, which sprintf("%x", .) rejects. This broke decoding for
# any sharelink built from a non-trivial script.
test_that("simple_hash does not overflow on long input", {
  for (n in c(8000, 20000, 100000)) {
    big <- strrep("x", n)

    expect_match(simple_hash(big), "^[0-9a-f]{8}$")
  }
})

test_that("simple_hash handles empty and unicode input", {
  expect_match(simple_hash(""), "^[0-9a-f]{8}$")
  expect_match(simple_hash("héllo 你好"), "^[0-9a-f]{8}$")
})

test_that("format_file_size renders bytes, KB, and MB", {
  expect_equal(format_file_size(0), "0 bytes")
  expect_equal(format_file_size(512), "512 bytes")
  expect_equal(format_file_size(2048), "2 KB")
  expect_equal(format_file_size(2 * 1024^2), "2 MB")
})

# Internal validation helpers.
#
#   is_*()                 -> TRUE/FALSE
#   check_*() / ensure_*() -> invisible(TRUE), or abort with a cli message.

#' Check if version is valid
#' @param version Version string to check
#' @return Logical value
#' @noRd
is_valid_version <- function(version) {
  if (!is.character(version) || length(version) != 1) {
    return(FALSE)
  }

  # Allow "latest"
  if (version == "latest") {
    return(TRUE)
  }

  # Check version format (v0.5.4 or greater)
  if (grepl("^v\\d+\\.\\d+\\.\\d+$", version)) {
    # Extract version numbers
    version_clean <- gsub("^v", "", version)
    version_parts <- as.numeric(strsplit(version_clean, "\\.")[[1]])

    # Check if >= v0.5.4
    if (version_parts[1] > 0) return(TRUE)
    if (version_parts[1] == 0 && version_parts[2] > 5) return(TRUE)
    if (version_parts[1] == 0 && version_parts[2] == 5 && version_parts[3] >= 4) return(TRUE)
  }

  return(FALSE)
}

#' Check if mode is valid
#' @param mode Mode specification (character vector or string)
#' @return Logical value
#' @noRd
is_valid_mode <- function(mode) {
  if (is.null(mode)) {
    return(TRUE)  # NULL is valid (means no mode restriction)
  }

  valid_components <- c("plot", "files", "terminal", "editor")

  if (is.character(mode)) {
    if (length(mode) == 1) {
      # String format like "plot-files-terminal"
      if (mode == "") return(FALSE)
      components <- strsplit(mode, "-")[[1]]
    } else {
      # Vector format like c("plot", "files", "terminal")
      components <- mode
    }

    # Check all components are valid and unique
    return(all(components %in% valid_components) &&
             length(components) == length(unique(components)) &&
             length(components) > 0)
  }

  return(FALSE)
}

#' Ensure mode is valid
#' @param mode Mode specification to check
#' @param arg_name Name of the argument (for error messages)
#' @return Invisible TRUE if valid, aborts if not
#' @noRd
check_valid_mode <- function(mode, arg_name = "mode") {
  if (!is_valid_mode(mode)) {
    cli::cli_abort(c(
      "Invalid {.arg {arg_name}} argument",
      "x" = "{.arg {arg_name}} must be NULL, a character vector of valid components, or a string with '-' separators",
      "i" = "Valid components: {.val {c('plot', 'files', 'terminal', 'editor')}}",
      "i" = "Examples: {.code c('plot', 'files')} or {.code 'plot-files-terminal'}"
    ))
  }
  invisible(TRUE)
}


#' Check if webR version is valid
#'
#' @param version Version string to check
#' @param arg_name Name of the argument (for error messages)
#'
#' @return
#' Invisible TRUE if valid, aborts if not
#'
#' @details
#' This function checks if the provided version is either "latest" or a valid
#' version string in the format "vX.Y.Z" where X, Y, Z are integers and the version is
#' greater than or equal to v0.5.4.
#'
#' @noRd
check_valid_version <- function(version, arg_name = "version") {
  if (!is_valid_version(version)) {
    cli::cli_abort(c(
      "Invalid {.arg {arg_name}} argument",
      "x" = "{.arg {arg_name}} must be 'latest' or a version >= v0.5.4 (e.g., 'v0.5.4', 'v1.0.0')",
      "i" = "You provided: {.val {version}}"
    ))
  }
  invisible(TRUE)
}

#' Check if input is a single character string
#' @param x Object to check
#' @return Logical value
#' @noRd
is_single_string <- function(x) {
  is.character(x) && length(x) == 1 && !is.na(x) && nzchar(x)
}

#' Ensure input is a single character string
#' @param x Object to check
#' @param arg_name Name of the argument (for error messages)
#' @return Invisible TRUE if valid, aborts if not
#' @noRd
check_single_string <- function(x, arg_name) {
  if (!is_single_string(x)) {
    cli::cli_abort(c(
      "Invalid {.arg {arg_name}} argument",
      "x" = "{.arg {arg_name}} must be a single non-empty character string",
      "i" = "You provided {.type {x}} of length {length(x)}"
    ))
  }
  invisible(TRUE)
}

#' Check if input is a single logical value
#' @param x Object to check
#' @return Logical value
#' @noRd
is_single_logical <- function(x) {
  is.logical(x) && length(x) == 1 && !is.na(x)
}

#' Ensure input is a single logical value
#' @param x Object to check
#' @param arg_name Name of the argument (for error messages)
#' @return Invisible TRUE if valid, aborts if not
#' @noRd
check_single_logical <- function(x, arg_name) {
  if (!is_single_logical(x)) {
    cli::cli_abort(c(
      "Invalid {.arg {arg_name}} argument",
      "x" = "{.arg {arg_name}} must be a single logical value (TRUE or FALSE)",
      "i" = "You provided {.type {x}} of length {length(x)}"
    ))
  }
  invisible(TRUE)
}

#' Check if input is a properly named list
#' @param x Object to check
#' @return Logical value
#' @noRd
is_named_list <- function(x) {
  is.list(x) && !is.null(names(x)) && all(names(x) != "") && all(nzchar(names(x)))
}

#' Check if input is a valid file path
#'
#' A file path is, structurally, a single non-empty string; the distinct name
#' and [check_valid_path()] wrapper exist only for the path-specific error text.
#' @param x Object to check
#' @return Logical value
#' @noRd
is_valid_path <- function(x) {
  is_single_string(x)
}

#' Ensure input is a valid file path
#' @param x Object to check
#' @param arg_name Name of the argument (for error messages)
#' @return Invisible TRUE if valid, aborts if not
#' @noRd
check_valid_path <- function(x, arg_name) {
  if (!is_valid_path(x)) {
    cli::cli_abort(c(
      "Invalid {.arg {arg_name}} argument",
      "x" = "{.arg {arg_name}} must be a single non-empty character string representing a file path",
      "i" = "You provided {.type {x}} of length {length(x)}"
    ))
  }
  invisible(TRUE)
}

#' Check if input is a character vector
#'
#' @param x Object to check
#' @return
#' Logical value
#'
#' @noRd
is_character_vector <- function(x) {
  is.character(x)
}

#' Ensure input is a character vector
#'
#' @param x Object to check
#' @param arg_name Name of the argument (for error messages)
#'
#' @return
#' Invisible TRUE if valid, aborts if not
#'
#' @noRd
check_character_vector <- function(x, arg_name) {
  if (!is_character_vector(x)) {
    cli::cli_abort(c(
      "Invalid {.arg {arg_name}} argument",
      "x" = "{.arg {arg_name}} must be a character vector",
      "i" = "You provided {.type {x}}"
    ))
  }
  invisible(TRUE)
}

#' Check if directory exists
#'
#' @param path Directory path to check
#'
#' @return
#' Logical value
#'
#' @noRd
has_directory <- function(path) {
  dir.exists(path)
}

#' Ensure directory exists
#' @param path Directory path to check
#' @param arg_name Name of the argument (for error messages)
#' @return Invisible TRUE if valid, aborts if not
#' @noRd
ensure_directory_exists <- function(path, arg_name = "directory_path") {
  if (!has_directory(path)) {
    cli::cli_abort(c(
      "Directory does not exist",
      "x" = "Cannot find directory {.path {path}}",
      "i" = "Please check the path and try again"
    ))
  }
  invisible(TRUE)
}

#' Ensure specified files exist in a named list
#' @param files_to_check Character vector of filenames to check
#' @param files_list Named list of files
#' @param files_arg_name Name of the files argument (for error messages)
#' @param check_arg_name Name of the checking argument (for error messages)
#' @return Invisible TRUE if valid, aborts if not
#' @noRd
ensure_files_in_list <- function(files_to_check, files_list,
                                 files_arg_name = "files",
                                 check_arg_name = "autorun_files") {
  missing_files <- setdiff(files_to_check, names(files_list))
  if (length(missing_files) > 0) {
    cli::cli_abort(c(
      "Files specified in {.arg {check_arg_name}} not found in {.arg {files_arg_name}}",
      "x" = "Missing files: {.file {missing_files}}",
      "i" = "Available files: {.file {names(files_list)}}"
    ))
  }
  invisible(TRUE)
}

#' Check if engine is valid for Shinylive
#' @param engine Engine specification ("r" or "python")
#' @return Logical value
#' @noRd
is_valid_shinylive_engine <- function(engine) {
  is.character(engine) && length(engine) == 1 && engine %in% c("r", "python")
}

#' Ensure engine is valid for Shinylive
#' @param engine Engine specification to check
#' @param arg_name Name of the argument (for error messages)
#' @return Invisible TRUE if valid, aborts if not
#' @noRd
check_valid_shinylive_engine <- function(engine, arg_name = "engine") {
  if (!is_valid_shinylive_engine(engine)) {
    cli::cli_abort(c(
      "Invalid {.arg {arg_name}} argument",
      "x" = "{.arg {arg_name}} must be either 'r' or 'python'",
      "i" = "You provided: {.val {engine}}"
    ))
  }
  invisible(TRUE)
}

#' Check if Shinylive mode is valid
#' @param mode Mode specification ("editor" or "app")
#' @return Logical value
#' @noRd
is_valid_shinylive_mode <- function(mode) {
  is.character(mode) && length(mode) == 1 && mode %in% c("editor", "app")
}

#' Ensure Shinylive mode is valid
#' @param mode Mode specification to check
#' @param arg_name Name of the argument (for error messages)
#' @return Invisible TRUE if valid, aborts if not
#' @noRd
check_valid_shinylive_mode <- function(mode, arg_name = "mode") {
  if (!is_valid_shinylive_mode(mode)) {
    cli::cli_abort(c(
      "Invalid {.arg {arg_name}} argument",
      "x" = "{.arg {arg_name}} must be either 'editor' or 'app'",
      "i" = "You provided: {.val {mode}}"
    ))
  }
  invisible(TRUE)
}

#' Check if URL is a valid Shinylive URL
#' @param url URL string to check
#' @return Logical value
#' @noRd
is_valid_shinylive_url <- function(url) {
  if (!is_single_string(url)) {
    return(FALSE)
  }

  grepl("^https://shinylive\\.io/(r|py)/(editor|app)/#", url)
}

#' Ensure URL is a valid Shinylive URL
#' @param url URL to check
#' @param arg_name Name of the argument (for error messages)
#' @return Invisible TRUE if valid, aborts if not
#' @noRd
check_valid_shinylive_url <- function(url, arg_name = "url") {
  if (!is_valid_shinylive_url(url)) {
    cli::cli_abort(c(
      "Invalid {.arg {arg_name}} argument",
      "x" = "{.arg {arg_name}} must be a valid Shinylive URL",
      "i" = "Expected format: https://shinylive.io/[r|py]/[editor|app]/#...",
      "i" = "You provided: {.val {url}}"
    ))
  }
  invisible(TRUE)
}

#' Check if URL is a valid webR URL
#' @param url URL string to check
#' @return Logical value
#' @noRd
is_valid_webr_url <- function(url) {
  if (!is_single_string(url)) {
    return(FALSE)
  }

  # Check if URL contains webR domain and has a fragment with code parameter
  has_webr_domain <- grepl("webr\\.r-wasm\\.org", url, ignore.case = TRUE)
  has_fragment <- grepl("#", url)
  has_code_param <- grepl("#.*code=", url)

  return(has_webr_domain && has_fragment && has_code_param)
}

#' Ensure URL is a valid webR URL
#' @param url URL string to check
#' @param arg_name Name of the argument (for error messages)
#' @return Invisible TRUE if valid, aborts if not
#' @noRd
check_valid_webr_url <- function(url, arg_name = "url") {
  if (!is_valid_webr_url(url)) {
    cli::cli_abort(c(
      "Invalid webR URL",
      "x" = "{.arg {arg_name}} must be a valid webR REPL sharelink",
      "i" = "URL should be from webr.r-wasm.org and contain encoded data",
      "i" = "Example: https://webr.r-wasm.org/latest/#code=..."
    ))
  }
  invisible(TRUE)
}

# Tests for validator functions

test_that("is_single_string works correctly", {
  expect_true(livelink:::is_single_string("hello"))
  expect_true(livelink:::is_single_string("a"))

  expect_false(livelink:::is_single_string(""))

expect_false(livelink:::is_single_string(c("a", "b")))
  expect_false(livelink:::is_single_string(123))
  expect_false(livelink:::is_single_string(NULL))
  expect_false(livelink:::is_single_string(NA_character_))
})

test_that("check_single_string aborts on invalid input", {
  expect_error(livelink:::check_single_string("", "arg"))
  expect_error(livelink:::check_single_string(123, "arg"))
  expect_error(livelink:::check_single_string(NULL, "arg"))

  expect_invisible(livelink:::check_single_string("valid", "arg"))
})

test_that("is_single_logical works correctly", {
  expect_true(livelink:::is_single_logical(TRUE))
  expect_true(livelink:::is_single_logical(FALSE))

  expect_false(livelink:::is_single_logical(c(TRUE, FALSE)))
  expect_false(livelink:::is_single_logical("TRUE"))
  expect_false(livelink:::is_single_logical(1))
  expect_false(livelink:::is_single_logical(NA))
})

test_that("check_single_logical aborts on invalid input", {
  expect_error(livelink:::check_single_logical("TRUE", "arg"))
  expect_error(livelink:::check_single_logical(NA, "arg"))

  expect_invisible(livelink:::check_single_logical(TRUE, "arg"))
})

test_that("is_valid_version works correctly", {
  expect_true(livelink:::is_valid_version("latest"))
  expect_true(livelink:::is_valid_version("v0.5.4"))
  expect_true(livelink:::is_valid_version("v0.5.5"))
  expect_true(livelink:::is_valid_version("v1.0.0"))

  expect_false(livelink:::is_valid_version("v0.5.3"))
  expect_false(livelink:::is_valid_version("v0.4.0"))
  expect_false(livelink:::is_valid_version("0.5.4"))
  expect_false(livelink:::is_valid_version("invalid"))
})

test_that("is_valid_mode works correctly", {
  expect_true(livelink:::is_valid_mode(NULL))
  expect_true(livelink:::is_valid_mode("editor"))
  expect_true(livelink:::is_valid_mode("plot"))
  expect_true(livelink:::is_valid_mode("editor-plot"))
  expect_true(livelink:::is_valid_mode(c("editor", "plot")))
  expect_true(livelink:::is_valid_mode(c("plot", "files", "terminal", "editor")))

  expect_false(livelink:::is_valid_mode("invalid"))
  expect_false(livelink:::is_valid_mode(""))
  expect_false(livelink:::is_valid_mode(c("editor", "editor"))) # duplicates
})

test_that("is_named_list works correctly", {
  expect_true(livelink:::is_named_list(list(a = 1, b = 2)))
  expect_true(livelink:::is_named_list(list("file.R" = "code")))

  expect_false(livelink:::is_named_list(list(1, 2)))
  expect_false(livelink:::is_named_list(list(a = 1, 2)))
  expect_false(livelink:::is_named_list(c(a = 1, b = 2)))
  expect_false(livelink:::is_named_list(NULL))
})

test_that("is_valid_shinylive_engine works correctly", {
  expect_true(livelink:::is_valid_shinylive_engine("r"))
  expect_true(livelink:::is_valid_shinylive_engine("python"))

  expect_false(livelink:::is_valid_shinylive_engine("R"))
  expect_false(livelink:::is_valid_shinylive_engine("Python"))
  expect_false(livelink:::is_valid_shinylive_engine("julia"))
})

test_that("is_valid_shinylive_mode works correctly", {
  expect_true(livelink:::is_valid_shinylive_mode("editor"))
  expect_true(livelink:::is_valid_shinylive_mode("app"))

  expect_false(livelink:::is_valid_shinylive_mode("Editor"))
  expect_false(livelink:::is_valid_shinylive_mode("both"))
})

test_that("is_valid_webr_url works correctly", {
  valid_url <- "https://webr.r-wasm.org/latest/#code=abc123"
  expect_true(livelink:::is_valid_webr_url(valid_url))

  expect_false(livelink:::is_valid_webr_url("https://example.com"))
  expect_false(livelink:::is_valid_webr_url("https://webr.r-wasm.org/latest/"))
  expect_false(livelink:::is_valid_webr_url(""))
  expect_false(livelink:::is_valid_webr_url(NULL))
})

test_that("is_valid_shinylive_url works correctly", {
  valid_r_url <- "https://shinylive.io/r/editor/#code=abc"
  valid_py_url <- "https://shinylive.io/py/app/#code=xyz"

  expect_true(livelink:::is_valid_shinylive_url(valid_r_url))
  expect_true(livelink:::is_valid_shinylive_url(valid_py_url))

  expect_false(livelink:::is_valid_shinylive_url("https://example.com"))
  expect_false(livelink:::is_valid_shinylive_url("https://shinylive.io/r/"))
})

#' Convert mode to string format
#' @param mode Mode specification (character vector, string, or NULL)
#' @return String in format "component1-component2-..." or NULL
#' @noRd
format_mode_string <- function(mode) {
  if (is.null(mode)) {
    return(NULL)
  }

  if (length(mode) == 1) {
    # Already a string, return as-is
    return(mode)
  } else {
    # Convert vector to string
    return(paste(mode, collapse = "-"))
  }
}

#' Build WebR URL with mode and code
#' @param base_url Base WebR URL
#' @param encoded_data Base64 encoded data
#' @param flags URL flags
#' @param mode Mode string or NULL
#' @return Complete WebR URL
#' @noRd
build_webr_url <- function(base_url, encoded_data, flags, mode = NULL) {
  if (is.null(mode)) {
    # No mode: a bare "#code=..." fragment
    paste0(base_url, "#code=", encoded_data, "&", flags)
  } else {
    # With mode: a "?mode='...'" query ahead of the "#code=" fragment
    mode_string <- format_mode_string(mode)
    paste0(base_url, "?mode='", mode_string, "'#code=", encoded_data, "&", flags)
  }
}

#' Encode webR share items into a URL-fragment payload
#'
#' The shared encoding tail for webR links: serialize the files to JSON, gzip,
#' and base64. `URLencode(reserved = TRUE)` is the non-obvious step: it
#' percent-escapes base64's `+` and `/`, which are URL-reserved and would
#' otherwise corrupt the `#code=` fragment.
#' @param share_items List of file items (each `name`/`path`/`text`, plus an
#'   optional `autorun`)
#' @return A URL-safe encoded string
#' @noRd
encode_webr_payload <- function(share_items) {
  json_data <- jsonlite::toJSON(share_items, auto_unbox = TRUE)
  compressed <- memCompress(charToRaw(json_data), type = "gzip")
  base64_data <- base64enc::base64encode(compressed)
  utils::URLencode(base64_data, reserved = TRUE)
}

#' Turn document chunks into shareable links
#'
#' @description
#' livelink plugs into 'knitr' two ways, and which you want depends on one
#' question: **should the code also run in your document?**
#'
#' **A chunk hook**, on an ordinary `r` chunk. The chunk runs as usual -- its
#' output, plots and all, appear in the rendered page -- and a link is added
#' underneath. Use this for code you want your reader to *see the result of* and
#' *also* be able to open and play with.
#'
#' ````
#' ```{r}
#' #| livelink: true
#' #| autorun: true
#' # Load the data
#' data(mtcars)
#' plot(mtcars$mpg, mtcars$wt)
#' ```
#' ````
#'
#' **An engine**, as ```` ```{livelink} ````. The chunk is displayed but **not**
#' run: only the link is produced. Use this for code your session cannot or should
#' not execute -- a Shiny app, something needing a package you have not installed,
#' or anything slow.
#'
#' ````
#' ```{livelink}
#' #| engine.target: shinylive-r
#' library(shiny)
#' shinyApp(fluidPage(), function(input, output) {})
#' ```
#' ````
#'
#' There is deliberately no `{shinylive-r}` or `{shinylive-py}` engine. 'knitr'
#' cannot dispatch a chunk whose engine name contains a hyphen (its chunk syntax
#' forbids it), and in Quarto such a cell is handed to the Shinylive extension
#' rather than to 'knitr'. Name Shinylive through `engine.target` instead.
#'
#' @details
#' Reach for either of these rather than expression input
#' (`webr_repl_link({ ... })`) inside a knitted document. 'knitr' evaluates
#' chunks through `evaluate::evaluate()`, which discards source references, and
#' comments live in those -- so comments inside a `{ }` expression are silently
#' dropped from the link when the document renders, and no `keep.source` setting
#' brings them back. Both the hook and the engine are handed the chunk's verbatim
#' source, so nothing is lost.
#'
#' Both are registered automatically when livelink is loaded, provided 'knitr' is
#' installed. Call these yourself only if you have reset `knitr::knit_hooks` or
#' `knitr::knit_engines`.
#'
#' # Chunk options
#'
#' \describe{
#'   \item{`livelink`}{Hook only. `true` for a webR link, or name the target
#'     directly: `"webr"`, `"shinylive-r"`, `"shinylive-py"`.}
#'   \item{`engine.target`}{Engine only. `"webr"` (default), `"shinylive-r"`, or
#'     `"shinylive-py"`.}
#'   \item{`autorun`}{Logical. Run the code as soon as the link opens. webR only.}
#'   \item{`panels`}{Character vector of webR panels, e.g. `c("editor", "plot")`.}
#'   \item{`mode`}{Shinylive only. Display mode, `"editor"` (default) or `"app"`.}
#'   \item{`filename`}{webR only. Name for the script file in the webR virtual
#'     filesystem (default `"script.R"`); must end in `.R` for `autorun` to work.}
#'   \item{`link.text`}{Text for the hyperlink. Defaults to `"Open in webR"` or
#'     `"Open in Shinylive"`.}
#'   \item{`link.only`}{Engine only. If `TRUE`, emit the link without the source.}
#' }
#'
#' # Setting options once
#'
#' These are ordinary 'knitr' chunk options, so `opts_chunk` sets them for a whole
#' document, and a single chunk opts out with `livelink: false`:
#'
#' ```
#' knitr::opts_chunk$set(livelink = TRUE, autorun = TRUE)
#' ```
#'
#' # `echo` does not gate the link
#'
#' It is natural to assume the code must be visible for a link to be made. It need
#' not be. `echo` controls whether the **source is shown in your page**; the link is
#' built from the chunk's source, which 'knitr' hands over either way. So
#' `echo: false` gives a working link whose code the reader simply cannot see.
#'
#' `eval: false` is the other half: the chunk is displayed but not run, which makes
#' an `r` chunk behave rather like the engine.
#'
#' @return Called for their side effect. Invisibly `TRUE` if registration
#'   happened, `FALSE` if 'knitr' is not installed.
#'
#' @seealso `vignette("links-in-documents", package = "livelink")` for the whole
#'   picture, and [webr_repl_link()] for why a braced expression loses comments in
#'   a knitted document.
#'
#' @examplesIf requireNamespace("knitr", quietly = TRUE)
#' # Both are registered on load; call directly only after resetting knitr's hooks.
#' use_livelink_hook()
#' use_livelink_engine()
#'
#' @name livelink-knitr
NULL

#' @rdname livelink-knitr
#' @export
use_livelink_hook <- function() {
  if (!requireNamespace("knitr", quietly = TRUE)) {
    return(invisible(FALSE))
  }

  knitr::knit_hooks$set(livelink = livelink_hook)
  invisible(TRUE)
}

#' @rdname livelink-knitr
#' @export
use_livelink_engine <- function() {
  if (!requireNamespace("knitr", quietly = TRUE)) {
    return(invisible(FALSE))
  }

  # A single engine, `livelink`, that targets webR by default. Shinylive is
  # reached with the `engine.target` chunk option, not a separate engine name:
  # 'knitr' cannot dispatch a `{shinylive-r}` chunk (its chunk syntax forbids a
  # hyphen in the engine name), and an underscore near-miss like `{shinylive_r}`
  # only misleads.
  knitr::knit_engines$set(livelink = livelink_engine)
  invisible(TRUE)
}

#' Resolve the link target from a chunk option
#'
#' @param x Value of the `livelink` or `engine.target` chunk option
#' @return One of "webr", "shinylive-r", "shinylive-py"
#' @noRd
livelink_target <- function(x) {
  target <- if (isTRUE(x)) {
    "webr"
  } else if (is.character(x) && length(x) == 1) {
    x
  } else {
    "webr"
  }

  if (!target %in% c("webr", "shinylive-r", "shinylive-py")) {
    cli::cli_abort(c(
      "Invalid livelink chunk target",
      "x" = "Got {.val {target}}",
      "i" = "Valid targets: {.val webr}, {.val shinylive-r}, {.val shinylive-py}"
    ))
  }

  target
}

#' Build the link a chunk asks for
#'
#' @param code The chunk's source, as one string
#' @param options knitr chunk options
#' @param target Resolved target
#' @return A link object
#' @noRd
livelink_from_chunk <- function(code, options, target) {
  switch(target,
    "webr" = webr_repl_link(
      code,
      filename = options[["filename"]] %||% "script.R",
      autorun  = isTRUE(options[["autorun"]]),
      panels   = options[["panels"]]
    ),
    "shinylive-r" = shinylive_r_link(
      code,
      mode = options[["mode"]] %||% "editor"
    ),
    "shinylive-py" = shinylive_py_link(
      code,
      mode = options[["mode"]] %||% "editor"
    )
  )
}

#' The markdown for a chunk's link
#'
#' A blank line either side keeps it a paragraph of its own, so the renderer sees
#' a link rather than a run-on of the preceding output.
#'
#' @param link A link object
#' @param options knitr chunk options
#' @param target Resolved target
#' @return A markdown string
#' @noRd
livelink_markdown <- function(link, options, target) {
  default_text <- if (target == "webr") "Open in webR" else "Open in Shinylive"
  link_text <- options[["link.text"]] %||% default_text

  paste0("\n\n[", link_text, "](", as.character(link), ")\n")
}

#' The knitr chunk hook
#'
#' Fires after an ordinary R chunk has run, so the chunk's own output is already
#' in the document; we only append the link. `options$code` is the chunk's
#' verbatim source, comments and all -- which is the only route by which comments
#' reach us from a knitted document.
#'
#' @param before TRUE before the chunk runs, FALSE after
#' @param options knitr chunk options
#' @param envir The chunk's evaluation environment (unused)
#' @return A markdown string appended after the chunk, or NULL
#' @noRd
livelink_hook <- function(before, options, envir) {
  # Nothing to add before the chunk runs, and `livelink: false` opts out.
  if (before || isFALSE(options[["livelink"]])) {
    return(NULL)
  }

  target <- livelink_target(options[["livelink"]])
  code <- paste(options$code, collapse = "\n")
  link <- livelink_from_chunk(code, options, target)

  livelink_markdown(link, options, target)
}

#' The knitr engine callback
#'
#' The chunk is shown but never run, so this is the tool for code the session
#' cannot execute. As with the hook, `options$code` is the verbatim source.
#'
#' @param options knitr chunk options
#' @return A string of knitr output
#' @noRd
livelink_engine <- function(options) {
  target <- livelink_target(options[["engine.target"]] %||% "webr")
  code <- paste(options$code, collapse = "\n")
  link <- livelink_from_chunk(code, options, target)

  # The link must travel as `extra`, which engine_output() appends verbatim.
  # Passing it as `out` routes it through the output hook, which prefixes it with
  # the chunk's `comment` string and wraps it in a code block -- so the reader
  # gets a literal `#> [Open in webR](https://...)` instead of a link they can
  # click.
  markdown_link <- livelink_markdown(link, options, target)

  if (isTRUE(options[["link.only"]])) {
    return(knitr::engine_output(options, code = NULL, out = NULL,
                                extra = markdown_link))
  }

  # The fenced source block takes its language from `options$engine`, which is
  # "livelink" here -- a language no highlighter knows. Name the real one so the
  # code is highlighted as the R or Python it is.
  options$engine <- if (target == "shinylive-py") "python" else "r"

  knitr::engine_output(options, code = options$code, out = NULL,
                       extra = markdown_link)
}

`%||%` <- function(x, y) if (is.null(x)) y else x

.onLoad <- function(libname, pkgname) {
  # Registering here (rather than making users call them) means a document can use
  # either mechanism after nothing more than library(livelink).
  use_livelink_hook()
  use_livelink_engine()
}

skip_if_not_installed("knitr")

# Render a document containing a ```{livelink} chunk and hand back the rendered
# markdown. This has to be a real knit: the whole point of the engine is that it
# behaves differently from expression input *under knitr*.
knit_livelink <- function(chunk_header, code_lines) {
  rmd <- c(
    "```{r}",
    "#| include: false",
    "library(livelink)",
    "```",
    "",
    chunk_header,
    code_lines,
    "```"
  )
  input <- withr::local_tempfile(fileext = ".Rmd")
  output <- withr::local_tempfile(fileext = ".md")
  writeLines(rmd, input)
  # Knit from a temp directory so figures and Rplots.pdf land there rather
  # than in tests/testthat: CRAN only allows writes to tempdir() during check.
  withr::local_dir(withr::local_tempdir())
  suppressMessages(knitr::knit(input, output = output, quiet = TRUE))
  readLines(output, warn = FALSE)
}

extract_url <- function(rendered) {
  line <- grep("\\]\\(https://", rendered, value = TRUE)[1]
  sub(".*\\]\\((https://[^)]+)\\).*", "\\1", line)
}

test_that("the engine is registered when livelink is loaded", {
  expect_true("livelink" %in% names(knitr::knit_engines$get()))
})

test_that("the chunk hook is registered when livelink is loaded", {
  expect_true("livelink" %in% names(knitr::knit_hooks$get()))
})

test_that("use_livelink_hook registers the hook", {
  expect_true(use_livelink_hook())
  expect_true("livelink" %in% names(knitr::knit_hooks$get()))
})

# The reason the hook exists: the engine replaces execution, so a ```{livelink}
# chunk shows a link but never any R output. The hook leaves the chunk running.
test_that("a hooked chunk still runs, and gains a link", {
  rendered <- knit_livelink("```{r}\n#| livelink: true", "mean(c(1, 2, 3))")
  text <- paste(rendered, collapse = "\n")

  # the chunk's own output is there ...
  expect_match(text, "[1] 2", fixed = TRUE)
  # ... and so is the link
  expect_match(text, "[Open in webR](https://webr.r-wasm.org/", fixed = TRUE)
})

test_that("comments survive into a hooked chunk's link", {
  rendered <- knit_livelink(
    "```{r}\n#| livelink: true",
    c("# a leading comment", "x <- 1  # a trailing comment")
  )

  code <- preview_webr_link(extract_url(rendered))$files_data[[1]]$text

  expect_match(code, "# a leading comment", fixed = TRUE)
  expect_match(code, "# a trailing comment", fixed = TRUE)
})

test_that("the hook honors autorun and panels", {
  rendered <- knit_livelink(
    "```{r}\n#| livelink: true\n#| autorun: true\n#| panels: [\"editor\", \"plot\"]",
    "plot(1:10)"
  )
  url <- extract_url(rendered)

  expect_match(url, "mode='editor-plot'", fixed = TRUE)
  expect_equal(preview_webr_link(url)$autorun_files, "script.R")
})

test_that("the hook can target Shinylive", {
  rendered <- knit_livelink(
    "```{r}\n#| livelink: shinylive-r\n#| eval: false",
    "library(shiny)"
  )

  expect_match(extract_url(rendered), "^https://shinylive\\.io/r/")
})

test_that("livelink: false opts a chunk out", {
  rendered <- knit_livelink("```{r}\n#| livelink: false", "mean(c(1, 2, 3))")
  text <- paste(rendered, collapse = "\n")

  expect_match(text, "[1] 2", fixed = TRUE)
  expect_false(grepl("Open in webR", text, fixed = TRUE))
})

test_that("an invalid hook target is rejected", {
  expect_error(
    knit_livelink("```{r}\n#| livelink: julia", "x <- 1"),
    "livelink chunk target"
  )
})

test_that("use_livelink_engine registers the engine", {
  expect_true(use_livelink_engine())
  expect_true("livelink" %in% names(knitr::knit_engines$get()))
})

# knitr's chunk syntax forbids a hyphen in the engine name (the chunk-detection
# regex is `[a-zA-Z0-9_]+`), so a `{shinylive-r}` chunk cannot dispatch to a knitr
# engine at all. Shinylive is named through engine.target instead. Guard against a
# well-meaning future re-introduction of a hyphenated engine that would silently
# never fire.
test_that("no hyphenated shinylive engine is registered", {
  engines <- names(knitr::knit_engines$get())
  expect_false(any(grepl("^shinylive", engines)))
})

# The reason this feature exists. knitr runs chunks through evaluate::evaluate(),
# which discards source references, so comments inside a `{ }` expression are lost
# when a document is rendered. The engine receives the chunk's verbatim source, so
# they survive.
test_that("comments in a livelink chunk survive into the generated link", {
  rendered <- knit_livelink(
    "```{livelink}",
    c(
      "# a leading comment",
      "data(mtcars)",
      "plot(mtcars$mpg, mtcars$wt)  # a trailing comment"
    )
  )

  code <- preview_webr_link(extract_url(rendered))$files_data[[1]]$text

  expect_match(code, "# a leading comment", fixed = TRUE)
  expect_match(code, "# a trailing comment", fixed = TRUE)
  expect_match(code, "plot(mtcars$mpg, mtcars$wt)", fixed = TRUE)
})

test_that("the engine emits both the source and a link by default", {
  rendered <- knit_livelink("```{livelink}", "plot(1:10)")
  text <- paste(rendered, collapse = "\n")

  expect_match(text, "plot(1:10)", fixed = TRUE)
  expect_match(text, "[Open in webR](https://webr.r-wasm.org/", fixed = TRUE)
})

# Regression: the link used to travel to engine_output() as `out`, which routes
# it through knitr's output hook. That prefixed it with the chunk's `comment`
# string and wrapped it in a code block, so the reader got a literal
# `#> [Open in webR](https://...)` rather than a link they could click.
test_that("the link is rendered as markdown, not as commented chunk output", {
  rendered <- knit_livelink("```{livelink}", "plot(1:10)")

  expect_false(any(grepl("#>\\s*\\[Open in webR\\]", rendered)))
  expect_true(any(grepl("^\\[Open in webR\\]\\(https://", rendered)))
})

# The fenced block used to be tagged with the engine name, `livelink`, which no
# highlighter knows -- so the code rendered unhighlighted.
test_that("the source block is fenced as the language it actually is", {
  r_chunk <- knit_livelink("```{livelink}", "plot(1:10)")
  expect_true(any(grepl("^```+\\s*r\\s*$", r_chunk)))
  expect_false(any(grepl("^```+\\s*livelink", r_chunk)))

  py_chunk <- knit_livelink(
    "```{livelink, engine.target='shinylive-py'}",
    "from shiny import App"
  )
  expect_true(any(grepl("^```+\\s*python\\s*$", py_chunk)))
})

test_that("link.only suppresses the source chunk", {
  rendered <- knit_livelink(
    "```{livelink, link.only=TRUE}",
    "secret_marker <- 1"
  )
  text <- paste(rendered, collapse = "\n")

  expect_match(text, "https://webr.r-wasm.org/")
  expect_false(grepl("secret_marker", text, fixed = TRUE))
})

test_that("autorun and panels chunk options reach the link", {
  rendered <- knit_livelink(
    "```{livelink, autorun=TRUE, panels=c('editor','plot')}",
    "plot(1:10)"
  )
  url <- extract_url(rendered)

  expect_match(url, "mode='editor-plot'", fixed = TRUE)

  preview <- preview_webr_link(url)
  expect_equal(preview$autorun_files, "script.R")
})

test_that("engine.target routes to Shinylive", {
  rendered <- knit_livelink(
    "```{livelink, engine.target='shinylive-r'}",
    "library(shiny)"
  )
  url <- extract_url(rendered)

  expect_match(url, "^https://shinylive\\.io/r/")
  expect_match(
    paste(rendered, collapse = "\n"),
    "Open in Shinylive",
    fixed = TRUE
  )
})

test_that("engine.target routes to Python Shinylive", {
  rendered <- knit_livelink(
    "```{livelink, engine.target='shinylive-py'}",
    "from shiny import App"
  )

  expect_match(extract_url(rendered), "^https://shinylive\\.io/py/")
})

test_that("link.text overrides the hyperlink label", {
  rendered <- knit_livelink(
    "```{livelink, link.text='Run this yourself'}",
    "plot(1:10)"
  )

  expect_match(
    paste(rendered, collapse = "\n"),
    "[Run this yourself](http",
    fixed = TRUE
  )
})

test_that("an invalid engine.target is rejected", {
  expect_error(
    knit_livelink("```{livelink, engine.target='julia'}", "x <- 1"),
    "livelink chunk target"
  )
})

test_that("filename chunk option names the file in the link", {
  rendered <- knit_livelink(
    "```{livelink, filename='analysis.R'}",
    "plot(1:10)"
  )

  preview <- preview_webr_link(extract_url(rendered))
  expect_equal(preview$files_data[[1]]$name, "analysis.R")
})

#' Turn a livelink container into a data frame
#'
#' The container classes coerce to a tidy data frame, so a folder of links or a
#' batch of decoded results can be tabulated, filtered, joined, or written to
#' CSV with the tools you already use. For a tibble, wrap the result:
#' `tibble::as_tibble(as.data.frame(x))`.
#'
#' @param x A `webr_directory`, `shinylive_directory`, `webr_decoded_batch`, or
#'   `shinylive_decoded_batch` object.
#' @param row.names A character vector of row names, or `NULL`.
#' @param optional Ignored; present for compatibility with the generic.
#' @param ... Ignored.
#'
#' @return
#' A data frame. For a directory, one row per generated link with columns
#' `filename` and `url`. For a decoded batch, one row per successfully decoded
#' URL with columns `name`, `url`, `total_files`, `total_size`, and
#' `output_dir` (URLs that failed to decode carry no result and are omitted; the
#' object's `total_urls` and `successful_urls` fields hold the counts).
#'
#' @examples
#' dir <- tempfile()
#' dir.create(dir)
#' writeLines("plot(1:10)",       file.path(dir, "one.R"))
#' writeLines("hist(rnorm(100))", file.path(dir, "two.R"))
#'
#' links <- webr_repl_directory(dir)
#' as.data.frame(links)
#'
#' @name as.data.frame.livelink
NULL

# A directory holds a named character vector: filename -> URL.
directory_to_data_frame <- function(x, row.names) {
  urls <- x$urls
  filename <- if (is.null(names(urls))) {
    rep(NA_character_, length(urls))
  } else {
    names(urls)
  }
  data.frame(
    filename = filename,
    url = unname(urls),
    stringsAsFactors = FALSE,
    row.names = row.names
  )
}

# A batch holds a named list of the decoded results (only the successes; a
# failed URL leaves no entry). Each result is itself a decoded object carrying
# its own url, counts, and output directory.
batch_to_data_frame <- function(x, row.names) {
  results <- x$results

  get_num <- function(field) {
    vapply(results, function(r) {
      v <- r[[field]]
      if (is.null(v)) NA_real_ else as.numeric(v)
    }, numeric(1))
  }
  get_chr <- function(field) {
    vapply(results, function(r) {
      v <- r[[field]]
      if (is.null(v)) NA_character_ else as.character(v)
    }, character(1))
  }

  name <- if (is.null(names(results))) seq_along(results) else names(results)

  data.frame(
    name = name,
    url = get_chr("url"),
    total_files = get_num("total_files"),
    total_size = get_num("total_size"),
    output_dir = get_chr("output_dir"),
    stringsAsFactors = FALSE,
    row.names = row.names
  )
}

#' @rdname as.data.frame.livelink
#' @export
as.data.frame.webr_directory <- function(x, row.names = NULL, optional = FALSE, ...) {
  directory_to_data_frame(x, row.names)
}

#' @rdname as.data.frame.livelink
#' @export
as.data.frame.shinylive_directory <- function(x, row.names = NULL, optional = FALSE, ...) {
  directory_to_data_frame(x, row.names)
}

#' @rdname as.data.frame.livelink
#' @export
as.data.frame.webr_decoded_batch <- function(x, row.names = NULL, optional = FALSE, ...) {
  batch_to_data_frame(x, row.names)
}

#' @rdname as.data.frame.livelink
#' @export
as.data.frame.shinylive_decoded_batch <- function(x, row.names = NULL, optional = FALSE, ...) {
  batch_to_data_frame(x, row.names)
}

#' Decode webR REPL link(s) to extract files to local directory
#'
#' Decodes webR REPL sharelinks to extract the embedded files and save them
#' to a local directory. This is the reverse operation of creating webR links.
#' Handles both single URLs and multiple URLs automatically.
#'
#' @param url Character string or vector containing webR REPL URL(s)
#' @param output_dir Character string specifying the output directory path.
#'   Defaults to a `webr_files` directory inside the session temporary
#'   directory; pass an explicit path to extract somewhere permanent.
#' @param overwrite Logical. Whether to overwrite existing files (default: FALSE)
#' @param create_subdir Logical. If `TRUE` (default), each decoded link is
#'   extracted into its own subdirectory under `output_dir` rather than directly
#'   into it. For a single URL the subdirectory is named `webr_<hash>`, where
#'   `<hash>` is a short fingerprint of the URL. For multiple URLs, see
#'   `name_dirs`. Set `FALSE` to extract straight into `output_dir`.
#' @param name_dirs Logical. For multiple URLs, controls how the per-link
#'   subdirectories are named: `TRUE` (default) numbers them `script_01`,
#'   `script_02`, ...; `FALSE` names each one `webr_<hash>` from the URL
#'   fingerprint. Ignored for a single URL, and ignored when
#'   `create_subdir = FALSE` (all files then extract into `output_dir`).
#'
#' @return
#' For a single URL, a `webr_decoded` object. For multiple URLs, a
#' `webr_decoded_batch` object.
#'
#' @seealso [preview_webr_link()] to inspect a link without writing files, and
#'   [webr_repl_link()], [webr_repl_project()], and [webr_repl_exercise()], which
#'   create the links this function decodes.
#'
#' @include utils.R
#' @export
#' @examples
#' # Round-trip: build a link, then decode it back to files
#' url <- as.character(webr_repl_link("plot(1:10)"))
#'
#' result <- decode_webr_link(url)
#' print(result)
#'
#' # Extract to a directory of your choosing
#' out <- file.path(tempdir(), "my_code")
#' decode_webr_link(url, output_dir = out, create_subdir = FALSE, overwrite = TRUE)
#' list.files(out)
#'
#' # Several links at once
#' urls <- c(url, as.character(webr_repl_link("hist(rnorm(100))")))
#' decode_webr_link(urls, output_dir = file.path(tempdir(), "my_scripts"))
decode_webr_link <- function(url,
                             output_dir = file.path(tempdir(), "webr_files"),
                             overwrite = FALSE,
                             create_subdir = TRUE,
                             name_dirs = TRUE) {

  # Validate basic inputs
  check_character_vector(url, "url")
  check_single_string(output_dir, "output_dir")
  check_single_logical(overwrite, "overwrite")
  check_single_logical(create_subdir, "create_subdir")
  check_single_logical(name_dirs, "name_dirs")

  # Handle single URL case
  if (length(url) == 1) {
    return(decode_single_webr_link(url, output_dir, overwrite, create_subdir))
  }

  # Handle multiple URLs case
  return(decode_multiple_webr_links(url, output_dir, overwrite, create_subdir, name_dirs))
}

#' Decode a single webR REPL link
#' @param url Single URL string
#' @param output_dir Output directory
#' @param overwrite Whether to overwrite files
#' @param create_subdir Whether to create subdirectory
#' @return webr_decoded object
#' @noRd
decode_single_webr_link <- function(url, output_dir, overwrite, create_subdir) {
  # Validate URL
  check_valid_webr_url(url, "url")

  # Decompress and parse URL data
  cli::cli_inform("Decompressing webR data...")
  url_data <- decompress_webr_url(url)

  mode <- url_data$mode
  version <- url_data$version
  flags <- url_data$flags
  files_data <- url_data$files_data

  cli::cli_inform("Parsing file data...")

  # Determine output directory
  final_output_dir <- if (create_subdir) {
    # Create subdirectory based on URL hash
    url_hash <- simple_hash(url, 8)
    file.path(output_dir, paste0("webr_", url_hash))
  } else {
    output_dir
  }

  # Create output directory
  if (!dir.exists(final_output_dir)) {
    dir.create(final_output_dir, recursive = TRUE, showWarnings = FALSE)
    cli::cli_inform("Created directory: {.path {final_output_dir}}")
  }

  # Decode and save files
  files_info <- decode_and_save_webr_files(files_data, final_output_dir, overwrite)

  # Summary
  cli::cli_inform(c(
    "v" = "Successfully decoded {nrow(files_info)} file{?s} to {.path {final_output_dir}}"
  ))

  # Return webr_decoded object
  new_webr_decoded(files_info, final_output_dir, url, mode, version, flags)
}

#' Decode multiple webR REPL links
#' @param urls Vector of URL strings
#' @param output_dir Base output directory
#' @param overwrite Whether to overwrite files
#' @param create_subdir Whether to create subdirectories
#' @param name_dirs Whether to use numbered directory names
#' @return webr_decoded_batch object
#' @noRd
decode_multiple_webr_links <- function(urls, output_dir, overwrite, create_subdir, name_dirs) {
  if (length(urls) == 0) {
    cli::cli_warn("No URLs provided")
    return(new_webr_decoded_batch(list(), output_dir, character(0)))
  }

  # Validate all URLs first
  for (i in seq_along(urls)) {
    tryCatch({
      check_valid_webr_url(urls[i], paste0("url[", i, "]"))
    }, error = function(e) {
      cli::cli_abort(c(
        "Invalid URL at position {i}",
        "x" = "{e$message}"
      ))
    })
  }

  cli::cli_inform("Processing {length(urls)} webR URL{?s}...")

  results <- list()

  for (i in seq_along(urls)) {
    url <- urls[i]

    tryCatch({
      # Determine subdirectory name
      if (!create_subdir) {
        # Use base output directory for each URL
        url_output_dir <- output_dir
        subdir_name <- paste0("url_", i)
      } else if (name_dirs) {
        subdir_name <- sprintf("script_%02d", i)
        url_output_dir <- file.path(output_dir, subdir_name)
      } else {
        url_hash <- simple_hash(url, 8)
        subdir_name <- paste0("webr_", url_hash)
        url_output_dir <- file.path(output_dir, subdir_name)
      }

      cli::cli_inform("")
      cli::cli_h3("Processing URL {i}/{length(urls)}: {subdir_name}")

      # Decode files using single URL logic
      decoded_result <- decode_single_webr_link(
        url = url,
        output_dir = url_output_dir,
        overwrite = overwrite,
        create_subdir = FALSE  # We handle subdirectory creation here
      )

      results[[subdir_name]] <- decoded_result

    }, error = function(e) {
      cli::cli_warn(c(
        "Failed to process URL {i}",
        "x" = "{e$message}"
      ))
    })
  }

  # A failed URL is warned about above and simply leaves no entry in `results`.
  successful_results <- results
  cli::cli_inform(c(
    "",
    "v" = "Successfully processed {length(successful_results)}/{length(urls)} URL{?s}"
  ))

  # Return webr_decoded_batch object
  new_webr_decoded_batch(results, output_dir, urls)
}

#' Decode and save files from webR data
#' @param files_data Parsed file data from webR
#' @param output_dir Directory to save files
#' @param overwrite Whether to overwrite existing files
#' @return Data frame with file information
#' @noRd
decode_and_save_webr_files <- function(files_data, output_dir, overwrite) {
  cli::cli_inform("Decoding {length(files_data)} file{?s}...")

  files_info <- data.frame(
    filename = character(0),
    path = character(0),
    autorun = logical(0),
    size_bytes = numeric(0),
    stringsAsFactors = FALSE
  )

  # Track reasons for skipped files
  skip_reasons <- list(
    invalid_structure = character(0),
    no_content = character(0),
    already_exists = character(0),
    save_failed = character(0)
  )

  for (i in seq_along(files_data)) {
    file_info <- files_data[[i]]

    # Validate file structure
    if (!all(c("name", "path") %in% names(file_info))) {
      cli::cli_warn("Skipping invalid file entry {i}: missing required fields")
      skip_reasons$invalid_structure <- c(skip_reasons$invalid_structure, paste0("entry_", i))
      next
    }

    # Check for content in either "text" or "data" field
    has_content <- ("text" %in% names(file_info)) || ("data" %in% names(file_info))
    if (!has_content) {
      # "name" is guaranteed present: the structure check above already skipped
      # any entry missing it.
      filename <- file_info$name
      cli::cli_warn("Skipping file entry {i}: no content found")
      skip_reasons$no_content <- c(skip_reasons$no_content, filename)
      next
    }

    filename <- file_info$name
    file_path <- file.path(output_dir, filename)
    autorun <- isTRUE(file_info$autorun)

    # Create subdirectories if needed
    file_dir <- dirname(file_path)
    if (!dir.exists(file_dir)) {
      dir.create(file_dir, recursive = TRUE, showWarnings = FALSE)
    }

    # Check if file exists and handle overwrite
    if (file.exists(file_path) && !overwrite) {
      cli::cli_warn("File already exists, skipping: {.file {filename}}")
      skip_reasons$already_exists <- c(skip_reasons$already_exists, filename)
      next
    }

    # Save file based on content type
    tryCatch({
      # Determine content and type
      if ("text" %in% names(file_info)) {
        # Text content in "text" field
        content <- file_info$text
        is_binary <- FALSE
      } else if ("data" %in% names(file_info)) {
        # Content in "data" field - could be text or binary
        content <- file_info$data

        # For webR, "data" field typically contains text content (converted from raw)
        # We can detect if it's likely binary by checking for non-printable characters
        is_binary <- detect_binary_content(content)
      }

      if (is_binary) {
        # Handle as binary data
        if (is.character(content) && grepl("^[0-9a-f ]+$", content)) {
          # Hex string format - convert back to binary
          hex_chars <- strsplit(gsub(" ", "", content), "")[[1]]
          if (length(hex_chars) %% 2 == 0) {
            hex_pairs <- paste0(hex_chars[seq(1, length(hex_chars), 2)],
                                hex_chars[seq(2, length(hex_chars), 2)])
            binary_data <- as.raw(strtoi(hex_pairs, 16L))
            writeBin(binary_data, file_path)
            file_size <- length(binary_data)
          } else {
            # Odd number of hex chars, treat as text
            writeLines(content, file_path, useBytes = TRUE)
            file_size <- nchar(content, type = "bytes")
          }
        } else {
          # Treat as text
          writeLines(content, file_path, useBytes = TRUE)
          file_size <- nchar(content, type = "bytes")
        }
      } else {
        # Handle as text data
        writeLines(content, file_path, useBytes = TRUE)
        file_size <- nchar(content, type = "bytes")
      }

      # Add to results
      files_info <- rbind(files_info, data.frame(
        filename = filename,
        path = file_path,
        autorun = autorun,
        size_bytes = file_size,
        stringsAsFactors = FALSE
      ))

      size_label <- format_file_size(file_size)
      autorun_label <- if (autorun) " (autorun)" else ""
      cli::cli_inform("  {.file {filename}} ({size_label}){autorun_label}")

    }, error = function(e) {
      cli::cli_warn(c(
        "Failed to save file: {.file {filename}}",
        "x" = "{conditionMessage(e)}"
      ))
      # `<<-`, not `<-`: a plain assignment here lands in the handler's frame and
      # is discarded, which left the save_failed reporting block unreachable.
      skip_reasons$save_failed <<- c(skip_reasons$save_failed, filename)
    })
  }

  # Add skip reasons as attribute for better error reporting
  attr(files_info, "skip_reasons") <- skip_reasons

  files_info
}

#' Detect if content is likely binary
#' @param content Character string to check
#' @return Logical indicating if content is likely binary
#' @noRd
detect_binary_content <- function(content) {
  if (!is.character(content) || length(content) == 0 || nchar(content) == 0) {
    return(FALSE)
  }

  raw_chars <- charToRaw(content)

  if (any(raw_chars == 0)) {
    return(TRUE)
  }

  # Well-formed UTF-8 is text, even though its bytes run above 126. Judging
  # "printable" byte by byte would call "café" and "你好" binary.
  if (!validUTF8(content)) {
    return(TRUE)
  }

  # Control characters (other than tab, newline, carriage return) are the real
  # signal that this is not text.
  control <- raw_chars < as.raw(32) & !(raw_chars %in% as.raw(c(9, 10, 13)))

  sum(control) / length(raw_chars) > 0.1
}

#' Normalize msgpack data from RcppMsgPack format to list format
#' @param msgpack_data Raw msgpack data from RcppMsgPack
#' @return Normalized list structure
#' @noRd
normalize_msgpack_data <- function(msgpack_data) {
  if (!is.list(msgpack_data)) {
    return(msgpack_data)
  }

  # Handle direct key-value structure at the root level
  if (all(c("key", "value") %in% names(msgpack_data))) {
    return(list(convert_keyvalue_to_list(msgpack_data)))
  }

  # Process each item in the list
  normalized <- lapply(msgpack_data, function(item) {
    # Check if this item has the RcppMsgPack key-value structure
    if (is.list(item) && all(c("key", "value") %in% names(item))) {
      return(convert_keyvalue_to_list(item))
    } else {
      # If not the expected structure, return as-is
      return(item)
    }
  })

  return(normalized)
}

#' Convert key-value structure to named list
#' @param kv_item Item with key and value lists
#' @return Named list
#' @noRd
convert_keyvalue_to_list <- function(kv_item) {
  keys <- kv_item$key
  values <- kv_item$value

  if (length(keys) != length(values)) {
    cli::cli_warn("Mismatched key-value lengths in msgpack data")
    return(kv_item)
  }

  result <- list()

  for (i in seq_along(keys)) {
    key <- as.character(keys[[i]])
    value <- values[[i]]

    # Handle different value types
    if (is.raw(value)) {
      # Take the fallback as the value of tryCatch; assigning to `result` inside
      # the handler would write into the handler's frame and be discarded.
      result[[key]] <- tryCatch(
        rawToChar(value),
        error = function(e) {
          cli::cli_warn("Failed to convert raw data for key '{key}' to character: {conditionMessage(e)}")
          paste(sprintf("%02x", as.integer(value)), collapse = " ")
        }
      )
    } else if (is.logical(value) && key == "autorun") {
      # Keep autorun as logical
      result[[key]] <- as.logical(value)
    } else {
      # Keep other values as-is
      result[[key]] <- value
    }
  }

  return(result)
}

#' Decompress and parse webR URL data
#' @param url webR URL
#' @return List containing mode, version, flags, and files_data
#' @noRd
decompress_webr_url <- function(url) {
  # Extract URL metadata
  url_parts <- parse_webr_url(url)

  mode <- url_parts$mode
  version <- url_parts$version
  encoded_data <- url_parts$encoded_data
  flags <- url_parts$flags

  if (is.null(encoded_data)) {
    cli::cli_abort(c(
      "No 'code' parameter found in URL",
      "x" = "URL must contain a 'code' parameter with the encoded file data"
    ))
  }

  # Decode URL encoding
  decoded_data <- utils::URLdecode(encoded_data)

  # Base64 decode
  tryCatch({
    binary_data <- base64enc::base64decode(decoded_data)
  }, error = function(e) {
    cli::cli_abort(c(
      "Error decoding base64 data",
      "x" = "Failed to decode base64: {e$message}",
      "i" = "The URL may be corrupted"
    ))
  })

  # Decompress if needed
  if (grepl("u", flags)) {
    # Uncompressed
    raw_data <- binary_data
  } else if (grepl("z", flags)) {
    # zlib/gzip compressed - try both formats.
    # The fallback must be the *value* of the tryCatch: assigning inside the
    # error handler writes into the handler's own frame, so `raw_data` would
    # never reach this one and the graceful path would die with
    # `object 'raw_data' not found`.
    raw_data <- tryCatch(
      memDecompress(binary_data, type = "gzip"),
      error = function(e1) {
        tryCatch(
          # Try without gzip headers (raw zlib)
          memDecompress(binary_data, type = "unknown"),
          error = function(e2) {
            cli::cli_abort(c(
              "Error during decompression",
              "x" = "Failed to decompress data: {conditionMessage(e1)}",
              "i" = "The URL may be corrupted or use an unsupported compression format"
            ))
          }
        )
      }
    )
  } else {
    cli::cli_abort(c(
      "Unknown compression flags",
      "x" = "Flags '{flags}' contain unknown compression options",
      "i" = "Supported flags: 'u' (uncompressed), 'z' (compressed)"
    ))
  }

  # Parse format
  if (grepl("m", flags)) {
    # msgpack format
    tryCatch({
      if (!requireNamespace("RcppMsgPack", quietly = TRUE)) {
        cli::cli_abort(c(
          "Package RcppMsgPack is required for msgpack format",
          "i" = "Install with: install.packages('RcppMsgPack')"
        ))
      }
      files_data_raw <- RcppMsgPack::msgpack_unpack(raw_data)

      # Normalize the msgpack data structure
      files_data <- normalize_msgpack_data(files_data_raw)

    }, error = function(e) {
      cli::cli_abort(c(
        "Error parsing msgpack data",
        "x" = "Failed to parse msgpack format: {e$message}",
        "i" = "The decoded data may not be valid msgpack"
      ))
    })
  } else if (grepl("j", flags)) {
    # JSON format
    tryCatch({
      json_string <- rawToChar(raw_data)
      files_data <- jsonlite::fromJSON(json_string, simplifyVector = FALSE)
    }, error = function(e) {
      cli::cli_abort(c(
        "Error parsing JSON data",
        "x" = "Failed to parse JSON format: {e$message}",
        "i" = "The decoded data may not be valid JSON"
      ))
    })
  } else {
    cli::cli_abort(c(
      "Unknown format flags",
      "x" = "Flags '{flags}' contain unknown format options",
      "i" = "Supported flags: 'm' (msgpack), 'j' (JSON)"
    ))
  }

  list(
    mode = mode,
    version = version,
    flags = flags,
    files_data = files_data
  )
}

#' Parse webR URL structure
#' @param url webR URL to parse
#' @return List with URL components
#' @noRd
parse_webr_url <- function(url) {
  # Extract version from URL path using a more robust approach
  version <- extract_version_from_url(url)

  # Split URL at fragment (#)
  url_parts <- strsplit(url, "#")[[1]]
  if (length(url_parts) < 2) {
    cli::cli_abort(c(
      "No encoded data found in URL",
      "x" = "URL must contain encoded data after the '#' symbol"
    ))
  }

  base_url <- url_parts[1]
  fragment <- url_parts[2]

  # Extract mode from base URL if present
  mode <- extract_mode_from_url(base_url)

  # Parse fragment parameters
  fragment_parts <- extract_webr_parameters(fragment)

  list(
    mode = mode,
    version = version,
    encoded_data = fragment_parts$code,
    flags = fragment_parts$flags
  )
}

#' Extract version from webR URL
#' @param url webR URL
#' @return Version string
#' @noRd
extract_version_from_url <- function(url) {
  # Use regmatches and regexec for more reliable capture group extraction
  pattern <- "webr\\.r-wasm\\.org/([^/]+)/"
  matches <- regexec(pattern, url)

  if (matches[[1]][1] == -1) {
    return("unknown")
  }

  # Extract the captured group (version part)
  match_result <- regmatches(url, matches)[[1]]

  if (length(match_result) >= 2) {
    version <- match_result[2]  # Second element is the captured group
  } else {
    version <- "unknown"
  }

  return(version)
}

#' Extract mode from webR URL
#' @param base_url Base part of URL before fragment
#' @return Mode string or NULL
#' @noRd
extract_mode_from_url <- function(base_url) {
  # Look for mode parameter in query string - simplified approach
  # Handle both quoted and unquoted mode values
  mode_patterns <- c(
    "mode=(['\"])([^'\"&]+)\\1",  # Quoted mode values
    "mode=([^&'\"]+)"             # Unquoted mode values
  )

  for (pattern in mode_patterns) {
    matches <- regexec(pattern, base_url)

    if (matches[[1]][1] != -1) {
      match_result <- regmatches(base_url, matches)[[1]]

      if (length(match_result) >= 3) {
        # Quoted pattern - mode is in third element
        mode <- match_result[3]
      } else if (length(match_result) >= 2) {
        # Unquoted pattern - mode is in second element
        mode <- match_result[2]
      } else {
        next
      }

      # Split back into the panel vector the caller passed to `panels`, so a
      # decoded link reports what was encoded rather than the wire format.
      return(strsplit(mode, "-", fixed = TRUE)[[1]])
    }
  }

  return(NULL)
}

#' Format panel vector for display using cli
#' @param mode Panel vector (like c("editor", "plot")) or string ("editor-plot")
#' @return Formatted string for display
#' @noRd
format_mode_for_display <- function(mode) {
  if (is.null(mode) || !is.character(mode) || length(mode) == 0) {
    return(NULL)
  }

  # Accept either the panel vector or the hyphen-joined wire format.
  components <- unlist(strsplit(mode, "-", fixed = TRUE))
  components <- tools::toTitleCase(components)

  # Use cli to format with proper conjunctions (Oxford comma included by default)
  cli::cli_vec(components)
}

#' Extract parameters from webR URL fragment
#' @param fragment URL fragment part after '#'
#' @return List with code and flags
#' @noRd
extract_webr_parameters <- function(fragment) {
  # Split by & to handle multiple parameters
  params <- strsplit(fragment, "&")[[1]]

  code <- NULL
  flags <- NULL

  for (param in params) {
    if (grepl("^code=", param)) {
      code <- sub("^code=", "", param)
    } else if (nchar(param) > 0 && !grepl("=", param)) {
      # Flags are typically at the end without = sign
      flags <- param
    }
  }

  # Default flags if not specified
  if (is.null(flags)) {
    flags <- "mz"  # msgpack + compressed
  }

  list(code = code, flags = flags)
}


#' Preview webR REPL link contents without writing files to disk
#'
#' @description
#' Decodes a webR URL and returns information about the embedded files
#' without actually saving them to disk. Use print method options to control display.
#'
#' @param url Character string containing the webR URL
#'
#' @return A `webr_preview` object (a list) with elements including `files_data`
#'   (the embedded files), `total_files`, `total_size`, `mode`, `version`,
#'   `flags`, and `autorun_files`. Its `print()` method accepts `show_content`
#'   and `max_content_length` to display file bodies.
#'
#' @seealso [decode_webr_link()] to extract the embedded files to disk once you
#'   are happy with the preview.
#'
#' @examples
#' url <- as.character(webr_repl_link("plot(1:10)"))
#'
#' # Inspect a link without writing anything to disk
#' preview <- preview_webr_link(url)
#'
#' # Default print (no content)
#' print(preview)
#'
#' # Show file contents
#' print(preview, show_content = TRUE)
#'
#' # Show file contents with custom length limit
#' print(preview, show_content = TRUE, max_content_length = 200)
#'
#' # Access the preview data
#' preview$files_data
#' preview$total_files
#'
#' @export
preview_webr_link <- function(url) {

  # Validate inputs
  check_valid_webr_url(url, "url")

  # Decompress and parse URL data
  tryCatch({
    url_data <- decompress_webr_url(url)

    mode <- url_data$mode
    version <- url_data$version
    flags <- url_data$flags
    files_data <- url_data$files_data

    # Calculate metadata
    total_size <- 0
    autorun_files <- character(0)

    for (i in seq_along(files_data)) {
      file_info <- files_data[[i]]

      # Calculate size
      if ("text" %in% names(file_info)) {
        size_bytes <- nchar(file_info$text, type = "bytes")
      } else if ("data" %in% names(file_info)) {
        if (is.character(file_info$data)) {
          size_bytes <- nchar(file_info$data, type = "bytes")
        } else {
          size_bytes <- length(file_info$data)
        }
      } else {
        size_bytes <- 0
      }

      total_size <- total_size + size_bytes

      # Track autorun files
      if (isTRUE(file_info$autorun)) {
        autorun_files <- c(autorun_files, file_info$name)
      }
    }

    # Return webr_preview object
    new_webr_preview(url, mode, version, flags, files_data, total_size, autorun_files)

  }, error = function(e) {
    cli::cli_abort(c(
      "Failed to preview webR link",
      "x" = "{e$message}"
    ))
  })
}

# livelink always WRITES json+gzip ("jz"), so every round-trip test only proves we
# can read links we made ourselves. webR's own native wire format is msgpack+zlib
# ("mz"). These tests build links in the formats webR emits and prove we can read
# them -- i.e. that livelink can decode a link produced by webR itself.

webr_url <- function(payload, flags) {
  encoded <- utils::URLencode(base64enc::base64encode(payload), reserved = TRUE)
  paste0("https://webr.r-wasm.org/latest/#code=", encoded, "&", flags)
}

msgpack_webr_url <- function(files, flags = "mz") {
  packed <- RcppMsgPack::msgpack_pack(files)
  payload <- if (grepl("z", flags)) memCompress(packed, type = "gzip") else packed
  webr_url(payload, flags)
}

json_webr_url <- function(files, flags = "jz") {
  json <- jsonlite::toJSON(files, auto_unbox = TRUE)
  payload <- if (grepl("z", flags)) {
    memCompress(charToRaw(json), type = "gzip")
  } else {
    charToRaw(json)
  }
  webr_url(payload, flags)
}

one_file <- list(list(
  name = "script.R",
  path = "/home/web_user/script.R",
  text = "# a comment\nplot(1:10)"
))

test_that("a msgpack link (webR's native 'mz' format) can be previewed", {
  preview <- preview_webr_link(msgpack_webr_url(one_file))

  expect_s3_class(preview, "webr_preview")
  expect_equal(preview$total_files, 1)
  expect_equal(preview$files_data[[1]]$name, "script.R")
  expect_equal(preview$files_data[[1]]$text, "# a comment\nplot(1:10)")
})

test_that("a msgpack link can be decoded to disk", {
  dir <- withr::local_tempdir()

  suppressMessages(
    decode_webr_link(msgpack_webr_url(one_file), output_dir = dir, overwrite = TRUE)
  )

  path <- list.files(dir, recursive = TRUE, full.names = TRUE)
  expect_length(path, 1)
  expect_equal(
    paste(readLines(path, warn = FALSE), collapse = "\n"),
    "# a comment\nplot(1:10)"
  )
})

test_that("a multi-file msgpack link round-trips", {
  files <- list(
    list(name = "main.R", path = "/home/web_user/main.R", text = "source('utils.R')"),
    list(name = "utils.R", path = "/home/web_user/utils.R", text = "f <- function() 42")
  )
  dir <- withr::local_tempdir()

  suppressMessages(
    decode_webr_link(msgpack_webr_url(files), output_dir = dir, overwrite = TRUE)
  )

  written <- list.files(dir, recursive = TRUE)
  expect_setequal(basename(written), c("main.R", "utils.R"))
})

test_that("uncompressed links are decoded", {
  # 'ju' = json, uncompressed
  preview <- preview_webr_link(json_webr_url(one_file, flags = "ju"))

  expect_equal(preview$files_data[[1]]$text, "# a comment\nplot(1:10)")
})

test_that("the autorun flag is recovered from the URL", {
  autorun_file <- list(list(
    name = "script.R",
    path = "/home/web_user/script.R",
    text = "plot(1:10)",
    autorun = TRUE
  ))

  preview <- preview_webr_link(json_webr_url(autorun_file, flags = "jza"))

  expect_equal(preview$autorun_files, "script.R")
})

test_that("a corrupt payload produces a readable error, not an internal one", {
  bad <- paste0(
    "https://webr.r-wasm.org/latest/#code=",
    utils::URLencode(base64enc::base64encode(charToRaw("not gzipped at all")), reserved = TRUE),
    "&jz"
  )

  err <- tryCatch(
    suppressWarnings(preview_webr_link(bad)),
    error = function(e) conditionMessage(e)
  )

  # Must not surface `object 'raw_data' not found` -- the decompression fallback
  # used to be assigned inside a tryCatch error handler, so it was discarded and
  # the graceful path died with an internal error instead.
  expect_false(grepl("raw_data", err))
  expect_false(grepl("not found", err))
  expect_match(err, "webR link|decompress|corrupt|JSON")
})

test_that("a URL with no code fragment is rejected", {
  expect_error(preview_webr_link("https://webr.r-wasm.org/latest/"), class = "rlang_error")
})

test_that("a non-webR URL is rejected", {
  expect_error(preview_webr_link("https://example.com/nope"))
})

# Coverage for the decode paths -- the least-tested half of the package, and where
# the simple_hash() overflow bug lived.

webr_urls <- function(n = 2) {
  vapply(
    seq_len(n),
    function(i) as.character(webr_repl_link(sprintf("x <- %d", i), filename = sprintf("s%d.R", i))),
    character(1)
  )
}

shiny_urls <- function(n = 2) {
  vapply(
    seq_len(n),
    function(i) as.character(shinylive_r_link(sprintf("library(shiny) # %d", i))),
    character(1)
  )
}

# ---- batch decoding -------------------------------------------------------

test_that("decoding several webR URLs returns a batch object", {
  dir <- withr::local_tempdir()

  batch <- suppressMessages(decode_webr_link(webr_urls(3), output_dir = dir))

  expect_s3_class(batch, "webr_decoded_batch")
  expect_equal(batch$total_urls, 3)
  expect_equal(batch$successful_urls, 3)
  expect_equal(batch$total_files, 3)
})

test_that("each URL in a batch lands in its own directory", {
  dir <- withr::local_tempdir()

  suppressMessages(decode_webr_link(webr_urls(2), output_dir = dir))

  # name_dirs = TRUE (default) -> script_01, script_02
  expect_setequal(list.dirs(dir, recursive = FALSE, full.names = FALSE),
                  c("script_01", "script_02"))
  expect_true(file.exists(file.path(dir, "script_01", "s1.R")))
  expect_true(file.exists(file.path(dir, "script_02", "s2.R")))
})

test_that("name_dirs = FALSE uses hash-named directories", {
  dir <- withr::local_tempdir()

  suppressMessages(decode_webr_link(webr_urls(2), output_dir = dir, name_dirs = FALSE))

  dirs <- list.dirs(dir, recursive = FALSE, full.names = FALSE)
  expect_length(dirs, 2)
  expect_true(all(grepl("^webr_[0-9a-f]{8}$", dirs)))
})

test_that("create_subdir = FALSE decodes a batch into one flat directory", {
  dir <- withr::local_tempdir()

  suppressMessages(
    decode_webr_link(webr_urls(2), output_dir = dir, create_subdir = FALSE)
  )

  expect_length(list.dirs(dir, recursive = FALSE), 0)
  expect_setequal(list.files(dir), c("s1.R", "s2.R"))
})

test_that("decoding several Shinylive URLs returns a batch object", {
  dir <- withr::local_tempdir()

  batch <- suppressMessages(decode_shinylive_link(shiny_urls(2), output_dir = dir))

  expect_s3_class(batch, "shinylive_decoded_batch")
  expect_equal(batch$total_urls, 2)
  expect_equal(batch$successful_urls, 2)
})

test_that("an invalid URL in a batch is reported, not silently dropped", {
  dir <- withr::local_tempdir()
  urls <- c(webr_urls(1), "https://example.com/not-a-webr-link")

  expect_error(
    suppressMessages(decode_webr_link(urls, output_dir = dir)),
    "Invalid URL"
  )
})

test_that("an empty URL vector warns and returns an empty batch", {
  dir <- withr::local_tempdir()

  expect_warning(
    batch <- suppressMessages(decode_webr_link(character(0), output_dir = dir)),
    "No URLs"
  )
  expect_s3_class(batch, "webr_decoded_batch")
  expect_equal(batch$total_urls, 0)
})

# ---- single-URL layout ----------------------------------------------------

test_that("create_subdir = FALSE writes straight into output_dir", {
  dir <- withr::local_tempdir()
  url <- as.character(webr_repl_link("plot(1:10)", filename = "flat.R"))

  suppressMessages(decode_webr_link(url, output_dir = dir, create_subdir = FALSE))

  expect_true(file.exists(file.path(dir, "flat.R")))
  expect_length(list.dirs(dir, recursive = FALSE), 0)
})

test_that("create_subdir = TRUE nests under a hash-named directory", {
  dir <- withr::local_tempdir()

  suppressMessages(decode_webr_link(as.character(webr_repl_link("x <- 1")), output_dir = dir))

  subdirs <- list.dirs(dir, recursive = FALSE, full.names = FALSE)
  expect_length(subdirs, 1)
  expect_match(subdirs, "^webr_[0-9a-f]{8}$")
})

# ---- overwrite -----------------------------------------------------------

test_that("existing files are skipped unless overwrite = TRUE", {
  dir <- withr::local_tempdir()
  url <- as.character(webr_repl_link("new_content <- 1", filename = "f.R"))

  writeLines("old_content <- 0", file.path(dir, "f.R"))

  # overwrite = FALSE: the file on disk must survive untouched
  suppressMessages(
    decode_webr_link(url, output_dir = dir, create_subdir = FALSE, overwrite = FALSE)
  )
  expect_equal(readLines(file.path(dir, "f.R"), warn = FALSE), "old_content <- 0")

  # overwrite = TRUE: it must be replaced
  suppressMessages(
    decode_webr_link(url, output_dir = dir, create_subdir = FALSE, overwrite = TRUE)
  )
  expect_equal(readLines(file.path(dir, "f.R"), warn = FALSE), "new_content <- 1")
})

test_that("a skipped file is reported without leaking cli templates", {
  dir <- withr::local_tempdir()
  url <- as.character(webr_repl_link("x <- 1", filename = "f.R"))
  writeLines("existing", file.path(dir, "f.R"))

  result <- suppressMessages(
    decode_webr_link(url, output_dir = dir, create_subdir = FALSE, overwrite = FALSE)
  )

  expect_equal(result$total_files, 0)

  out <- paste(capture.output(print(result), type = "message"), collapse = "\n")

  expect_match(out, "already exist")
  expect_match(out, "f.R", fixed = TRUE)
  expect_match(out, "overwrite = TRUE", fixed = TRUE)
  # An unevaluated {length(...)} in the output means cli globbed the template.
  expect_false(grepl("{length(", out, fixed = TRUE))
})

test_that("many skipped files are truncated in the report", {
  dir <- withr::local_tempdir()
  files <- setNames(as.list(paste0("x <- ", 1:5)), paste0("f", 1:5, ".R"))
  url <- as.character(webr_repl_project(files))

  for (nm in names(files)) writeLines("existing", file.path(dir, nm))

  result <- suppressMessages(
    decode_webr_link(url, output_dir = dir, create_subdir = FALSE, overwrite = FALSE)
  )
  out <- paste(capture.output(print(result), type = "message"), collapse = "\n")

  expect_match(out, "and 2 more", fixed = TRUE)
  expect_false(grepl("{length(", out, fixed = TRUE))
})

# ---- binary content -------------------------------------------------------

test_that("detect_binary_content spots binary payloads", {
  expect_false(detect_binary_content("plain text\nwith newlines"))
  expect_false(detect_binary_content(""))

  # Control characters are the real signal.
  expect_true(detect_binary_content(rawToChar(as.raw(c(1, 2, 3, 4, 5, 6, 7, 8)))))
})

# Regression: "printable" was judged byte by byte, so every byte above 126 counted
# as non-printable -- which made ordinary UTF-8 text look like binary data.
test_that("UTF-8 text is not mistaken for binary", {
  expect_false(detect_binary_content("caf\u00e9"))
  expect_false(detect_binary_content("h\u00e9llo w\u00f6rld"))
  expect_false(detect_binary_content("\u4f60\u597d\u4e16\u754c"))
})

test_that("a Shinylive binary file is decoded from base64", {
  dir <- withr::local_tempdir()
  payload <- as.raw(c(0x89, 0x50, 0x4e, 0x47, 0x0d, 0x0a, 0x1a, 0x0a))  # PNG magic

  files <- list(
    list(name = "app.R", content = "library(shiny)", type = "text"),
    list(name = "logo.png", content = base64enc::base64encode(payload), type = "binary")
  )
  json <- jsonlite::toJSON(files, auto_unbox = TRUE)
  url <- paste0(
    "https://shinylive.io/r/editor/#code=",
    gsub("/", "-", lzstring::compressToEncodedURIComponent(as.character(json)))
  )

  suppressMessages(decode_shinylive_link(url, output_dir = dir, create_subdir = FALSE))

  written <- readBin(file.path(dir, "logo.png"), "raw", n = 8)
  expect_identical(written, payload)
})

test_that("a Shinylive preview reports file types", {
  url <- as.character(shinylive_r_link("library(shiny)"))

  preview <- preview_shinylive_link(url)

  expect_equal(preview$file_types, "text")
  expect_gt(preview$total_size, 0)
})

# ---- Python Shinylive -----------------------------------------------------

test_that("a Python Shinylive link previews and decodes", {
  dir <- withr::local_tempdir()
  url <- as.character(shinylive_py_link("from shiny import App\napp = App()"))

  preview <- preview_shinylive_link(url)
  expect_equal(preview$engine, "python")

  suppressMessages(decode_shinylive_link(url, output_dir = dir, create_subdir = FALSE))
  expect_equal(
    paste(readLines(file.path(dir, "app.py"), warn = FALSE), collapse = "\n"),
    "from shiny import App\napp = App()"
  )
})

test_that("a multi-file Shinylive app decodes every file", {
  dir <- withr::local_tempdir()
  url <- as.character(shinylive_r_link(list(
    "app.R"   = "library(shiny)",
    "utils.R" = "f <- function() 42",
    "data.csv" = "a,b\n1,2"
  )))

  suppressMessages(decode_shinylive_link(url, output_dir = dir, create_subdir = FALSE))

  expect_setequal(list.files(dir), c("app.R", "utils.R", "data.csv"))
  expect_equal(
    paste(readLines(file.path(dir, "data.csv"), warn = FALSE), collapse = "\n"),
    "a,b\n1,2"
  )
})

# ---- URL parsing error branches -------------------------------------------

test_that("malformed webR URLs are rejected with useful errors", {
  expect_error(preview_webr_link("not a url"), "webR")
  expect_error(preview_webr_link("https://example.com/#code=abc&jz"), "webR")
  expect_error(decode_webr_link(42), class = "rlang_error")
})

test_that("malformed Shinylive URLs are rejected", {
  expect_error(preview_shinylive_link("https://example.com/#code=abc"), "Shinylive")
  expect_error(preview_shinylive_link("not a url"), "Shinylive")
})

test_that("a webR URL with no code fragment is rejected", {
  expect_error(preview_webr_link("https://webr.r-wasm.org/latest/#flags=jz"))
})

# ---- URL parsing happy paths ----------------------------------------------

test_that("version and panels are parsed back out of a URL", {
  url <- as.character(
    webr_repl_link("x <- 1", version = "v0.5.4", panels = c("editor", "plot"))
  )

  preview <- preview_webr_link(url)

  expect_equal(preview$version, "v0.5.4")
  expect_equal(preview$mode, c("editor", "plot"))
})

test_that("Shinylive mode and engine are parsed back out of a URL", {
  preview <- preview_shinylive_link(as.character(shinylive_r_link("library(shiny)", mode = "app")))

  expect_equal(preview$mode, "app")
  expect_equal(preview$engine, "r")
})

# Tests for decode functions

test_that("preview_webr_link returns webr_preview object", {
  # Create a link and then preview it
  link <- webr_repl_link("x <- 1 + 1\nprint(x)")
  preview <- preview_webr_link(link$url)

  expect_s3_class(preview, "webr_preview")
  expect_equal(preview$total_files, 1)
  expect_true(preview$total_size > 0)
  expect_equal(preview$version, "latest")
})

test_that("preview_webr_link extracts correct file content", {
  original_code <- "# Test comment\nresult <- 42"
  link <- webr_repl_link(original_code)
  preview <- preview_webr_link(link$url)

  # Check file data structure
  expect_equal(length(preview$files_data), 1)
  file_info <- preview$files_data[[1]]
  expect_equal(file_info$name, "script.R")
})

test_that("preview_webr_link handles autorun files", {
  link <- webr_repl_link("print('auto')", autorun = TRUE)
  preview <- preview_webr_link(link$url)

  expect_equal(length(preview$autorun_files), 1)
})

test_that("preview_webr_link validates URL", {
  expect_error(
    preview_webr_link("https://example.com"),
    "valid webR"
  )
  expect_error(
    preview_webr_link(""),
    "valid webR"
  )
})

test_that("decode_webr_link creates output directory and files", {
  skip_on_cran() # Skip file I/O tests on CRAN

  link <- webr_repl_link("test_code <- 123")
  temp_dir <- tempfile("webr_decode_test")

  result <- decode_webr_link(link$url, output_dir = temp_dir, create_subdir = FALSE)

  expect_s3_class(result, "webr_decoded")
  expect_true(dir.exists(temp_dir))
  expect_true(file.exists(file.path(temp_dir, "script.R")))

  # Cleanup
  unlink(temp_dir, recursive = TRUE)
})

test_that("decode_webr_link respects overwrite parameter", {
  skip_on_cran()

  link <- webr_repl_link("original <- 1")
  temp_dir <- tempfile("webr_overwrite_test")
  dir.create(temp_dir, recursive = TRUE)

  # First decode
  decode_webr_link(link$url, output_dir = temp_dir, create_subdir = FALSE)

  # Second decode without overwrite should skip (suppress warning about existing file)
  link2 <- webr_repl_link("modified <- 2")
  result <- suppressWarnings(
    decode_webr_link(link2$url, output_dir = temp_dir, create_subdir = FALSE, overwrite = FALSE)
  )

  # File should not be overwritten (files_info will show skipped)
  expect_equal(nrow(result$files_info), 0) # Skipped due to existing file

  # Cleanup
  unlink(temp_dir, recursive = TRUE)
})

test_that("preview_shinylive_link returns shinylive_preview object", {
  link <- shinylive_r_link("shinyApp(fluidPage(), function(i, o) {})")
  preview <- preview_shinylive_link(link$url)

  expect_s3_class(preview, "shinylive_preview")
  expect_equal(preview$total_files, 1)
  expect_equal(preview$engine, "r")
  expect_equal(preview$mode, "editor")
})

test_that("preview_shinylive_link validates URL", {
  expect_error(
    preview_shinylive_link("https://example.com"),
    "valid Shinylive"
  )
})

test_that("decode_shinylive_link creates output directory and files", {
  skip_on_cran()

  link <- shinylive_r_link("shinyApp(fluidPage('test'), function(i, o) {})")
  temp_dir <- tempfile("shinylive_decode_test")

  result <- decode_shinylive_link(link$url, output_dir = temp_dir, create_subdir = FALSE)

  expect_s3_class(result, "shinylive_decoded")
  expect_true(dir.exists(temp_dir))
  expect_true(file.exists(file.path(temp_dir, "app.R")))

  # Cleanup
  unlink(temp_dir, recursive = TRUE)
})

test_that("round-trip webr encoding and decoding preserves content", {
  skip_on_cran()

  original_code <- "# Round trip test\nresult <- sum(1:10)\nprint(result)"
  link <- webr_repl_link(original_code, filename = "roundtrip.R")
  temp_dir <- tempfile("webr_roundtrip")

  result <- decode_webr_link(link$url, output_dir = temp_dir, create_subdir = FALSE)
  decoded_content <- readLines(file.path(temp_dir, "roundtrip.R"), warn = FALSE)

  expect_equal(paste(decoded_content, collapse = "\n"), original_code)

  # Cleanup
  unlink(temp_dir, recursive = TRUE)
})

test_that("decode_shinylive_link creates files from URL", {
  skip_on_cran()

  link <- shinylive_r_link("shinyApp(fluidPage('test'), function(i, o) {})")
  temp_dir <- tempfile("shinylive_roundtrip")

  result <- decode_shinylive_link(link$url, output_dir = temp_dir, create_subdir = FALSE)

  # Verify file was created
  expect_true(file.exists(file.path(temp_dir, "app.R")))
  expect_s3_class(result, "shinylive_decoded")

  # Cleanup
  unlink(temp_dir, recursive = TRUE)
})

#' Create a webR REPL sharelink from R code
#'
#' Generates a shareable URL for R code that can be executed in the webR environment.
#' Supports expressions, file paths, character strings, and clipboard input.
#'
#' @param input Code input. Can be:
#'   - R expression (no quotes needed): `webr_repl_link({ plot(1:10) })`
#'   - Character string: R code to execute
#'   - File path: Path to R file to read
#'   - NULL: Read from clipboard (requires clipr package)
#' @param filename Name for the file (default: `"script.R"`)
#' @param path Full path where the file will be placed in webR. If NULL (default),
#'   the file is placed at `"/home/web_user/{filename}"`.
#' @param autorun Logical. Whether to auto-execute the code when link is opened (default: `FALSE`).
#'   Only R files (`.R`) can be auto-executed.
#' @param panels Character vector or string specifying which webR interface panels to show.
#'   Valid panels: `"plot"`, `"files"`, `"terminal"`, `"editor"`. Can be `c("plot", "files")` or `"plot-files"`.
#'   If NULL (default), shows all panels.
#' @param version webR version to use (`"latest"` or specific version >= "v0.5.4")
#' @param base_url webR application URL. If NULL, uses global option or builds from version
#'
#' @return webr_link object containing the webR sharelink and metadata
#'
#' @section Comments in expression input:
#' Expression input recovers your source from R's *source references*, which R only
#' attaches when `keep.source` is enabled. Comments therefore survive in an
#' interactive session, but are dropped when the calling code is parsed without
#' source references -- notably inside a knitted 'Quarto' or 'R Markdown' document,
#' because 'knitr' evaluates chunks through `evaluate::evaluate()`, which discards
#' them. No `keep.source` setting recovers them there.
#'
#' If you need comments preserved, pass the code as a string or a file path, or write
#' it as a chunk in the document -- see [livelink-knitr] and
#' `vignette("links-in-documents", package = "livelink")`.
#'
#' @seealso [webr_repl_project()] for multi-file projects and
#'   [webr_repl_exercise()] for exercise and solution pairs; [livelink-knitr] to
#'   give a document chunk its own link; `vignette("getting-started", package =
#'   "livelink")` for an introduction.
#'
#' @export
#' @examples
#' # Expression input (no quotes needed!)
#' webr_repl_link({
#'   plot(1:10)
#'   summary(mtcars)
#' })
#'
#' # Traditional string input
#' webr_repl_link("plot(1:10)")
#'
#' # Choose which panels the REPL shows
#' webr_repl_link({ hist(rnorm(100)) }, panels = c("plot", "editor"))
#'
#' # Run the code as soon as the link opens
#' webr_repl_link("plot(1:10)", autorun = TRUE)
#'
#' # File path input
#' script <- tempfile(fileext = ".R")
#' writeLines("plot(1:10)", script)
#' webr_repl_link(script)
#'
#' # Read the code from the clipboard
#' if (interactive()) {
#'   webr_repl_link()
#' }
webr_repl_link <- function(input = NULL,
                           filename = "script.R",
                           path = NULL,
                           autorun = FALSE,
                           panels = NULL,
                           version = "latest",
                           base_url = NULL) {

  # Capture expression if provided (check expression form without evaluating input)
  x_expr <- substitute(input)
  code <- if (!missing(input) && is_brace_call(x_expr)) {
    process_input(x_expr = x_expr)
  } else {
    process_input(input = input)
  }

  check_single_string(filename, "filename")
  check_single_logical(autorun, "autorun")
  check_valid_mode(panels, "panels")
  check_valid_version(version, "version")

  if (is.null(path)) {
    path <- paste0("/home/web_user/", filename)
  } else {
    check_valid_path(path, "path")
  }

  if (is.null(base_url)) {
    base_url <- get_webr_base_url(version)
  } else {
    check_single_string(base_url, "base_url")
  }

  is_r_file <- grepl("\\.R$", filename, ignore.case = TRUE)
  if (autorun && !is_r_file) {
    cli::cli_warn(c(
      "{.arg autorun} was ignored",
      "!" = "Only R files can be auto-executed, but {.file {filename}} is not one.",
      "i" = "Give {.arg filename} a {.code .R} extension to enable autorun."
    ))
  }
  autorun_enabled <- autorun && is_r_file

  share_item <- list(
    name = filename,
    path = path,
    text = code
  )

  if (autorun_enabled) {
    share_item$autorun <- TRUE
  }

  encoded_data <- encode_webr_payload(list(share_item))

  # The `a` flag must agree with the per-item autorun set above, or the link
  # claims to autorun a file that carries no autorun instruction.
  flags <- if (autorun_enabled) "jza" else "jz"

  url <- build_webr_url(base_url, encoded_data, flags, panels)

  new_webr_link(url, filename, path, panels, version, autorun_enabled)
}

#' Create webR REPL sharelink for multiple files
#'
#' Creates a webR sharelink for projects with multiple R files, data files, or other resources.
#' Supports named lists and file path vectors as input.
#'
#' @param input Input for multiple files. Can be:
#'   - Named list of braced expressions, so each file is written as R rather than
#'     as a string full of escaped newlines:
#'     `list("main.R" = { plot(1:10) }, "utils.R" = { f <- function() 42 })`
#'   - Named list of strings: `list("main.R" = code1, "utils.R" = code2)`
#'   - Vector of file paths: `c("main.R", "utils.R", "data.csv")`
#'
#'   The two list forms mix freely, which is what you want for a project holding
#'   both code and, say, a `README.md`.
#' @param autorun_files Character vector of filenames to auto-execute when project loads, or "all" to autorun all R files (default: none)
#' @param base_path Base directory path for all files (default: `"/home/web_user/"`)
#' @param panels Character vector or string specifying which webR interface panels to show.
#'   Valid panels: `"plot"`, `"files"`, `"terminal"`, `"editor"`. Can be `c("plot", "files")` or `"plot-files"`.
#'   If NULL (default), shows all panels.
#' @param version webR version to use (`"latest"` or specific version >= "v0.5.4")
#' @param base_url webR application URL. If NULL, uses global option or builds from version
#'
#' @return webr_project object containing the webR sharelink for the multi-file project
#'
#' @section Writing a project as code:
#' A file's contents can be given as a `{ ... }` block instead of a string, which
#' spares you escaping every newline and quote:
#'
#' ```
#' webr_repl_project(list(
#'   "main.R"    = { source("utils.R"); summarise(mtcars) },
#'   "utils.R"   = { summarise <- function(d) summary(d) },
#'   "README.md" = "# Analysis"
#' ))
#' ```
#'
#' The blocks are **never evaluated** -- they are source to ship, not code to run
#' -- so an assignment inside one leaves nothing behind in your session.
#'
#' Two things to know. Comments inside `{ }` survive in an interactive session but
#' not in a knitted document (see [webr_repl_link()]). And a `library()` call
#' inside a block is visible to `R CMD check`, which will report the package as an
#' undeclared dependency of *yours*; in a vignette or an example, use a string for
#' code that loads packages.
#'
#' @seealso [webr_repl_link()] for the single-file case.
#'
#' @export
#' @examples
#' # Each file written as R, rather than as an escaped string
#' webr_repl_project(list(
#'   "main.R"  = { source("utils.R"); summarise(mtcars) },
#'   "utils.R" = { summarise <- function(d) summary(d) }
#' ))
#'
#' # Strings still work, and the two forms mix
#' files <- list(
#'   "main.R" = "source('utils.R')\nresult <- analyze_data(mtcars)",
#'   "utils.R" = "analyze_data <- function(data) { summary(data) }",
#'   "README.md" = "# My Analysis\nThis project analyzes the mtcars dataset."
#' )
#' webr_repl_project(files, autorun_files = "main.R")
#'
#' # Autorun every R file in the project
#' webr_repl_project(files, autorun_files = "all")
#'
#' # File paths input
#' project_dir <- tempfile()
#' dir.create(project_dir)
#' main <- file.path(project_dir, "main.R")
#' utils <- file.path(project_dir, "utils.R")
#' writeLines("source('utils.R')", main)
#' writeLines("# utils", utils)
#' webr_repl_project(c(main, utils))
webr_repl_project <- function(input,
                              autorun_files = character(0),
                              base_path = "/home/web_user/",
                              panels = NULL,
                              version = "latest",
                              base_url = NULL) {

  # Captured, not forced: a literal list() may name each file's contents as a
  # `{ ... }` block, and forcing it would run those blocks instead of shipping
  # them.
  x_expr <- substitute(input)
  processed_files <- process_project_input(
    input = input, x_expr = x_expr, env = parent.frame()
  )

  check_character_vector(autorun_files, "autorun_files")
  check_valid_path(base_path, "base_path")
  check_valid_mode(panels, "panels")
  check_valid_version(version, "version")

  autorun_all <- length(autorun_files) == 1 && autorun_files == "all"

  if (length(autorun_files) > 0 && !autorun_all) {
    ensure_files_in_list(autorun_files, processed_files, "input", "autorun_files")
  }

  if (is.null(base_url)) {
    base_url <- get_webr_base_url(version)
  } else {
    check_single_string(base_url, "base_url")
  }

  if (!grepl("/$", base_path)) {
    base_path <- paste0(base_path, "/")
  }

  build_webr_project(processed_files, autorun_files, base_path, panels,
                     version, base_url)
}


#' Bundle a set of named file contents into one webR project link
#'
#' The encoding core shared by [webr_repl_project()] and
#' [webr_repl_directory()]'s `single_link` mode. Inputs are already validated and
#' normalized: `processed_files` is a named list of file contents, `base_path`
#' ends in `/`, and `base_url` is resolved.
#'
#' @param processed_files Named list mapping filename to its content string
#' @param autorun_files Character vector of files to autorun, or `"all"`
#' @param base_path Normalized base directory path
#' @param panels Panels to show, or NULL
#' @param version WebR version
#' @param base_url Resolved WebR base URL
#' @return A `webr_project` object
#' @noRd
build_webr_project <- function(processed_files, autorun_files, base_path,
                               panels, version, base_url) {
  autorun_all <- length(autorun_files) == 1 && autorun_files == "all"

  share_items <- mapply(function(content, filename) {
    item <- list(
      name = filename,
      path = paste0(base_path, filename),
      text = content
    )

    should_autorun <- grepl("\\.R$", filename, ignore.case = TRUE) &&
      (autorun_all || filename %in% autorun_files)

    if (should_autorun) {
      item$autorun <- TRUE
    }

    item
  }, processed_files, names(processed_files), SIMPLIFY = FALSE, USE.NAMES = FALSE)

  encoded_data <- encode_webr_payload(share_items)

  # Set the `a` flag whenever any file carries autorun, not only for
  # `autorun_files = "all"` -- a named autorun list used to encode the per-item
  # flag but never the URL flag that activates it.
  any_autorun <- any(vapply(share_items, function(i) isTRUE(i$autorun), logical(1)))
  flags <- if (any_autorun) "jza" else "jz"

  url <- build_webr_url(base_url, encoded_data, flags, panels)

  new_webr_project(url, processed_files, base_path, panels, version, autorun_files)
}


#' Create paired exercise and solution webR REPL links
#'
#' Generates a pair of webR links for educational purposes: one for student exercises
#' (without autorun) and one for solutions (with autorun enabled).
#'
#' @param exercise_text Character string containing the exercise code with placeholders or TODOs
#' @param solution_text Character string containing the complete solution code
#' @param exercise_name Base name for the exercise (will create `"name_exercise.R"` and `"name_solution.R"`)
#' @param base_path Base directory path for files (default: `"/home/web_user/"`)
#' @param version webR version to use ("latest" or specific version >= "v0.5.4")
#' @param base_url webR application URL. If NULL, uses global option or builds from version
#'
#' @return webr_exercise object holding the paired `exercise` and `solution` links
#'
#' @seealso [webr_repl_link()], which this builds on;
#'   `vignette("teaching", package = "livelink")` for using links in a course.
#'
#' @examples
#' exercise_code <- "
#' # Exercise: Calculate mean of mtcars$mpg
#' # TODO: Complete the line below
#' mean_mpg <- 0
#' print(mean_mpg)
#' "
#'
#' solution_code <- "
#' # Solution: Calculate mean of mtcars$mpg
#' mean_mpg <- mean(mtcars$mpg)
#' print(mean_mpg)
#' "
#'
#' links <- webr_repl_exercise(exercise_code, solution_code, "basic_stats")
#' links$exercise
#' links$solution
#'
#' # Custom path and version
#' webr_repl_exercise(exercise_code, solution_code, "stats",
#'                    base_path = "/exercises/", version = "v0.5.4")
#'
#' @export
webr_repl_exercise <- function(exercise_text,
                               solution_text,
                               exercise_name,
                               base_path = "/home/web_user/",
                               version = "latest",
                               base_url = NULL) {

  check_single_string(exercise_text, "exercise_text")
  check_single_string(solution_text, "solution_text")
  check_single_string(exercise_name, "exercise_name")
  check_valid_path(base_path, "base_path")
  check_valid_version(version, "version")

  if (is.null(base_url)) {
    base_url <- get_webr_base_url(version)
  } else {
    check_single_string(base_url, "base_url")
  }

  if (!grepl("/$", base_path)) {
    base_path <- paste0(base_path, "/")
  }

  exercise_filename <- paste0(exercise_name, "_exercise.R")
  solution_filename <- paste0(exercise_name, "_solution.R")

  # Exercise link (no autorun - student works on it)
  exercise_link <- webr_repl_link(
    exercise_text,
    filename = exercise_filename,
    path = paste0(base_path, exercise_filename),
    autorun = FALSE,
    version = version,
    base_url = base_url
  )

  # Solution link (with autorun - shows complete solution)
  solution_link <- webr_repl_link(
    solution_text,
    filename = solution_filename,
    path = paste0(base_path, solution_filename),
    autorun = TRUE,
    version = version,
    base_url = base_url
  )

  new_webr_exercise(exercise_link, solution_link)
}

test_that("set_webr_base_url sets and resets the global option", {
  withr::local_options(livelink.base_url = NULL)

  set_webr_base_url("https://example.org/webr/")
  expect_equal(getOption("livelink.base_url"), "https://example.org/webr/")

  set_webr_base_url(NULL)
  expect_null(getOption("livelink.base_url"))
})

test_that("set_webr_base_url returns its value invisibly", {
  withr::local_options(livelink.base_url = NULL)

  expect_invisible(set_webr_base_url("https://example.org/webr/"))
  expect_equal(withVisible(set_webr_base_url("https://example.org/webr/"))$value,
               "https://example.org/webr/")
})

test_that("set_webr_base_url validates its input", {
  withr::local_options(livelink.base_url = NULL)

  expect_error(set_webr_base_url(42))
  expect_error(set_webr_base_url(c("a", "b")))
})

test_that("link functions honor the global base URL option", {
  withr::local_options(livelink.base_url = "https://example.org/webr/")

  expect_match(as.character(webr_repl_link("plot(1:10)")),
               "^https://example\\.org/webr/")
  expect_match(as.character(webr_repl_project(list("a.R" = "x <- 1"))),
               "^https://example\\.org/webr/")
})

test_that("the default base URL is used when no option is set", {
  withr::local_options(livelink.base_url = NULL)

  expect_match(as.character(webr_repl_link("plot(1:10)")),
               "^https://webr\\.r-wasm\\.org/")
})

test_that("an explicit base_url argument beats the global option", {
  withr::local_options(livelink.base_url = "https://option.example/webr/")

  url <- as.character(webr_repl_link("x <- 1", base_url = "https://arg.example/webr/"))

  expect_match(url, "^https://arg\\.example/webr/")
})

# Tests for S3 class methods

# webr_link class tests
test_that("webr_link has correct structure", {
  result <- webr_repl_link("x <- 1")

  expect_s3_class(result, "webr_link")
  expect_named(result, c("url", "filename", "path", "mode", "version", "autorun"))
})

test_that("print.webr_link returns invisibly", {
  result <- webr_repl_link("x <- 1")

  expect_invisible(print(result))
  expect_identical(print(result), result)
})

test_that("as.character.webr_link returns URL", {
  result <- webr_repl_link("x <- 1")

  expect_equal(as.character(result), result$url)
})

# webr_project class tests
test_that("webr_project has correct structure", {
  skip_on_cran()
  # Create temp files for testing

  temp_dir <- tempfile("project_test")
  dir.create(temp_dir)
  writeLines("# main code", file.path(temp_dir, "main.R"))
  writeLines("# utils code", file.path(temp_dir, "utils.R"))

  files <- c(file.path(temp_dir, "main.R"), file.path(temp_dir, "utils.R"))
  result <- webr_repl_project(files)

  expect_s3_class(result, "webr_project")
  expect_named(result, c("url", "files", "base_path", "mode", "version", "autorun_files"))

  unlink(temp_dir, recursive = TRUE)
})

test_that("print.webr_project returns invisibly", {
  skip_on_cran()
  temp_dir <- tempfile("project_print")
  dir.create(temp_dir)
  writeLines("# code", file.path(temp_dir, "main.R"))

  result <- webr_repl_project(file.path(temp_dir, "main.R"))

  expect_invisible(print(result))
  expect_identical(print(result), result)

  unlink(temp_dir, recursive = TRUE)
})

test_that("as.character.webr_project returns URL", {
  skip_on_cran()
  temp_dir <- tempfile("project_char")
  dir.create(temp_dir)
  writeLines("# code", file.path(temp_dir, "main.R"))

  result <- webr_repl_project(file.path(temp_dir, "main.R"))

  expect_equal(as.character(result), result$url)

  unlink(temp_dir, recursive = TRUE)
})

# webr_exercise class tests
test_that("webr_exercise has correct structure", {
  result <- livelink:::webr_repl_exercise("# Exercise", "# Solution", "test")

  expect_s3_class(result, "webr_exercise")
  expect_named(result, c("exercise", "solution"))
  expect_s3_class(result$exercise, "webr_link")
  expect_s3_class(result$solution, "webr_link")
})

test_that("webr_exercise creates correct filenames", {
  result <- livelink:::webr_repl_exercise("ex", "sol", "stats")

  expect_equal(result$exercise$filename, "stats_exercise.R")
  expect_equal(result$solution$filename, "stats_solution.R")
})

test_that("webr_exercise solution has autorun enabled", {
  result <- livelink:::webr_repl_exercise("ex", "sol", "test")

  expect_false(result$exercise$autorun)
  expect_true(result$solution$autorun)
})

test_that("print.webr_exercise returns invisibly", {
  result <- livelink:::webr_repl_exercise("ex", "sol", "test")

  expect_invisible(print(result))
  expect_identical(print(result), result)
})

# shinylive_link class tests
test_that("shinylive_link has correct structure", {
  result <- shinylive_r_link("shinyApp(fluidPage(), function(i, o) {})")

  expect_s3_class(result, "shinylive_link")
  expect_named(result, c("url", "files", "engine", "mode"))
})

test_that("print.shinylive_link returns invisibly", {
  result <- shinylive_r_link("shinyApp(fluidPage(), function(i, o) {})")

  expect_invisible(print(result))
})

test_that("as.character.shinylive_link returns URL", {
  result <- shinylive_r_link("shinyApp(fluidPage(), function(i, o) {})")

  expect_equal(as.character(result), result$url)
})

# shinylive_project class tests
test_that("shinylive_project has correct structure", {
  skip_on_cran()
  temp_dir <- tempfile("shinylive_proj")
  dir.create(temp_dir)
  writeLines("# app code", file.path(temp_dir, "app.R"))
  writeLines("# helpers", file.path(temp_dir, "utils.R"))

  files <- c(file.path(temp_dir, "app.R"), file.path(temp_dir, "utils.R"))
  result <- shinylive_project(files, engine = "r")

  expect_s3_class(result, "shinylive_project")
  expect_named(result, c("url", "files", "engine", "mode"))

  unlink(temp_dir, recursive = TRUE)
})

test_that("print.shinylive_project returns invisibly", {
  skip_on_cran()
  temp_dir <- tempfile("shinylive_print")
  dir.create(temp_dir)
  writeLines("# app code", file.path(temp_dir, "app.R"))

  result <- shinylive_project(file.path(temp_dir, "app.R"), engine = "r")

  expect_invisible(print(result))
  expect_identical(print(result), result)

  unlink(temp_dir, recursive = TRUE)
})

test_that("as.character.shinylive_project returns URL", {
  skip_on_cran()
  temp_dir <- tempfile("shinylive_char")
  dir.create(temp_dir)
  writeLines("# app code", file.path(temp_dir, "app.R"))

  result <- shinylive_project(file.path(temp_dir, "app.R"), engine = "r")

  expect_equal(as.character(result), result$url)

  unlink(temp_dir, recursive = TRUE)
})

# webr_preview class tests
test_that("webr_preview has correct structure", {
  link <- webr_repl_link("x <- 1")
  result <- preview_webr_link(link$url)

  expect_s3_class(result, "webr_preview")
  expect_true("url" %in% names(result))
  expect_true("files_data" %in% names(result))
  expect_true("total_files" %in% names(result))
  expect_true("total_size" %in% names(result))
})

test_that("print.webr_preview returns invisibly", {
  link <- webr_repl_link("x <- 1")
  result <- preview_webr_link(link$url)

  expect_invisible(print(result))
  expect_identical(print(result), result)
})

# repl_urls generic tests
test_that("repl_urls works for all link types", {
  webr_link <- webr_repl_link("x <- 1")
  shinylive_link <- shinylive_r_link("shinyApp(fluidPage(), function(i, o) {})")

  expect_type(repl_urls(webr_link), "character")
  expect_type(repl_urls(shinylive_link), "character")

  expect_equal(repl_urls(webr_link), webr_link$url)
  expect_equal(repl_urls(shinylive_link), shinylive_link$url)
})

# Tests for webr_repl_link functions

test_that("webr_repl_link creates valid webr_link object", {
  result <- webr_repl_link("1 + 1")

  expect_s3_class(result, "webr_link")
  expect_true(grepl("webr.r-wasm.org", result$url))
  expect_true(grepl("#code=", result$url))
  expect_equal(result$filename, "script.R")
  expect_equal(result$path, "/home/web_user/script.R")
  expect_equal(result$version, "latest")
  expect_false(result$autorun)
})

test_that("webr_repl_link respects custom filename", {
  result <- webr_repl_link("x <- 1", filename = "analysis.R")

  expect_equal(result$filename, "analysis.R")
  expect_equal(result$path, "/home/web_user/analysis.R")
})

test_that("webr_repl_link respects custom path", {
  result <- webr_repl_link("x <- 1", path = "/custom/path/script.R")

  expect_equal(result$path, "/custom/path/script.R")
})

test_that("webr_repl_link respects autorun", {
  result <- webr_repl_link("print('hello')", autorun = TRUE)

  expect_true(result$autorun)
  expect_true(grepl("&jza", result$url)) # 'a' flag for autorun
})

test_that("webr_repl_link respects mode parameter", {
  result <- webr_repl_link("x <- 1", panels = "editor")

  expect_equal(result$mode, "editor")
  expect_true(grepl("mode='editor'", result$url))
})

test_that("webr_repl_link respects version parameter", {
  result <- webr_repl_link("x <- 1", version = "v0.5.4")

  expect_equal(result$version, "v0.5.4")
  expect_true(grepl("v0.5.4", result$url))
})

test_that("webr_repl_link validates inputs", {
  expect_error(webr_repl_link("x", filename = 123))
  expect_error(webr_repl_link("x", autorun = "yes"))
  expect_error(webr_repl_link("x", panels = "invalid"))
  expect_error(webr_repl_link("x", version = "v0.4.0"))
})

test_that("webr_repl_project creates valid webr_project object", {
  skip_on_cran()
  temp_dir <- tempfile("webr_proj")
  dir.create(temp_dir)
  writeLines("source('utils.R')", file.path(temp_dir, "main.R"))
  writeLines("helper <- function() 1", file.path(temp_dir, "utils.R"))

  files <- c(file.path(temp_dir, "main.R"), file.path(temp_dir, "utils.R"))
  result <- webr_repl_project(files)

  expect_s3_class(result, "webr_project")
  expect_true(grepl("webr.r-wasm.org", result$url))
  expect_equal(length(result$files), 2)
  expect_true("main.R" %in% result$files)
  expect_true("utils.R" %in% result$files)

  unlink(temp_dir, recursive = TRUE)
})

test_that("webr_repl_project respects autorun_files", {
  skip_on_cran()
  temp_dir <- tempfile("webr_autorun")
  dir.create(temp_dir)
  writeLines("print('main')", file.path(temp_dir, "main.R"))
  writeLines("print('utils')", file.path(temp_dir, "utils.R"))

  files <- c(file.path(temp_dir, "main.R"), file.path(temp_dir, "utils.R"))
  result <- webr_repl_project(files, autorun_files = "main.R")

  expect_equal(result$autorun_files, "main.R")

  unlink(temp_dir, recursive = TRUE)
})

test_that("webr_repl_project validates autorun_files exist in files", {
  skip_on_cran()
  temp_dir <- tempfile("webr_validate")
  dir.create(temp_dir)
  writeLines("code", file.path(temp_dir, "main.R"))

  expect_error(
    webr_repl_project(file.path(temp_dir, "main.R"), autorun_files = "nonexistent.R"),
    "not found"
  )

  unlink(temp_dir, recursive = TRUE)
})

test_that("repl_urls extracts URL from webr_link", {
  result <- webr_repl_link("1 + 1")
  url <- repl_urls(result)

  expect_type(url, "character")
  expect_equal(url, result$url)
})

test_that("as.character works for webr_link", {
  result <- webr_repl_link("1 + 1")
  url <- as.character(result)

  expect_type(url, "character")
  expect_equal(url, result$url)
})

#' Extract URLs as character vector
#'
#' @param x Link object
#' @param ... Additional arguments
#'
#' @return
#' A character vector of URLs. Most objects yield a single URL; exercise objects
#' return a length-2 named vector (`exercise`, `solution`); directory and batch
#' objects return one URL per file.
#'
#' @seealso [repl_urls()] for the same extraction as a named generic you can call
#'   explicitly.
#'
#' @export
as.character.webr_link <- function(x, ...) {
  x$url
}

#' @rdname as.character.webr_link
#' @export
as.character.shinylive_link <- function(x, ...) {
  x$url
}

#' @rdname as.character.webr_link
#' @export
as.character.webr_project <- function(x, ...) {
  x$url
}

#' @rdname as.character.webr_link
#' @export
as.character.webr_exercise <- function(x, ...) {
  c(exercise = x$exercise$url, solution = x$solution$url)
}

#' @rdname as.character.webr_link
#' @export
as.character.webr_directory <- function(x, ...) {
  x$urls
}

#' @rdname as.character.webr_link
#' @export
as.character.webr_decoded <- function(x, ...) {
  x$url
}

#' @rdname as.character.webr_link
#' @export
as.character.webr_decoded_batch <- function(x, ...) {
  x$urls
}

#' @rdname as.character.webr_link
#' @export
as.character.webr_preview <- function(x, ...) {
  x$url
}

#' @rdname as.character.webr_link
#' @export
as.character.shinylive_project <- function(x, ...) {
  x$url
}

#' @rdname as.character.webr_link
#' @export
as.character.shinylive_directory <- function(x, ...) {
  x$urls
}

#' @rdname as.character.webr_link
#' @export
as.character.shinylive_decoded <- function(x, ...) {
  x$url
}

#' @rdname as.character.webr_link
#' @export
as.character.shinylive_decoded_batch <- function(x, ...) {
  x$urls
}

#' @rdname as.character.webr_link
#' @export
as.character.shinylive_preview <- function(x, ...) {
  x$url
}

#' Extract shareable URLs from livelink objects
#'
#' @description
#' Generic function to extract the shareable URL(s) from any livelink object,
#' covering both webR REPL and Shinylive results. Provides a clear way to get
#' just the URLs for sharing or further processing.
#'
#' @param x A livelink object. Supported classes are `webr_link`, `webr_project`,
#'   `webr_exercise`, `webr_directory`, `webr_decoded`, `webr_decoded_batch`,
#'   `webr_preview`, `shinylive_link`, `shinylive_project`, `shinylive_directory`,
#'   `shinylive_decoded`, `shinylive_decoded_batch`, and `shinylive_preview`.
#' @param ... Additional arguments passed to methods
#' @return A character vector of URLs. Most objects yield a single URL; exercise
#'   objects return a length-2 named vector (`exercise`, `solution`); directory
#'   and batch objects return one URL per file.
#'
#' @seealso The `as.character()` methods (for example [as.character.webr_link()]),
#'   which `repl_urls()` delegates to.
#'
#' @examples
#' # Single link
#' link <- webr_repl_link("plot(1:10)")
#' repl_urls(link)
#'
#' # Exercise (returns named vector)
#' exercise <- webr_repl_exercise("# TODO", "plot(1:10)", "test")
#' repl_urls(exercise)
#'
#' # Shinylive links work the same way
#' repl_urls(shinylive_r_link("library(shiny)"))
#'
#' # Decoded files (returns the original URL)
#' decoded <- decode_webr_link(as.character(link))
#' repl_urls(decoded)
#'
#' @export
repl_urls <- function(x, ...) {
  UseMethod("repl_urls")
}

# The classes repl_urls understands. Every one carries an as.character() method
# returning its URL(s), and repl_urls is a named alias for exactly that, so the
# default method delegates rather than repeating thirteen identical bodies.
livelink_url_classes <- c(
  "webr_link", "webr_project", "webr_exercise", "webr_directory",
  "webr_decoded", "webr_decoded_batch", "webr_preview",
  "shinylive_link", "shinylive_project", "shinylive_directory",
  "shinylive_decoded", "shinylive_decoded_batch", "shinylive_preview"
)

#' @rdname repl_urls
#' @export
repl_urls.default <- function(x, ...) {
  if (inherits(x, livelink_url_classes)) {
    return(as.character(x, ...))
  }
  cli::cli_abort(c(
    "Cannot extract URLs from object of class {.cls {class(x)}}",
    "i" = "Supported classes: {.cls {livelink_url_classes}}"
  ))
}

# Generated by data-raw/aspell.R. R CMD check --as-cran reads this file to
# accept the words in .aspell/words.rds during its spelling pass.
Rd_files <- vignettes <- R_files <- description <-
  list(encoding = "UTF-8", dictionaries = c("en_stats", "words"))

#' Decode Shinylive link(s) to extract files to local directory
#'
#' @description
#' Decodes Shinylive sharelinks to extract the embedded files and save them
#' to a local directory. This is the reverse operation of creating Shinylive links.
#' Handles both single URLs and multiple URLs automatically.
#'
#' @param url Character string or vector containing Shinylive URL(s)
#' @param output_dir Character string specifying the output directory path.
#'   Defaults to a `shinylive_files` directory inside the session temporary
#'   directory; pass an explicit path to extract somewhere permanent.
#' @param overwrite Logical. Whether to overwrite existing files (default: FALSE)
#' @param create_subdir Logical. If `TRUE` (default), each decoded link is
#'   extracted into its own subdirectory under `output_dir` rather than directly
#'   into it. For a single URL the subdirectory is named `shinylive_<hash>`, where
#'   `<hash>` is a short fingerprint of the URL. For multiple URLs, see
#'   `name_dirs`. Set `FALSE` to extract straight into `output_dir`; with multiple
#'   URLs this means identically named files (such as `app.R`) collide, so a later
#'   app overwrites an earlier one when `overwrite = TRUE`, or is skipped when
#'   `overwrite = FALSE`.
#' @param name_dirs Logical. For multiple URLs, controls how the per-link
#'   subdirectories are named: `TRUE` (default) numbers them `app_01`, `app_02`,
#'   ...; `FALSE` names each one `shinylive_<hash>` from the URL fingerprint.
#'   Ignored for a single URL, and ignored when `create_subdir = FALSE`.
#'
#' @return For a single URL, a `shinylive_decoded` object. For multiple URLs, a
#'   `shinylive_decoded_batch` object.
#'
#' @seealso [preview_shinylive_link()] to inspect a link without writing files;
#'   [shinylive_r_link()] and [shinylive_py_link()] to create links.
#'
#' @examples
#' # Round-trip: build a link, then decode it back to files
#' url <- as.character(shinylive_r_link("library(shiny)"))
#'
#' result <- decode_shinylive_link(url)
#' print(result)
#'
#' # Extract to a directory of your choosing
#' out <- file.path(tempdir(), "my_app")
#' decode_shinylive_link(url, output_dir = out, create_subdir = FALSE, overwrite = TRUE)
#' list.files(out)
#'
#' # Both engines at once
#' urls <- c(url, as.character(shinylive_py_link("from shiny import App")))
#' decode_shinylive_link(urls, output_dir = file.path(tempdir(), "my_apps"))
#'
#' @export
decode_shinylive_link <- function(url,
                                  output_dir = file.path(tempdir(), "shinylive_files"),
                                  overwrite = FALSE,
                                  create_subdir = TRUE,
                                  name_dirs = TRUE) {

  # Validate basic inputs
  check_character_vector(url, "url")
  check_single_string(output_dir, "output_dir")
  check_single_logical(overwrite, "overwrite")
  check_single_logical(create_subdir, "create_subdir")
  check_single_logical(name_dirs, "name_dirs")

  # Handle single URL case
  if (length(url) == 1) {
    return(decode_single_shinylive_link(url, output_dir, overwrite, create_subdir))
  }

  # Handle multiple URLs case
  return(decode_multiple_shinylive_links(url, output_dir, overwrite, create_subdir, name_dirs))
}

#' Decode a single Shinylive link
#' @param url Single URL string
#' @param output_dir Output directory
#' @param overwrite Whether to overwrite files
#' @param create_subdir Whether to create subdirectory
#' @return shinylive_decoded object
#' @noRd
decode_single_shinylive_link <- function(url, output_dir, overwrite, create_subdir) {
  # Validate URL
  check_valid_shinylive_url(url, "url")

  # Decompress and parse URL data
  cli::cli_inform("Decompressing Shinylive data...")
  url_data <- decompress_shinylive_url(url)

  engine <- url_data$engine
  mode <- url_data$mode
  files_data <- url_data$files_data

  cli::cli_inform("Parsing file data...")

  # Determine output directory
  final_output_dir <- if (create_subdir) {
    # Create subdirectory based on URL hash
    url_hash <- simple_hash(url, 8)
    file.path(output_dir, paste0("shinylive_", url_hash))
  } else {
    output_dir
  }

  # Create output directory
  if (!dir.exists(final_output_dir)) {
    dir.create(final_output_dir, recursive = TRUE, showWarnings = FALSE)
    cli::cli_inform("Created directory: {.path {final_output_dir}}")
  }

  # Decode and save files
  files_info <- decode_and_save_files(files_data, final_output_dir, overwrite)

  # Summary
  cli::cli_inform(c(
    "v" = "Successfully decoded {nrow(files_info)} file{?s} to {.path {final_output_dir}}"
  ))

  # Return shinylive_decoded object
  new_shinylive_decoded(files_info, final_output_dir, url, engine, mode)
}

#' Decode multiple Shinylive links
#' @param urls Vector of URL strings
#' @param output_dir Base output directory
#' @param overwrite Whether to overwrite files
#' @param create_subdir Whether to create subdirectories
#' @param name_dirs Whether to use numbered directory names
#' @return shinylive_decoded_batch object
#' @noRd
decode_multiple_shinylive_links <- function(urls, output_dir, overwrite, create_subdir, name_dirs) {
  if (length(urls) == 0) {
    cli::cli_warn("No URLs provided")
    return(new_shinylive_decoded_batch(list(), output_dir, character(0)))
  }

  # Validate all URLs first
  for (i in seq_along(urls)) {
    tryCatch({
      check_valid_shinylive_url(urls[i], paste0("url[", i, "]"))
    }, error = function(e) {
      cli::cli_abort(c(
        "Invalid URL at position {i}",
        "x" = "{e$message}"
      ))
    })
  }

  cli::cli_inform("Processing {length(urls)} Shinylive URL{?s}...")

  results <- list()

  for (i in seq_along(urls)) {
    url <- urls[i]

    tryCatch({
      # Determine subdirectory name
      if (!create_subdir) {
        # Use base output directory for each URL
        url_output_dir <- output_dir
        subdir_name <- paste0("url_", i)
      } else if (name_dirs) {
        subdir_name <- sprintf("app_%02d", i)
        url_output_dir <- file.path(output_dir, subdir_name)
      } else {
        url_hash <- simple_hash(url, 8)
        subdir_name <- paste0("shinylive_", url_hash)
        url_output_dir <- file.path(output_dir, subdir_name)
      }

      cli::cli_inform("")
      cli::cli_h3("Processing URL {i}/{length(urls)}: {subdir_name}")

      # Decode files using single URL logic
      decoded_result <- decode_single_shinylive_link(
        url = url,
        output_dir = url_output_dir,
        overwrite = overwrite,
        create_subdir = FALSE  # We handle subdirectory creation here
      )

      results[[subdir_name]] <- decoded_result

    }, error = function(e) {
      cli::cli_warn(c(
        "Failed to process URL {i}",
        "x" = "{e$message}"
      ))
    })
  }

  # A failed URL is warned about above and simply leaves no entry in `results`.
  successful_results <- results
  cli::cli_inform(c(
    "",
    "v" = "Successfully processed {length(successful_results)}/{length(urls)} URL{?s}"
  ))

  # Return shinylive_decoded_batch object
  new_shinylive_decoded_batch(results, output_dir, urls)
}

#' Decode and save files from Shinylive data
#' @param files_data Parsed file data from Shinylive
#' @param output_dir Directory to save files
#' @param overwrite Whether to overwrite existing files
#' @return Data frame with file information
#' @noRd
decode_and_save_files <- function(files_data, output_dir, overwrite) {
  cli::cli_inform("Decoding {length(files_data)} file{?s}...")

  files_info <- data.frame(
    filename = character(0),
    path = character(0),
    type = character(0),
    size_bytes = numeric(0),
    stringsAsFactors = FALSE
  )

  for (i in seq_along(files_data)) {
    file_info <- files_data[[i]]

    # Validate file structure
    if (!all(c("name", "content") %in% names(file_info))) {
      cli::cli_warn("Skipping invalid file entry {i}: missing required fields")
      next
    }

    filename <- file_info$name
    content <- file_info$content
    file_type <- if (!is.null(file_info$type)) file_info$type else "text"

    # Determine file path
    file_path <- file.path(output_dir, filename)

    # Create subdirectories if needed
    file_dir <- dirname(file_path)
    if (!dir.exists(file_dir)) {
      dir.create(file_dir, recursive = TRUE, showWarnings = FALSE)
    }

    # Check if file exists and handle overwrite
    if (file.exists(file_path) && !overwrite) {
      cli::cli_warn("File already exists, skipping: {.file {filename}}")
      next
    }

    # Save file based on type
    tryCatch({
      if (file_type == "binary") {
        # Decode base64 and write binary data
        binary_data <- base64enc::base64decode(content)
        writeBin(binary_data, file_path)
        file_size <- length(binary_data)
      } else {
        # Write text data
        writeLines(content, file_path, useBytes = TRUE)
        file_size <- nchar(content, type = "bytes")
      }

      # Add to results
      files_info <- rbind(files_info, data.frame(
        filename = filename,
        path = file_path,
        type = file_type,
        size_bytes = file_size,
        stringsAsFactors = FALSE
      ))

      size_label <- format_file_size(file_size)
      cli::cli_inform("  {.file {filename}} ({file_type}, {size_label})")

    }, error = function(e) {
      cli::cli_warn(c(
        "Failed to save file: {.file {filename}}",
        "x" = "{e$message}"
      ))
    })
  }

  files_info
}

#' Decompress and parse Shinylive URL data
#' @param url Shinylive URL
#' @return List containing engine, mode, and files_data
#' @noRd
decompress_shinylive_url <- function(url) {
  # Extract URL metadata
  url_parts <- strsplit(url, "/")[[1]]
  engine <- if ("py" %in% url_parts) "python" else "r"
  mode <- if ("app" %in% url_parts) "app" else "editor"

  # Extract the encoded data from URL
  url_fragment_parts <- strsplit(url, "#")[[1]]
  if (length(url_fragment_parts) < 2) {
    cli::cli_abort(c(
      "No encoded data found in URL",
      "x" = "URL must contain encoded data after the '#' symbol"
    ))
  }

  # Parse URL parameters
  params_string <- url_fragment_parts[2]
  encoded_data <- extract_code_parameter(params_string)

  if (is.null(encoded_data)) {
    cli::cli_abort(c(
      "No 'code' parameter found in URL",
      "x" = "URL must contain a 'code' parameter with the encoded file data"
    ))
  }

  # Decompress the data
  tryCatch({

    # Decompress using LZ-string
    json_string <- lzstring::decompressFromEncodedURIComponent(encoded_data)

  }, error = function(e) {
    cli::cli_abort(c(
      "Error during decompression",
      "x" = "Failed to decompress LZ-string data: {e$message}",
      "i" = "The URL may be corrupted or use an unsupported encoding format"
    ))
  })

  # Parse JSON
  tryCatch({
    files_data <- jsonlite::fromJSON(json_string, simplifyVector = FALSE)
  }, error = function(e) {
    cli::cli_abort(c(
      "Error parsing JSON data",
      "x" = "Failed to parse file data: {e$message}",
      "i" = "The decoded data may not be valid JSON"
    ))
  })

  list(
    engine = engine,
    mode = mode,
    files_data = files_data
  )
}

#' Extract code parameter from Shinylive URL fragment
#' @param params_string URL fragment part after '#'
#' @return Encoded data string or NULL if not found
#' @noRd
extract_code_parameter <- function(params_string) {
  # Handle both formats:
  # 1. Direct code parameter: code=...
  # 2. With other parameters: h=0&code=...

  # Split by & to handle multiple parameters
  params <- strsplit(params_string, "&")[[1]]

  for (param in params) {
    if (grepl("^code=", param)) {
      return(sub("^code=", "", param))
    }
  }

  return(NULL)
}

#' Preview Shinylive link contents without writing files to disk
#'
#' @description
#' Decodes a Shinylive URL and returns information about the embedded files
#' without actually saving them to disk. Use print method options to control display.
#'
#' @param url Character string containing the Shinylive URL
#'
#' @return shinylive_preview object with file information and metadata
#'
#' @seealso [decode_shinylive_link()] to extract files to disk;
#'   [shinylive_r_link()] and [shinylive_py_link()] to create links.
#'
#' @examples
#' url <- as.character(shinylive_r_link("library(shiny)"))
#'
#' # Inspect a link without writing anything to disk
#' preview <- preview_shinylive_link(url)
#'
#' # Default print (no content)
#' print(preview)
#'
#' # Show file contents
#' print(preview, show_content = TRUE)
#'
#' # Show file contents with custom length limit
#' print(preview, show_content = TRUE, max_content_length = 200)
#'
#' # Access the preview data
#' preview$files_data
#' preview$total_files
#'
#' @export
preview_shinylive_link <- function(url) {

  # Validate inputs
  check_valid_shinylive_url(url, "url")

  # Decompress and parse URL data
  tryCatch({
    url_data <- decompress_shinylive_url(url)

    engine <- url_data$engine
    mode <- url_data$mode
    files_data <- url_data$files_data

    # Calculate metadata
    total_size <- 0
    file_types <- character(0)

    for (i in seq_along(files_data)) {
      file_info <- files_data[[i]]
      content <- file_info$content
      file_type <- if (!is.null(file_info$type)) file_info$type else "text"

      # Byte size: for binary, derive it from the base64 length rather than
      # decoding the content; for text, count the bytes directly.
      if (file_type == "binary") {
        size_bytes <- calculate_base64_size(content)
      } else {
        size_bytes <- nchar(content, type = "bytes")
      }

      total_size <- total_size + size_bytes
      file_types <- c(file_types, file_type)
    }

    # Return shinylive_preview object
    new_shinylive_preview(url, engine, mode, files_data, total_size, unique(file_types))

  }, error = function(e) {
    cli::cli_abort(c(
      "Failed to preview Shinylive link",
      "x" = "{e$message}"
    ))
  })
}

#' Render livelink objects as links in knitted documents
#'
#' @description
#' 'knitr' calls `knit_print()` on the last value of a chunk. Without these
#' methods a link object would fall back to [print()], dumping cli console output
#' (a header, box glyphs, metadata) into the rendered page. These methods instead
#' emit a clickable Markdown link, which is almost always what you want when a
#' link object is the visible result of a chunk.
#'
#' A single link becomes `[Open in webR](url)` (or Shinylive); a project the
#' same. A directory or an exercise, which carry several named URLs, become a
#' bulleted list, one titled link per entry.
#'
#' These fire only inside 'knitr'. Call [print()] explicitly for the full cli
#' description, or [as.character()] for the bare URL.
#'
#' @param x A livelink object (a link, project, exercise, or directory).
#' @param ... Ignored.
#'
#' @return A `knit_asis` object (via [knitr::asis_output()]).
#'
#' @seealso [livelink-knitr] for the chunk hook and engine, [format.livelink()]
#'   and [as.character.webr_link()] for other renderings, and
#'   `vignette("links-in-documents", package = "livelink")`.
#'
#' @name knit_print.livelink
NULL

# Build the Markdown. A lone unnamed URL renders as one labelled link; a named
# vector (a directory, an exercise) renders as a titled bullet per link.
livelink_knit_markdown <- function(urls, label) {
  if (length(urls) == 0) {
    return(knitr::asis_output("\n_(no links)_\n"))
  }

  if (length(urls) == 1 && is.null(names(urls))) {
    md <- paste0("[", label, "](", urls, ")")
  } else {
    titles <- names(urls)
    if (is.null(titles)) titles <- rep(label, length(urls))
    md <- paste(paste0("- [", titles, "](", urls, ")"), collapse = "\n")
  }

  knitr::asis_output(paste0("\n", md, "\n"))
}

#' @rdname knit_print.livelink
#' @exportS3Method knitr::knit_print
knit_print.webr_link <- function(x, ...) {
  livelink_knit_markdown(as.character(x), "Open in webR")
}

#' @rdname knit_print.livelink
#' @exportS3Method knitr::knit_print
knit_print.webr_project <- function(x, ...) {
  livelink_knit_markdown(as.character(x), "Open in webR")
}

#' @rdname knit_print.livelink
#' @exportS3Method knitr::knit_print
knit_print.webr_exercise <- function(x, ...) {
  livelink_knit_markdown(as.character(x), "Open in webR")
}

#' @rdname knit_print.livelink
#' @exportS3Method knitr::knit_print
knit_print.webr_directory <- function(x, ...) {
  livelink_knit_markdown(as.character(x), "Open in webR")
}

#' @rdname knit_print.livelink
#' @exportS3Method knitr::knit_print
knit_print.shinylive_link <- function(x, ...) {
  livelink_knit_markdown(as.character(x), "Open in Shinylive")
}

#' @rdname knit_print.livelink
#' @exportS3Method knitr::knit_print
knit_print.shinylive_project <- function(x, ...) {
  livelink_knit_markdown(as.character(x), "Open in Shinylive")
}

#' @rdname knit_print.livelink
#' @exportS3Method knitr::knit_print
knit_print.shinylive_directory <- function(x, ...) {
  livelink_knit_markdown(as.character(x), "Open in Shinylive")
}

#' Create a webr_link object
#' @param url WebR URL
#' @param filename Original filename
#' @param path Full path in WebR
#' @param mode Interface mode
#' @param version WebR version
#' @param autorun Whether autorun is enabled
#' @return webr_link object
#' @noRd
new_webr_link <- function(url, filename, path, mode, version, autorun) {
  structure(
    list(
      url = url,
      filename = filename,
      path = path,
      mode = mode,
      version = version,
      autorun = autorun
    ),
    class = "webr_link"
  )
}

#' Create a webr_project object
#' @param url WebR URL
#' @param files Named list of file contents
#' @param base_path Base directory path
#' @param mode Interface mode
#' @param version WebR version
#' @param autorun_files Files with autorun enabled
#' @return webr_project object
#' @noRd
new_webr_project <- function(url, files, base_path, mode, version, autorun_files) {
  structure(
    list(
      url = url,
      files = names(files),
      base_path = base_path,
      mode = mode,
      version = version,
      autorun_files = autorun_files
    ),
    class = "webr_project"
  )
}

#' Create a webr_exercise object
#' @param exercise webr_link object for exercise
#' @param solution webr_link object for solution
#' @return webr_exercise object
#' @noRd
new_webr_exercise <- function(exercise, solution) {
  structure(
    list(
      exercise = exercise,
      solution = solution
    ),
    class = "webr_exercise"
  )
}

#' Create a webr_directory object
#'
#' @param urls Named character vector of URLs
#' @param base_path Base directory path in WebR
#' @param mode Interface mode
#' @param version WebR version
#' @param source_directory Original source directory
#'
#' @return
#' webr_directory object
#'
#' @noRd
new_webr_directory <- function(urls, base_path, mode, version, source_directory) {
  structure(
    list(
      urls = urls,
      base_path = base_path,
      mode = mode,
      version = version,
      source_directory = source_directory
    ),
    class = "webr_directory"
  )
}

#' Create a webr_decoded object
#' @param files_info Data frame with file information
#' @param output_dir Output directory path
#' @param url Original webR URL
#' @param mode Interface mode
#' @param version WebR version
#' @param flags Encoding flags
#' @return webr_decoded object
#' @noRd
new_webr_decoded <- function(files_info, output_dir, url, mode, version, flags) {
  structure(
    list(
      files_info = files_info,
      output_dir = output_dir,
      url = url,
      mode = mode,
      version = version,
      flags = flags,
      # Same summary fields as shinylive_decoded, so callers can treat a decoded
      # result the same way regardless of which environment it came from.
      total_files = nrow(files_info),
      total_size = sum(files_info$size_bytes, na.rm = TRUE)
    ),
    class = "webr_decoded"
  )
}

#' Create a decoded-batch object
#'
#' webr_decoded_batch and shinylive_decoded_batch are field-for-field identical,
#' differing only in their class tag, so both are built here.
#' @param results List of decoded objects
#' @param base_dir Base output directory
#' @param urls Original URLs
#' @param subclass Class tag for the returned object
#' @return An object of class `subclass`
#' @noRd
new_decoded_batch <- function(results, base_dir, urls, subclass) {
  successful <- results[!vapply(results, is.null, logical(1))]

  structure(
    list(
      results = results,
      base_dir = base_dir,
      urls = urls,
      total_urls = length(urls),
      successful_urls = length(successful),
      total_files = sum(vapply(successful, function(x) x$total_files, numeric(1))),
      total_size = sum(vapply(successful, function(x) x$total_size, numeric(1)))
    ),
    class = subclass
  )
}

#' Create a webr_decoded_batch object
#' @param results List of webr_decoded objects
#' @param base_dir Base output directory
#' @param urls Original URLs
#' @return webr_decoded_batch object
#' @noRd
new_webr_decoded_batch <- function(results, base_dir, urls) {
  new_decoded_batch(results, base_dir, urls, "webr_decoded_batch")
}

#' Create a webr_preview object
#' @param url Original webR URL
#' @param mode Interface mode
#' @param version WebR version
#' @param flags Encoding flags
#' @param files_data Raw file data
#' @param total_size Total size in bytes
#' @param autorun_files Files with autorun enabled
#' @return webr_preview object
#' @noRd
new_webr_preview <- function(url, mode, version, flags, files_data, total_size, autorun_files) {
  structure(
    list(
      url = url,
      mode = mode,
      version = version,
      flags = flags,
      files_data = files_data,
      total_files = length(files_data),
      total_size = total_size,
      autorun_files = autorun_files
    ),
    class = "webr_preview"
  )
}


#' Create a shinylive_link object
#' @param url Shinylive URL
#' @param files Named list of file contents
#' @param engine Engine type ("r" or "python")
#' @param mode Shinylive mode ("editor" or "app")
#' @return shinylive_link object
#' @noRd
new_shinylive_link <- function(url, files, engine, mode) {
  structure(
    list(
      url = url,
      files = names(files),
      engine = engine,
      mode = mode
    ),
    class = "shinylive_link"
  )
}

#' Create a shinylive_project object
#' @param url Shinylive URL
#' @param files Named list of file contents
#' @param engine Engine type ("r" or "python")
#' @param mode Shinylive mode ("editor" or "app")
#' @return shinylive_project object
#' @noRd
new_shinylive_project <- function(url, files, engine, mode) {
  structure(
    list(
      url = url,
      files = names(files),
      engine = engine,
      mode = mode
    ),
    class = "shinylive_project"
  )
}

#' Create a shinylive_directory object
#' @param urls Named character vector of URLs
#' @param engine Engine type ("r" or "python")
#' @param mode Shinylive mode ("editor" or "app")
#' @param source_directory Original source directory
#' @return shinylive_directory object
#' @noRd
new_shinylive_directory <- function(urls, engine, mode, source_directory) {
  structure(
    list(
      urls = urls,
      engine = engine,
      mode = mode,
      source_directory = source_directory
    ),
    class = "shinylive_directory"
  )
}

#' Create a shinylive_decoded object
#' @param files_info Data frame with file information
#' @param output_dir Directory where files were saved
#' @param url Original Shinylive URL
#' @param engine Engine type
#' @param mode Shinylive mode
#' @return shinylive_decoded object
#' @noRd
new_shinylive_decoded <- function(files_info, output_dir, url, engine, mode) {
  structure(
    list(
      files_info = files_info,
      output_dir = output_dir,
      url = url,
      engine = engine,
      mode = mode,
      total_files = nrow(files_info),
      total_size = sum(files_info$size_bytes, na.rm = TRUE)
    ),
    class = "shinylive_decoded"
  )
}

#' Create a shinylive_decoded_batch object
#' @param results List of shinylive_decoded objects
#' @param base_dir Base output directory
#' @param urls Original URLs
#' @return shinylive_decoded_batch object
#' @noRd
new_shinylive_decoded_batch <- function(results, base_dir, urls) {
  new_decoded_batch(results, base_dir, urls, "shinylive_decoded_batch")
}

#' Create a shinylive_preview object
#' @param url Original Shinylive URL
#' @param engine Engine type
#' @param mode Shinylive mode
#' @param files_data Raw file data
#' @param total_size Total size in bytes
#' @param file_types Unique file types found
#' @return shinylive_preview object
#' @noRd
new_shinylive_preview <- function(url, engine, mode, files_data, total_size, file_types) {
  structure(
    list(
      url = url,
      engine = engine,
      mode = mode,
      files_data = files_data,
      total_files = length(files_data),
      total_size = total_size,
      file_types = file_types
    ),
    class = "shinylive_preview"
  )
}

# S3 print methods

#' Print method for webr_link objects
#' @param x webr_link object
#' @param ... Additional arguments (ignored)
#' @return Invisibly returns the object
#' @export
print.webr_link <- function(x, ...) {
  cli::cli_h2("webR Link")
  cli::cli_text("{.url {x$url}}")
  cli::cli_text("")

  cli::cli_text("File: {.file {x$filename}} \u2192 {.path {x$path}}")

  formatted_mode <- format_mode_for_display(x$mode)
  if (!is.null(formatted_mode)) {
    cli::cli_text("Interface: {.val {formatted_mode}}")
  }

  cli::cli_text("Version: {.val {x$version}}")
  cli::cli_text("Autorun: {.val {x$autorun}}")

  invisible(x)
}

#' Print method for webr_project objects
#' @param x webr_project object
#' @param ... Additional arguments (ignored)
#' @return Invisibly returns the object
#' @export
print.webr_project <- function(x, ...) {
  cli::cli_h2("webR Project")
  cli::cli_text("{.url {x$url}}")
  cli::cli_text("")

  cli::cli_text("Files ({length(x$files)}):")
  for (file in x$files) {
    path <- paste0(x$base_path, file)
    if (file %in% x$autorun_files) {
      cli::cli_text("  {.file {file}} \u2192 {.path {path}} {.emph (autorun)}")
    } else {
      cli::cli_text("  {.file {file}} \u2192 {.path {path}}")
    }
  }

  formatted_mode <- format_mode_for_display(x$mode)
  if (!is.null(formatted_mode)) {
    cli::cli_text("")
    cli::cli_text("Interface: {.val {formatted_mode}}")
  }

  cli::cli_text("Version: {.val {x$version}}")

  invisible(x)
}

#' Print method for webr_exercise objects
#' @param x webr_exercise object
#' @param ... Additional arguments (ignored)
#' @return Invisibly returns the object
#' @export
print.webr_exercise <- function(x, ...) {
  cli::cli_h2("webR Exercise")
  cli::cli_text("")

  cli::cli_h3("Exercise")
  cli::cli_text("{.url {x$exercise$url}}")
  cli::cli_text("File: {.file {x$exercise$filename}} \u2192 {.path {x$exercise$path}}")
  cli::cli_text("")

  cli::cli_h3("Solution")
  cli::cli_text("{.url {x$solution$url}}")
  cli::cli_text("File: {.file {x$solution$filename}} \u2192 {.path {x$solution$path}} {.emph (autorun)}")
  cli::cli_text("")

  formatted_mode <- format_mode_for_display(x$exercise$mode)
  if (!is.null(formatted_mode)) {
    cli::cli_text("Interface: {.val {formatted_mode}}")
  }

  cli::cli_text("Version: {.val {x$exercise$version}}")

  invisible(x)
}

#' Print method for webr_directory objects
#' @param x webr_directory object
#' @param ... Additional arguments (ignored)
#' @return Invisibly returns the object
#' @export
print.webr_directory <- function(x, ...) {
  cli::cli_h2("webR Directory Links")
  cli::cli_text("Source: {.path {x$source_directory}}")
  cli::cli_text("")

  cli::cli_text("Generated {length(x$urls)} link{?s}:")
  for (i in seq_along(x$urls)) {
    filename <- names(x$urls)[i]
    path <- paste0(x$base_path, filename)
    cli::cli_text("  {.file {filename}} \u2192 {.path {path}}")
    cli::cli_text("    {.url {x$urls[i]}}")
    if (i < length(x$urls)) cli::cli_text("")
  }

  formatted_mode <- format_mode_for_display(x$mode)
  if (!is.null(formatted_mode)) {
    cli::cli_text("")
    cli::cli_text("Interface: {.val {formatted_mode}}")
  }

  cli::cli_text("Version: {.val {x$version}}")

  invisible(x)
}



#' Print method for webr_decoded objects
#' @param x webr_decoded object
#' @param ... Additional arguments (ignored)
#' @return Invisibly returns the object
#' @export
print.webr_decoded <- function(x, ...) {
  cli::cli_h2("webR Decoded Files")
  cli::cli_text("Source: {.url {x$url}}")
  cli::cli_text("Output: {.path {x$output_dir}}")
  cli::cli_text("")

  if (nrow(x$files_info) == 0) {
    # Show detailed reasons why no files were decoded
    skip_reasons <- attr(x$files_info, "skip_reasons")

    if (!is.null(skip_reasons)) {
      reasons_found <- sum(lengths(skip_reasons)) > 0

      if (reasons_found) {
        cli::cli_text("No files were saved due to the following issues:")

        if (length(skip_reasons$invalid_structure) > 0) {
          cli::cli_text("  {cli::symbol$cross} {length(skip_reasons$invalid_structure)} file{?s} had invalid structure (missing name or path)")
        }

        if (length(skip_reasons$no_content) > 0) {
          files_text <- truncate_file_list(skip_reasons$no_content)
          cli::cli_text("  {cli::symbol$cross} {length(skip_reasons$no_content)} file{?s} had no content: {files_text}")
        }

        if (length(skip_reasons$already_exists) > 0) {
          files_text <- truncate_file_list(skip_reasons$already_exists)
          cli::cli_text("  {cli::symbol$cross} {length(skip_reasons$already_exists)} file{?s} already {?exists/exist}: {files_text}")
          cli::cli_text("    {cli::symbol$info} Use {.code overwrite = TRUE} to replace existing files")
        }

        if (length(skip_reasons$save_failed) > 0) {
          files_text <- truncate_file_list(skip_reasons$save_failed)
          cli::cli_text("  {cli::symbol$cross} {length(skip_reasons$save_failed)} file{?s} failed to save: {files_text}")
        }
      } else {
        cli::cli_text("No files were found in the decoded data.")
      }
    } else {
      cli::cli_text("No files were found in the decoded data.")
    }
  } else {
    cli::cli_text("Files ({nrow(x$files_info)}):")
    for (i in seq_len(nrow(x$files_info))) {
      file_info <- x$files_info[i, ]
      size_label <- format_file_size(file_info$size_bytes)
      autorun_label <- if (file_info$autorun) " {.emph (autorun)}" else ""
      cli::cli_text("  {.file {file_info$filename}} ({size_label}){autorun_label}")
    }
  }

  cli::cli_text("")

  # Always show metadata
  if (!is.null(x$mode)) {
    formatted_mode <- format_mode_for_display(x$mode)
    if (!is.null(formatted_mode)) {
      cli::cli_text("Interface: {.val {formatted_mode}}")
    }
  }

  cli::cli_text("Version: {.val {x$version}}")
  cli::cli_text("Encoding: {.val {x$flags}}")

  invisible(x)
}

#' Print method for webr_decoded_batch objects
#' @param x webr_decoded_batch object
#' @param ... Additional arguments (ignored)
#' @return Invisibly returns the object
#' @export
print.webr_decoded_batch <- function(x, ...) {
  cli::cli_h2("webR Decoded Batch")
  cli::cli_text("Base directory: {.path {x$base_dir}}")
  cli::cli_text("Total URLs: {length(x$urls)}")
  cli::cli_text("")

  successful_results <- x$results[!sapply(x$results, is.null)]

  if (length(successful_results) == 0) {
    cli::cli_text("No URLs were successfully processed.")
    return(invisible(x))
  }

  cli::cli_text("Successfully processed {length(successful_results)} URL{?s}:")

  total_files_saved <- 0

  for (i in seq_along(successful_results)) {
    result_name <- names(successful_results)[i]
    result <- successful_results[[i]]
    file_count <- nrow(result$files_info)
    total_files_saved <- total_files_saved + file_count

    if (file_count > 0) {
      cli::cli_text("  {.file {result_name}}: {file_count} file{?s}")
    } else {
      # Check for skip reasons
      skip_reasons <- attr(result$files_info, "skip_reasons")
      if (!is.null(skip_reasons) && sum(lengths(skip_reasons)) > 0) {
        cli::cli_text("  {.file {result_name}}: No files saved (see individual warnings)")
      } else {
        cli::cli_text("  {.file {result_name}}: No files found")
      }
    }

    cli::cli_text("    {.path {result$output_dir}}")
    if (i < length(successful_results)) cli::cli_text("")
  }

  failed_count <- length(x$urls) - length(successful_results)
  if (failed_count > 0) {
    cli::cli_text("")
    cli::cli_text("Failed: {failed_count} URL{?s}")
  }

  cli::cli_text("")
  cli::cli_text("Summary: {total_files_saved} total file{?s} saved across {length(successful_results)} URL{?s}")

  invisible(x)
}

#' Print method for webr_preview objects
#' @param x webr_preview object
#' @param show_content Logical. Whether to print the contents of each file (default: FALSE)
#' @param max_content_length Maximum number of characters of content to show per file (default: 500)
#' @param ... Additional arguments (ignored)
#' @return Invisibly returns the object
#' @export
print.webr_preview <- function(x, show_content = FALSE, max_content_length = 500, ...) {
  cli::cli_h2("webR Link Preview")
  cli::cli_text("URL: {.url {x$url}}")
  cli::cli_text("")

  total_size_label <- format_file_size(x$total_size)
  cli::cli_text("Files: {x$total_files}")
  cli::cli_text("Total size: {total_size_label}")

  if (length(x$autorun_files) > 0) {
    cli::cli_text("Autorun files: {length(x$autorun_files)}")
  }

  if (!is.null(x$mode)) {
    formatted_mode <- format_mode_for_display(x$mode)
    if (!is.null(formatted_mode)) {
      cli::cli_text("Interface: {.val {formatted_mode}}")
    }
  }

  cli::cli_text("Version: {.val {x$version}}")
  cli::cli_text("Encoding: {.val {x$flags}}")
  cli::cli_text("")

  # Show file details
  for (i in seq_along(x$files_data)) {
    file_info <- x$files_data[[i]]
    filename <- file_info$name
    autorun_indicator <- if (isTRUE(file_info$autorun)) " {.emph (autorun)}" else ""

    # Calculate file size and get content
    if ("text" %in% names(file_info)) {
      file_size <- nchar(file_info$text, type = "bytes")
      content <- file_info$text
      is_text <- TRUE
    } else if ("data" %in% names(file_info)) {
      content <- file_info$data
      if (is.character(content)) {
        file_size <- nchar(content, type = "bytes")
        is_text <- !detect_binary_content(content)
      } else {
        file_size <- length(content)
        content <- "[Binary data]"
        is_text <- FALSE
      }
    } else {
      file_size <- 0
      content <- "[No content]"
      is_text <- FALSE
    }

    size_label <- format_file_size(file_size)
    cli::cli_text("{.file {filename}} ({size_label}){autorun_indicator}")

    if (show_content && is_text && nchar(content) > 0) {
      # Show truncated content
      display_content <- if (nchar(content) <= max_content_length) {
        content
      } else {
        paste0(substr(content, 1, max_content_length), "...")
      }

      # Add indentation to all lines and show as verbatim
      content_lines <- strsplit(display_content, "\n")[[1]]
      indented_lines <- paste0("  ", content_lines)
      indented_content <- paste(indented_lines, collapse = "\n")

      cli::cli_verbatim(indented_content)
    }

    if (i < length(x$files_data)) cli::cli_text("")
  }

  if (!show_content && any(sapply(x$files_data, function(f) {
    ("text" %in% names(f)) || ("data" %in% names(f) && is.character(f$data))
  }))) {
    cli::cli_text("")
    cli::cli_text("{.emph Use print(preview, show_content = TRUE) to see file contents}")
  }

  invisible(x)
}

#' Print method for shinylive_preview objects
#' @param x shinylive_preview object
#' @param show_content Logical. Whether to print the contents of each file (default: FALSE)
#' @param max_content_length Maximum number of characters of content to show per file (default: 500)
#' @param ... Additional arguments (ignored)
#' @return Invisibly returns the object
#' @export
print.shinylive_preview <- function(x, show_content = FALSE, max_content_length = 500, ...) {
  engine_name <- if (x$engine == "python") "Python" else "R"
  cli::cli_h2("Shinylive {engine_name} Preview")
  cli::cli_text("Source: {.url {x$url}}")
  cli::cli_text("")

  total_size_label <- format_file_size(x$total_size)
  cli::cli_text("Files: {x$total_files}")
  cli::cli_text("Total size: {total_size_label}")
  cli::cli_text("Engine: {.val {engine_name}}")
  cli::cli_text("Mode: {.val {x$mode}}")
  cli::cli_text("")

  for (i in seq_along(x$files_data)) {
    file_info <- x$files_data[[i]]
    is_text <- !identical(file_info$type, "binary")
    content <- if (is.null(file_info$content)) "" else file_info$content
    file_size <- if (is_text) {
      nchar(content, type = "bytes")
    } else {
      calculate_base64_size(content)
    }

    size_label <- format_file_size(file_size)
    type_label <- if (is_text) "text" else "binary"
    cli::cli_text("{.file {file_info$name}} ({type_label}, {size_label})")

    if (show_content && is_text && nchar(content) > 0) {
      display_content <- if (nchar(content) <= max_content_length) {
        content
      } else {
        paste0(substr(content, 1, max_content_length), "...")
      }
      indented <- paste0("  ", strsplit(display_content, "\n")[[1]])
      cli::cli_verbatim(paste(indented, collapse = "\n"))
    }

    if (i < length(x$files_data)) cli::cli_text("")
  }

  if (!show_content && x$total_files > 0) {
    cli::cli_text("")
    cli::cli_text("{.emph Use print(preview, show_content = TRUE) to see file contents}")
  }

  invisible(x)
}

#' Print method for shinylive_decoded objects
#' @param x shinylive_decoded object
#' @param ... Additional arguments (ignored)
#' @return Invisibly returns the object
#' @export
print.shinylive_decoded <- function(x, ...) {
  engine_name <- if (x$engine == "python") "Python" else "R"
  cli::cli_h2("Shinylive {engine_name} Decoded Files")
  cli::cli_text("Source: {.url {x$url}}")
  cli::cli_text("Output: {.path {x$output_dir}}")
  cli::cli_text("")

  if (x$total_files == 0) {
    cli::cli_text("No files were saved.")
    return(invisible(x))
  }

  cli::cli_text("Files ({x$total_files}):")
  for (i in seq_len(nrow(x$files_info))) {
    file_info <- x$files_info[i, ]
    size_label <- format_file_size(file_info$size_bytes)
    cli::cli_text("  {.file {file_info$filename}} ({size_label})")
  }

  cli::cli_text("")
  cli::cli_text("Total size: {format_file_size(x$total_size)}")
  cli::cli_text("Mode: {.val {x$mode}}")

  invisible(x)
}

#' Print method for shinylive_decoded_batch objects
#' @param x shinylive_decoded_batch object
#' @param ... Additional arguments (ignored)
#' @return Invisibly returns the object
#' @export
print.shinylive_decoded_batch <- function(x, ...) {
  cli::cli_h2("Shinylive Decoded Batch")
  cli::cli_text("Base directory: {.path {x$base_dir}}")
  cli::cli_text("Total URLs: {x$total_urls}")
  cli::cli_text("")

  if (x$successful_urls == 0) {
    cli::cli_text("No URLs were successfully processed.")
    return(invisible(x))
  }

  cli::cli_text("Successfully processed {x$successful_urls} URL{?s}:")
  for (result in x$results) {
    if (is.null(result)) next
    cli::cli_text("  {.path {result$output_dir}} ({result$total_files} file{?s})")
  }

  cli::cli_text("")
  cli::cli_text("Total files: {x$total_files}")
  cli::cli_text("Total size: {format_file_size(x$total_size)}")

  invisible(x)
}

#' Print method for shinylive_link objects
#' @param x shinylive_link object
#' @param ... Additional arguments (ignored)
#' @return Invisibly returns the object
#' @export
print.shinylive_link <- function(x, ...) {
  engine_name <- if (x$engine == "python") "Python" else "R"
  cli::cli_h2("Shinylive {engine_name} App")
  cli::cli_text("{.url {x$url}}")
  cli::cli_text("")

  cli::cli_text("Files ({length(x$files)}):")
  for (file in x$files) {
    cli::cli_text("  {.file {file}}")
  }

  cli::cli_text("")
  cli::cli_text("Engine: {.val {engine_name}}")
  cli::cli_text("Mode: {.val {x$mode}}")

  invisible(x)
}

#' Print method for shinylive_project objects
#' @param x shinylive_project object
#' @param ... Additional arguments (ignored)
#' @return Invisibly returns the object
#' @export
print.shinylive_project <- function(x, ...) {
  engine_name <- if (x$engine == "python") "Python" else "R"
  cli::cli_h2("Shinylive {engine_name} Project")
  cli::cli_text("{.url {x$url}}")
  cli::cli_text("")

  cli::cli_text("Files ({length(x$files)}):")
  for (file in x$files) {
    cli::cli_text("  {.file {file}}")
  }

  cli::cli_text("")
  cli::cli_text("Engine: {.val {engine_name}}")
  cli::cli_text("Mode: {.val {x$mode}}")

  invisible(x)
}

#' Print method for shinylive_directory objects
#' @param x shinylive_directory object
#' @param ... Additional arguments (ignored)
#' @return Invisibly returns the object
#' @export
print.shinylive_directory <- function(x, ...) {
  engine_name <- if (x$engine == "python") "Python" else "R"
  cli::cli_h2("Shinylive {engine_name} Directory")
  cli::cli_text("Source: {.path {x$source_directory}}")
  cli::cli_text("")

  if (length(x$urls) == 0) {
    cli::cli_text("No apps found.")
    return(invisible(x))
  }

  cli::cli_text("Generated {length(x$urls)} app{?s}:")
  for (i in seq_along(x$urls)) {
    app_name <- names(x$urls)[i]
    cli::cli_text("  {.file {app_name}}")
    cli::cli_text("    {.url {x$urls[i]}}")
    if (i < length(x$urls)) cli::cli_text("")
  }

  cli::cli_text("")
  cli::cli_text("Engine: {.val {engine_name}}")
  cli::cli_text("Mode: {.val {x$mode}}")

  invisible(x)
}

---
title: "Getting Started with livelink"
vignette: >
  %\VignetteIndexEntry{Getting Started with livelink}
  %\VignetteEngine{quarto::html}
  %\VignetteEncoding{UTF-8}
format:
  html:
    embed-resources: true
knitr:
  opts_chunk:
    collapse: true
    comment: '#>'
---

```{=html}
<style>
.ll-fig{display:block;margin:1.5rem auto;max-width:100%;height:auto}
.ll-dark{display:none}
@media (prefers-color-scheme:dark){.ll-light{display:none}.ll-dark{display:block}}
[data-bs-theme=light] .ll-light{display:block}
[data-bs-theme=light] .ll-dark{display:none}
[data-bs-theme=dark] .ll-light{display:none}
[data-bs-theme=dark] .ll-dark{display:block}
</style>
```

```{r setup}
#| include: false
library(livelink)
```

## The idea

Sharing R code usually means asking someone to install something. livelink asks
nothing: it packs your code *into a URL*, and the recipient clicks it and is looking
at a running R session in their browser.

There is no server and no upload. The code is compressed and encoded into the part of
the URL after the `#`, which browsers never send anywhere. The link is the payload.

![Your code is serialized to JSON, gzipped, base64-encoded, and placed in the fragment of a webR URL after the `#`. Opening the link runs it in the browser.](../man/figures/pipeline-light.svg){.ll-fig .ll-light}
![](../man/figures/pipeline-dark.svg){.ll-fig .ll-dark aria-hidden="true"}

Two destinations are supported:

- **[webR](https://docs.r-wasm.org/webr/latest/)**: a full R console and editor in the browser.
- **[Shinylive](https://shiny.posit.co/py/get-started/shinylive.html)**: Shiny apps, R or Python, with no server.

## Your first link

Pass some code. Get a link.

```{r}
webr_repl_link("plot(1:10)")
```

In a rendered document like this one the object becomes a clickable link, so you can
open it right from the page. At an interactive console it prints its details instead,
and `as.character()` gives you the bare URL string either way.

You can skip the quotes entirely and hand it a braced expression, which is easier to
write and lets your editor highlight the code:

```{r}
webr_repl_link({
  data(mtcars)
  plot(mtcars$mpg, mtcars$wt)
})
```

::: {.callout-note}
## Comments and expression input

Expression input reconstructs your code from R's *source references*. Those exist in
an interactive session, but not when a document is knitted, so comments inside
`{ }` are dropped from the link when this vignette is rendered.

If your code has comments worth keeping, pass it as a string or a file path, or write
it as a chunk. See [Links inside documents](links-in-documents.html).
:::

You can also point at a file:

```{r}
script <- tempfile(fileext = ".R")
writeLines(c("# fit a model", "lm(mpg ~ wt, data = mtcars)"), script)

webr_repl_link(script)
```

...or, in an interactive session, copy code to your clipboard and call
`webr_repl_link()` with no arguments at all.

## Making it run on arrival

By default the recipient sees your code, ready to run. Set `autorun = TRUE` and it
runs the moment the page opens, which is right for a demo and wrong for an exercise.

```{r}
webr_repl_link("hist(rnorm(1000))", autorun = TRUE)
```

## Choosing what they see

webR opens with four panels: an **editor** for your code, a **terminal** for the R
console, a **files** browser, and a **plot** pane. `panels` names the ones you want,
and the rest never appear.

![webR has four panels. `panels` names the ones you want, and the rest never appear. Omit it and all four are shown.](../man/figures/panels-light.svg){.ll-fig .ll-light}
![](../man/figures/panels-dark.svg){.ll-fig .ll-dark aria-hidden="true"}

```{r}
webr_repl_link("plot(1:10)", panels = c("editor", "plot"))
```

The names are joined into the URL's query string, `?mode='editor-plot'`, so you can
read a link and tell what the recipient will land on. Omit `panels` and webR shows
everything:

```{r}
webr_repl_link("plot(1:10)")
```

Trimming the panels is worth doing when the terminal would only invite someone to
wander off. [Teaching with livelink](teaching.html) leans on it.

## A Shiny app instead

`shinylive_r_link()` does the same for a Shiny application.

```{r}
app <- "
library(shiny)
ui <- fluidPage(
  sliderInput('bins', 'Bins:', 1, 50, 30),
  plotOutput('hist')
)
server <- function(input, output) {
  output$hist <- renderPlot(hist(faithful$waiting, breaks = input$bins))
}
shinyApp(ui, server)
"

shinylive_r_link(app, mode = "app")
```

`mode = "app"` shows the running application. `mode = "editor"` shows the code beside
it, which is what you want when the point is to teach. Python apps work the same way
through `shinylive_py_link()`.

::: {.callout-tip}
## Why a string, and not a braced expression?

`shinylive_r_link({ ... })` does work. But a Shiny app is source text destined for
*another* runtime, and putting `library(shiny)` inside braces makes it look like R
code your own package runs: `R CMD check` scans vignettes for `library()` calls and
will report shiny as an undeclared dependency of livelink, which it is not.

Braces are for R you want executed in webR. For a Shiny app, a string or a file path
says what you mean.
:::

## What a link actually looks like

Nothing here is magic, and the URL is readable if you know where to look.

![A webR URL in five parts: the base address, the version, the panels to show, the `#code=` fragment carrying your files, and the flags saying how the payload was encoded.](../man/figures/url-anatomy-light.svg){.ll-fig .ll-light}
![](../man/figures/url-anatomy-dark.svg){.ll-fig .ll-dark aria-hidden="true"}

The flags at the end say how the payload was encoded: `m` for msgpack, `j` for JSON,
`z` for gzip, `u` for uncompressed, `a` for autorun. livelink writes `mz`, the same
format webR's own share button writes. It reads all of them, including the `jz` links
earlier versions produced.

## Going the other way

Because the files live in the URL, any link can be turned back into files, with no
network access and no cooperation from a server.

```{r}
url <- as.character(webr_repl_link("plot(1:10)"))

preview_webr_link(url)
```

`preview_webr_link()` inspects a link *without writing anything*, which is the
sensible way to open a link someone sent you. `decode_webr_link()` extracts the files
once you have decided you trust them.

## Links inside a document

If you write in Quarto or R Markdown, a chunk can carry its own link. Add
`livelink: true` to an ordinary chunk and it keeps running exactly as before, output
and plots and all, with a link added underneath:

````
```{{r}}
#| livelink: true
#| autorun: true
data(mtcars)
plot(mtcars$mpg, mtcars$wt)   # a comment that survives
```
````

Your comments reach the link, which they do not when a `{ }` expression is knitted.
[Links inside documents](links-in-documents.html) covers this properly: the hook, the
engine, every chunk option, and how to set them once for a whole document.

## Where to next

- **[Multi-file projects and Shiny apps](webr-and-shinylive.html)**: projects, apps in R and Python, and whole directories.
- **[Decoding and previewing links](decoding-links.html)**: turning a link back into files, and inspecting one before you trust it.
- **[Links inside documents](links-in-documents.html)**: the chunk hook and the chunk engine, and every option they take.
- **[Teaching with livelink](teaching.html)**: exercise and solution pairs, whole course directories, and embedding links in notes.

---
title: "Decoding and previewing links"
vignette: >
  %\VignetteIndexEntry{Decoding and previewing links}
  %\VignetteEngine{quarto::html}
  %\VignetteEncoding{UTF-8}
format:
  html:
    embed-resources: true
knitr:
  opts_chunk:
    collapse: true
    comment: '#>'
---

```{=html}
<style>
.ll-fig{display:block;margin:1.5rem auto;max-width:100%;height:auto}
.ll-dark{display:none}
@media (prefers-color-scheme:dark){.ll-light{display:none}.ll-dark{display:block}}
[data-bs-theme=light] .ll-light{display:block}
[data-bs-theme=light] .ll-dark{display:none}
[data-bs-theme=dark] .ll-light{display:none}
[data-bs-theme=dark] .ll-dark{display:block}
</style>
```

```{r setup}
#| include: false
library(livelink)
```

A livelink URL is not a pointer to a file on a server. The files *are* the URL --
compressed, encoded, and carried in the fragment after `#`. Nothing is uploaded
anywhere, which is what makes these links work offline and forever.

That has a useful consequence: any link can be turned back into the files it came
from, by anyone, with no network access.

![Encoding and decoding are inverse operations. The files you put into a link are the files you get back, byte for byte, with no server and no network.](../man/figures/roundtrip-light.svg){.ll-fig .ll-light}
![](../man/figures/roundtrip-dark.svg){.ll-fig .ll-dark aria-hidden="true"}

## Look before you leap

Someone sends you a link. Before you let it write anything to your disk, you can
inspect it. There are two ways to open a link, and they differ in exactly one way:
whether anything touches your filesystem.

![One link, two operations. `preview_webr_link()` decodes it in memory and reports what is inside, writing nothing. `decode_webr_link()` extracts the files to a directory.](../man/figures/preview-decode-light.svg){.ll-fig .ll-light}
![](../man/figures/preview-decode-dark.svg){.ll-fig .ll-dark aria-hidden="true"}

`preview_webr_link()` decodes the URL in memory and tells you what is inside,
without creating a single file.

```{r}
url <- as.character(webr_repl_link("hist(rnorm(100))", filename = "plot.R"))

preview <- preview_webr_link(url)
preview
```

The preview object carries the details, so you can query it rather than read it:

```{r}
preview$total_files
preview$files_data[[1]]$name
```

To see the actual code, ask for it:

```{r}
print(preview, show_content = TRUE)
```

This is the honest way to open a link from a stranger. You get to read the code
before you decide whether to run it.

## Extracting the files

Once you are satisfied, `decode_webr_link()` writes the files out.

```{r}
out <- file.path(tempdir(), "recovered")

result <- decode_webr_link(url, output_dir = out)
result
```

By default it writes into a subdirectory of your session's temporary directory, so a
stray call cannot litter your project. Pass `output_dir` when you want the files
somewhere permanent:

```{r}
#| eval: false
decode_webr_link(url, output_dir = "~/analysis", create_subdir = FALSE)
```

`create_subdir = FALSE` puts the files directly in `output_dir` instead of nesting
them under a hash-named folder.

Existing files are never clobbered silently. If a file is already there, it is
skipped and reported:

```{r}
decode_webr_link(url, output_dir = out)
```

Pass `overwrite = TRUE` when replacing them is what you actually want.

## The round trip is lossless

What goes in comes back out, byte for byte, including multi-file projects and
non-ASCII text.

```{r}
project <- list(
  "main.R"    = "source('utils.R')\nrun()",
  "utils.R"   = "run <- function() 42",
  "README.md" = "# Analysis\n\nWith an accent: café"
)

link <- webr_repl_project(project)

recovered_dir <- file.path(tempdir(), "roundtrip")
invisible(decode_webr_link(as.character(link), output_dir = recovered_dir))

files <- list.files(recovered_dir, recursive = TRUE, full.names = TRUE)
recovered <- setNames(
  lapply(files, function(f) paste(readLines(f, warn = FALSE), collapse = "\n")),
  basename(files)
)

identical(recovered[sort(names(recovered))], project[sort(names(project))])
```

## Shinylive links decode too

The same two functions exist for Shiny apps, and they work the same way.

```{r}
app_url <- as.character(shinylive_r_link("library(shiny)\nshinyApp(fluidPage(), function(i, o) {})"))

preview_shinylive_link(app_url)
```

```{r}
decode_shinylive_link(app_url, output_dir = file.path(tempdir(), "app"))
```

Python apps are no different. `preview_shinylive_link()` reports the engine, so you
can tell what you are looking at before you extract it:

```{r}
py_url <- as.character(shinylive_py_link("from shiny import App, ui"))

preview_shinylive_link(py_url)$engine
```

## Many links at once

Both decoders accept a vector of URLs and return a batch object. This is the useful
shape for grading a stack of student submissions, or auditing a set of links from a
course page.

```{r}
urls <- c(
  as.character(webr_repl_link("plot(1:10)", filename = "one.R")),
  as.character(webr_repl_link("hist(rnorm(50))", filename = "two.R"))
)

batch <- decode_webr_link(urls, output_dir = file.path(tempdir(), "batch"))
batch
```

Each URL gets its own subdirectory, so submissions cannot overwrite one another.

## Reading links livelink did not create

webR's own share dialog encodes files with **msgpack**, and so does livelink, so a
link from either writes the same format. Older livelink links used JSON. The flags at
the end of a URL say which is which: `mz` is msgpack + gzip, `jz` is JSON + gzip, and
`a` means autorun.

![The flags record how a link was encoded. livelink decodes msgpack and JSON links alike, whoever wrote them, into the same files.](../man/figures/dialects-light.svg){.ll-fig .ll-light}
![](../man/figures/dialects-dark.svg){.ll-fig .ll-dark aria-hidden="true"}

livelink reads all of them. A link produced by webR itself decodes exactly like one
produced here, and you do not have to know or care which tool made it.

## Getting just the URL

Every object livelink returns (links, projects, previews, decoded results) answers
to `repl_urls()`:

```{r}
repl_urls(preview)
repl_urls(batch)
```

which is handy when you want to write links out to a file, or drop them into a
document, without unpicking the object.

---
title: "Multi-file projects and Shiny apps"
vignette: >
  %\VignetteIndexEntry{Multi-file projects and Shiny apps}
  %\VignetteEngine{quarto::html}
  %\VignetteEncoding{UTF-8}
format:
  html:
    embed-resources: true
knitr:
  opts_chunk:
    collapse: true
    comment: '#>'
---

```{=html}
<style>
.ll-fig{display:block;margin:1.5rem auto;max-width:100%;height:auto}
.ll-dark{display:none}
@media (prefers-color-scheme:dark){.ll-light{display:none}.ll-dark{display:block}}
[data-bs-theme=light] .ll-light{display:block}
[data-bs-theme=light] .ll-dark{display:none}
[data-bs-theme=dark] .ll-light{display:none}
[data-bs-theme=dark] .ll-dark{display:block}
</style>
```

```{r setup}
#| include: false
library(livelink)
```

Real work is rarely one file. A script sources a helper, an app is split across
`ui.R` and `server.R`, an analysis needs its data. All of it fits in one link.

This article covers webR projects, Shiny apps in R and Python, and turning a whole
directory into links.

# webR projects

`webr_repl_project()` packs several files into a single link. webR unpacks them into
its filesystem, so `source("utils.R")` finds `utils.R` exactly where it expects it.

![A named list or a vector of file paths goes in, and one URL carrying every file comes out, which webR unpacks into its filesystem.](../man/figures/projects-light.svg){.ll-fig .ll-light}
![](../man/figures/projects-dark.svg){.ll-fig .ll-dark aria-hidden="true"}

## Describing the files

Two input shapes, and you can use whichever is closer to hand.

A **named list**, where the names become the filenames. Write each file as R, in
braces. No escaped newlines, no fighting with quotes, and your editor keeps
highlighting it:

```{r}
webr_repl_project(list(
  "main.R"    = { source("utils.R"); summarise(mtcars) },
  "utils.R"   = { summarise <- function(d) summary(d) },
  "README.md" = "# Analysis\n\nRun main.R to start."
))
```

Non-R files stay strings, and the two forms mix freely, since a project is usually code
*and* something else.

::: {.callout-important}
## Write the list inside the call

The braces only work when the list is written *in* the call. `list()` is an ordinary
function, so if you assign it to a variable first, R evaluates its arguments. The
block runs, right there, and livelink never sees the source:

```r
# WRONG: the block runs now, and actually tries to source("utils.R")
project <- list("main.R" = { source("utils.R"); summarise(mtcars) })
webr_repl_project(project)

# RIGHT: written in the call, the block is captured and never run
webr_repl_project(list("main.R" = { source("utils.R"); summarise(mtcars) }))
```

livelink notices the first case and stops with an error rather than encoding whatever
the block happened to return. A list of *strings* in a variable is fine, of course.
That is the older form, and it still works.
:::

Because the blocks are captured rather than evaluated, an assignment inside one leaves
nothing behind in your session. `{ summarise <- function(d) summary(d) }` defines
nothing here. It is source to ship, not code to run.

Two caveats carry over from single-file expressions. Comments inside `{ }` survive
interactively but not in a knitted document, and a `library()` call inside a block is
visible to `R CMD check`, which will report that package as an undeclared dependency
of *yours*. Use a string for code that loads packages.

When a project has to be reused across several calls, as it is through the rest of
this section, keep it as a list of **strings** in a variable. A list of strings is a
plain value, so it assigns without surprises:

```{r}
project <- list(
  "main.R"    = "source('utils.R')\nsummarise(mtcars)",
  "utils.R"   = "summarise <- function(d) summary(d)",
  "README.md" = "# Analysis\n\nRun main.R to start."
)

webr_repl_project(project)
```

A **vector of file paths**, when the code is already on disk:

```{r}
dir <- file.path(tempdir(), "project")
dir.create(dir, showWarnings = FALSE)
writeLines("source('utils.R')",  file.path(dir, "main.R"))
writeLines("f <- function() 42", file.path(dir, "utils.R"))

webr_repl_project(c(file.path(dir, "main.R"), file.path(dir, "utils.R")))
```

Files are read and named by their basename, so the paths on your machine never leave
it.

## Running something on arrival

`autorun_files` names the files to execute when the link opens. The link then carries
webR's `a` flag, which is what actually triggers the run:

```{r}
link <- webr_repl_project(project, autorun_files = "main.R")

preview_webr_link(as.character(link))$autorun_files
```

Pass `"all"` to run every R file in the project:

```{r}
webr_repl_project(project, autorun_files = "all")
```

Naming a file that is not in the project is an error, rather than a link that quietly
does nothing:

```{r}
#| error: true
webr_repl_project(project, autorun_files = "nonexistent.R")
```

Only `.R` files can autorun. Listing a `.csv` is harmless, and simply is not run.

## Where the files land

By default everything is placed in `/home/web_user/`, which is webR's working
directory, so relative paths in your code just work. `base_path` moves them:

```{r}
webr_repl_project(project, base_path = "/home/web_user/analysis/")
```

Move them and relative `source()` calls have to move with them, so leave this alone
unless you have a reason.

# Shiny apps

![`shinylive_r_link()` and `shinylive_py_link()` pick the R or Python path, and `mode` picks whether a visitor lands on the code beside the app, or the running app alone.](../man/figures/shinylive-light.svg){.ll-fig .ll-light}
![](../man/figures/shinylive-dark.svg){.ll-fig .ll-dark aria-hidden="true"}

Write the app as a braced expression, exactly as you would write it anywhere. No
quotes around the whole thing, no escaped newlines, and your editor keeps highlighting
the code:

```{r}
#| eval: false
shinylive_r_link({
  library(shiny)

  ui <- fluidPage(
    titlePanel("Old Faithful"),
    sliderInput("bins", "Bins:", 1, 50, 30),
    plotOutput("hist")
  )

  server <- function(input, output) {
    output$hist <- renderPlot({
      hist(faithful$waiting, breaks = input$bins, col = "steelblue")
    })
  }

  shinyApp(ui, server)
})
```

That is real, working code. It is marked `eval: false` here for one narrow reason,
which does not apply to your own documents.

::: {.callout-note}
## Why that chunk is not run

Only to keep this package's vignette honest about its dependencies. `R CMD check`
reads the live code in a vignette and, finding `library(shiny)` in a braced
expression, would decide livelink depends on shiny. It does not. Marking the chunk
`eval: false` keeps that call out of the scanned code.

Your own course notes are not a package under check, so there is no such constraint.
Write the app as an expression and let it run.
:::

When you want the link built and inspected here, a string sidesteps the scan, so this
vignette can execute it. A bare string becomes `app.R`:

```{r}
app <- "library(shiny)\nshinyApp(fluidPage(), function(input, output) {})"

shinylive_r_link(app)
```

For an app split across files, the same choice applies. Write each file in braces:

```{r}
#| eval: false
shinylive_r_link(list(
  "app.R"    = { library(shiny); source("ui.R"); source("server.R"); shinyApp(ui, server) },
  "ui.R"     = { ui <- fluidPage(titlePanel("Split app")) },
  "server.R" = { server <- function(input, output) {} }
))
```

## Editor or app?

`mode` decides what a visitor lands on.

```{r}
# code and app side by side -- for teaching and review
shinylive_r_link(app, mode = "editor")
```

```{r}
# the running application only -- for the people who just want to use it
shinylive_r_link(app, mode = "app")
```

In `app` mode you can also drop the Shinylive header, which is what you want when the
app is going into an `<iframe>` on someone else's page:

```{r}
shinylive_r_link(app, mode = "app", header = FALSE)
```

`header` does nothing in `editor` mode, where there is no header to hide.

::: {.callout-note}
## `mode` is not `panels`

Shinylive has two display *modes*, `"editor"` and `"app"`. webR has four *panels* you
can show or hide. They are different arguments because they are different ideas: one
picks a layout, the other picks which parts of a layout appear.
:::

## Python

Python apps work the same way, and a bare string becomes `app.py`:

```{r}
py_app <- "
from shiny import App, render, ui

app_ui = ui.page_fluid(
    ui.h2('Hello from Python'),
    ui.output_text('greeting'),
)

def server(input, output, session):
    @render.text
    def greeting():
        return 'Running in the browser, no server required.'

app = App(app_ui, server)
"

shinylive_py_link(py_app, mode = "app")
```

The engine is in the URL, `shinylive.io/r/` versus `shinylive.io/py/`, so a link
carries its own language with it.

`shinylive_project()` is the same thing with the engine named explicitly, which is
useful when the engine is a variable rather than a decision you already made:

```{r}
shinylive_project(
  list("app.py" = py_app, "utils.py" = "def helper():\n    return 42"),
  engine = "python"
)
```

# A directory at a time

Each Shiny app in its own folder is the usual layout, and `shinylive_directory()`
turns the lot into links in one pass:

```{r}
apps <- file.path(tempdir(), "apps")
dir.create(file.path(apps, "histogram"), recursive = TRUE, showWarnings = FALSE)
dir.create(file.path(apps, "scatter"),   recursive = TRUE, showWarnings = FALSE)
writeLines(app, file.path(apps, "histogram", "app.R"))
writeLines(app, file.path(apps, "scatter",   "app.R"))

links <- shinylive_directory(apps, engine = "r", mode = "app")

repl_urls(links)
```

Every subdirectory holding an `app.R` (or `app.py`, for Python) becomes one link,
named after the folder. Helpers, data and CSS in that folder are bundled in with it.
Binary files are skipped with a warning, because a link is the wrong place for them.

A directory with no apps in it warns and hands back an empty result, rather than
failing:

```{r}
empty <- file.path(tempdir(), "no-apps")
dir.create(empty, showWarnings = FALSE)

shinylive_directory(empty, engine = "r")
```

`webr_repl_directory()` is the webR counterpart, and by default it works the other way
round: it gives each R script in a directory *its own* link, rather than bundling a
folder into one. That suits a folder of course scripts, and
[Teaching with livelink](teaching.html) puts it to work.

When you do want the whole directory in a single link, pass `single_link = TRUE`. The
files are bundled exactly as `webr_repl_project()` would bundle them, and you get one
`webr_project` back instead of a directory of links:

```{r}
scripts <- file.path(tempdir(), "scripts")
dir.create(scripts, showWarnings = FALSE)
writeLines("source('utils.R')", file.path(scripts, "main.R"))
writeLines("f <- function() 42", file.path(scripts, "utils.R"))

webr_repl_directory(scripts, single_link = TRUE, panels = c("editor", "plot"))
```

With `autorun = TRUE` alongside it, every R file in the bundle runs on arrival.

# How big can a link get?

The files travel *inside* the URL, so the ceiling is whatever a browser accepts in a
URL. Current browsers handle tens of thousands of characters, which is a lot of R.

The payload is gzipped before it is encoded, and source code compresses well. Real R
source, rather than a toy example:

```{r}
src <- paste(
  unlist(lapply(c("lm", "glm", "aov"), function(f) {
    paste(deparse(get(f, envir = asNamespace("stats"))), collapse = "\n")
  })),
  collapse = "\n\n"
)

url <- as.character(webr_repl_link(src, filename = "stats.R"))

c(source_chars = nchar(src), url_chars = nchar(url))
```

The link comes out *smaller than the code it carries*, because gzip more than pays for
what base64 costs.

Binary data is the case that does not work: it does not compress, and base64 inflates
it by a third. Do not put a PNG in a link. Have the code fetch it, or point at it.

---
title: "Teaching with livelink"
vignette: >
  %\VignetteIndexEntry{Teaching with livelink}
  %\VignetteEngine{quarto::html}
  %\VignetteEncoding{UTF-8}
format:
  html:
    embed-resources: true
knitr:
  opts_chunk:
    collapse: true
    comment: '#>'
---

```{=html}
<style>
.ll-fig{display:block;margin:1.5rem auto;max-width:100%;height:auto}
.ll-dark{display:none}
@media (prefers-color-scheme:dark){.ll-light{display:none}.ll-dark{display:block}}
[data-bs-theme=light] .ll-light{display:block}
[data-bs-theme=light] .ll-dark{display:none}
[data-bs-theme=dark] .ll-light{display:none}
[data-bs-theme=dark] .ll-dark{display:block}
</style>
```

```{r setup}
#| include: false
library(livelink)
```

The hardest part of teaching R is not R. It is the twenty minutes at the start of a
workshop spent installing R, then RStudio, then a package that fails to compile on
one student's laptop.

livelink sidesteps that. A link opens a working R session in the browser, with your
code already in the editor. Nothing to install, nothing to configure, and it works on
a school Chromebook.

## Exercise and solution pairs

`webr_repl_exercise()` is built for the most common teaching shape: give students a
scaffold with the answer removed, and keep the worked solution behind a second link.

![One call returns two links. They differ by a single flag: the solution carries `a` for autorun and computes the answer on arrival, while the exercise waits.](../man/figures/exercise-light.svg){.ll-fig .ll-light}
![](../man/figures/exercise-dark.svg){.ll-fig .ll-dark aria-hidden="true"}

```{r}
exercise <- "
# Exercise: what is the mean mpg of the cars in mtcars?
# Replace the 0 below with your answer.

mean_mpg <- 0
print(mean_mpg)
"

solution <- "
# Solution
mean_mpg <- mean(mtcars$mpg)
print(mean_mpg)
"

links <- webr_repl_exercise(exercise, solution, "mean_mpg")
links
```

These are written as plain strings on purpose. An exercise is mostly its instructions,
and those live in comments, so a braced `{ }` expression would drop them the moment
these notes are knitted. [Links inside
documents](links-in-documents.html#why-not-a-braced-expression) explains why.

The two links differ in one deliberate way: the solution **autoruns** and the exercise
does not. Students landing on the exercise see the code sitting there waiting for
them. Students landing on the solution see the answer already computed.

```{r}
links$exercise$autorun
links$solution$autorun
```

Hand out the exercise link in class, and hold the solution link back until afterwards, or
put it behind a collapsed callout in your notes.

```{r}
repl_urls(links)
```

## Choosing what students see

A full webR session shows an editor, a plot pane, a file browser, and a terminal. For
a first-year class that is a lot of surface area, and the terminal in particular
invites students to wander off.

`panels` controls which of them appear.

![webR has four panels. `panels` names the ones you want, and the rest never appear.](../man/figures/panels-light.svg){.ll-fig .ll-light}
![](../man/figures/panels-dark.svg){.ll-fig .ll-dark aria-hidden="true"}

```{r}
# A focused exercise: just write code and see the plot
webr_repl_link("plot(1:10)", panels = c("editor", "plot"))
```

```{r}
#| eval: false
# The full environment, for a project week
webr_repl_link(code, panels = c("editor", "plot", "files", "terminal"))
```

For a lecture demo, where students should see the *result* and not the machinery,
combine `autorun` with a plot-only view:

```{r}
webr_repl_link(
  "hist(rnorm(1000), col = 'steelblue', main = 'The central limit theorem, again')",
  autorun = TRUE,
  panels  = "plot"
)
```

Only `.R` files can autorun. Ask for it on something else and livelink tells you
rather than quietly producing a link that does nothing:

```{r}
#| error: true
webr_repl_link("some data", filename = "data.csv", autorun = TRUE)
```

## A whole folder of scripts

Course material tends to live as a directory of scripts, and livelink gives you two
ways to turn that folder into links. Which you want comes down to a single question:
are the scripts **independent**, or **parts of one thing**?

`webr_repl_directory()` gives each script *its own* link. That is the right call for a
folder of standalone exercises: one link per file, each opening just that script.

![Each script becomes its own link, named after the file. That is the mirror image of `webr_repl_project()`, which bundles a whole folder into one link.](../man/figures/course-light.svg){.ll-fig .ll-light}
![](../man/figures/course-dark.svg){.ll-fig .ll-dark aria-hidden="true"}

```{r}
# A small course directory
course <- file.path(tempdir(), "course")
dir.create(course, showWarnings = FALSE)
writeLines("plot(1:10)",           file.path(course, "01-plotting.R"))
writeLines("mean(mtcars$mpg)",     file.path(course, "02-summaries.R"))
writeLines("lm(mpg ~ wt, mtcars)", file.path(course, "03-models.R"))

links <- webr_repl_directory(course, panels = c("editor", "plot"))
links
```

The result is named by file, so it drops straight into a table of contents:

```{r}
urls <- repl_urls(links)
cat(sprintf("- [%s](%s)", names(urls), urls), sep = "\n")
```

Use `pattern` to pick out a subset, say the exercises but not the solutions:

```{r}
webr_repl_directory(course, pattern = "^0[12]")
```

### One link per file, or one link for the folder?

`webr_repl_project()` is the other half of the pair. It packs several files into a
*single* link, so a script that `source()`s its neighbour still finds it once webR
unpacks them side by side. Reach for it when the files belong together rather than
standing alone.

The two are mirror images, and the table says which to reach for:

| You want | Call |
|---|---|
| each script openable on its own | `webr_repl_directory(dir)` |
| the whole folder bundled into one link | `webr_repl_directory(dir, single_link = TRUE)` |
| a hand-picked set of files bundled into one link | `webr_repl_project(files)` |

`single_link = TRUE` is the bridge between them: point it at a folder and it bundles
the lot into one link, exactly as `webr_repl_project()` would, without your having to
list the files by hand.

```{r}
# the same course folder, as one bundled link instead of three
webr_repl_directory(course, single_link = TRUE, panels = c("editor", "plot"))
```

## Embedding links in your notes

If you write course notes in Quarto or R Markdown, you do not have to build the link
in one chunk and paste the URL into another. livelink plugs into knitr, so a chunk can
carry its own link.

![A `livelink` chunk renders as the code plus a link that runs it, comments and all. A braced expression cannot manage that in a knitted document.](../man/figures/engine-light.svg){.ll-fig .ll-light}
![](../man/figures/engine-dark.svg){.ll-fig .ll-dark aria-hidden="true"}

Add `livelink: true` to an ordinary `r` chunk. It runs as it always did, the plot
appears in your notes, and a link is added underneath. Here it is, running in this
very vignette:

```{r}
#| livelink: true
#| autorun: true
#| panels: ["editor", "plot"]
# Load the data
data(mtcars)
plot(mtcars$mpg, mtcars$wt)   # a comment students will actually see
```

Students read the result *and* get to go and poke at it. Follow the link: the comments
are still there, which they would not be had you written `webr_repl_link({ ... })`.

For a folder of course notes, set it once and let every chunk carry a link:

```{r}
#| eval: false
knitr::opts_chunk$set(livelink = TRUE, autorun = TRUE)
```

[Links inside documents](links-in-documents.html) covers the rest: the engine, for code
your notes should not run, every chunk option, and why `echo` is not what stands between
you and a link.

## Collecting work back in

Students send you links. `decode_webr_link()` accepts a vector of them and writes each
into its own directory, so a stack of submissions becomes a folder you can grade.

```{r}
submissions <- c(
  as.character(webr_repl_link("mean(mtcars$mpg)",   filename = "student_a.R")),
  as.character(webr_repl_link("median(mtcars$mpg)", filename = "student_b.R"))
)

decode_webr_link(submissions, output_dir = file.path(tempdir(), "submissions"))
```

If you would rather read the code than write it to disk, `preview_webr_link()` shows
you a submission without touching your filesystem at all, which is the safer default
when the code came from someone else.

# print.webr_preview honors show_content

    Code
      print(preview)
    Message
      
      -- webR Link Preview --
      
      URL:
      <https://webr.r-wasm.org/latest/#code=eJyLrlbKS8xNVbJSKk4uyiwo0QtS0lEqSCzJAIroZ%2BTnpuqXpybFlxanFukjKShJrSgBaUlNLkotic9NLMpOLVKw0VUwMVKqjQUAjrwblA%3D%3D&jz>
      
      Files: 1
      Total size: 19 bytes
      Version: "latest"
      Encoding: "jz"
      
      'script.R' (19 bytes)
      
      Use print(preview, show_content = TRUE) to see file contents

---

    Code
      print(preview, show_content = TRUE)
    Message
      
      -- webR Link Preview --
      
      URL:
      <https://webr.r-wasm.org/latest/#code=eJyLrlbKS8xNVbJSKk4uyiwo0QtS0lEqSCzJAIroZ%2BTnpuqXpybFlxanFukjKShJrSgBaUlNLkotic9NLMpOLVKw0VUwMVKqjQUAjrwblA%3D%3D&jz>
      
      Files: 1
      Total size: 19 bytes
      Version: "latest"
      Encoding: "jz"
      
      'script.R' (19 bytes)
        secret_marker <- 42

# print.shinylive_preview honors show_content

    Code
      print(preview)
    Message
      
      -- Shinylive R Preview --
      
      Source:
      <https://shinylive.io/r/editor/#code=NobwRAdghgtgpmAXGKAHVA6ASmANGAYwHsIAXOMpMAZzgICc5SB9GKegaznoAIAeALQ8ALACY8YUgE9UCZOQAepMAF8AukA>
      
      Files: 1
      Total size: 19 bytes
      Engine: "R"
      Mode: "editor"
      
      'app.R' (text, 19 bytes)
      
      Use print(preview, show_content = TRUE) to see file contents

---

    Code
      print(preview, show_content = TRUE)
    Message
      
      -- Shinylive R Preview --
      
      Source:
      <https://shinylive.io/r/editor/#code=NobwRAdghgtgpmAXGKAHVA6ASmANGAYwHsIAXOMpMAZzgICc5SB9GKegaznoAIAeALQ8ALACY8YUgE9UCZOQAepMAF8AukA>
      
      Files: 1
      Total size: 19 bytes
      Engine: "R"
      Mode: "editor"
      
      'app.R' (text, 19 bytes)
        secret_marker <- 42

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extract-link.R
\name{repl_urls}
\alias{repl_urls}
\alias{repl_urls.default}
\title{Extract shareable URLs from livelink objects}
\usage{
repl_urls(x, ...)

\method{repl_urls}{default}(x, ...)
}
\arguments{
\item{x}{A livelink object. Supported classes are \code{webr_link}, \code{webr_project},
\code{webr_exercise}, \code{webr_directory}, \code{webr_decoded}, \code{webr_decoded_batch},
\code{webr_preview}, \code{shinylive_link}, \code{shinylive_project}, \code{shinylive_directory},
\code{shinylive_decoded}, \code{shinylive_decoded_batch}, and \code{shinylive_preview}.}

\item{...}{Additional arguments passed to methods}
}
\value{
A character vector of URLs. Most objects yield a single URL; exercise
objects return a length-2 named vector (\code{exercise}, \code{solution}); directory
and batch objects return one URL per file.
}
\description{
Generic function to extract the shareable URL(s) from any livelink object,
covering both webR REPL and Shinylive results. Provides a clear way to get
just the URLs for sharing or further processing.
}
\examples{
# Single link
link <- webr_repl_link("plot(1:10)")
repl_urls(link)

# Exercise (returns named vector)
exercise <- webr_repl_exercise("# TODO", "plot(1:10)", "test")
repl_urls(exercise)

# Shinylive links work the same way
repl_urls(shinylive_r_link("library(shiny)"))

# Decoded files (returns the original URL)
decoded <- decode_webr_link(as.character(link))
repl_urls(decoded)

}
\seealso{
The \code{as.character()} methods (for example \code{\link[=as.character.webr_link]{as.character.webr_link()}}),
which \code{repl_urls()} delegates to.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/webr-repl-links.R
\name{webr_repl_project}
\alias{webr_repl_project}
\title{Create webR REPL sharelink for multiple files}
\usage{
webr_repl_project(
  input,
  autorun_files = character(0),
  base_path = "/home/web_user/",
  panels = NULL,
  version = "latest",
  base_url = NULL
)
}
\arguments{
\item{input}{Input for multiple files. Can be:
\itemize{
\item Named list of braced expressions, so each file is written as R rather than
as a string full of escaped newlines:
\code{list("main.R" = { plot(1:10) }, "utils.R" = { f <- function() 42 })}
\item Named list of strings: \code{list("main.R" = code1, "utils.R" = code2)}
\item Vector of file paths: \code{c("main.R", "utils.R", "data.csv")}
}

The two list forms mix freely, which is what you want for a project holding
both code and, say, a \code{README.md}.}

\item{autorun_files}{Character vector of filenames to auto-execute when project loads, or "all" to autorun all R files (default: none)}

\item{base_path}{Base directory path for all files (default: \code{"/home/web_user/"})}

\item{panels}{Character vector or string specifying which webR interface panels to show.
Valid panels: \code{"plot"}, \code{"files"}, \code{"terminal"}, \code{"editor"}. Can be \code{c("plot", "files")} or \code{"plot-files"}.
If NULL (default), shows all panels.}

\item{version}{webR version to use (\code{"latest"} or specific version >= "v0.5.4")}

\item{base_url}{webR application URL. If NULL, uses global option or builds from version}
}
\value{
webr_project object containing the webR sharelink for the multi-file project
}
\description{
Creates a webR sharelink for projects with multiple R files, data files, or other resources.
Supports named lists and file path vectors as input.
}
\section{Writing a project as code}{

A file's contents can be given as a \code{{ ... }} block instead of a string, which
spares you escaping every newline and quote:

\if{html}{\out{<div class="sourceCode">}}\preformatted{webr_repl_project(list(
  "main.R"    = \{ source("utils.R"); summarise(mtcars) \},
  "utils.R"   = \{ summarise <- function(d) summary(d) \},
  "README.md" = "# Analysis"
))
}\if{html}{\out{</div>}}

The blocks are \strong{never evaluated} -- they are source to ship, not code to run
-- so an assignment inside one leaves nothing behind in your session.

Two things to know. Comments inside \code{{ }} survive in an interactive session but
not in a knitted document (see \code{\link[=webr_repl_link]{webr_repl_link()}}). And a \code{library()} call
inside a block is visible to \verb{R CMD check}, which will report the package as an
undeclared dependency of \emph{yours}; in a vignette or an example, use a string for
code that loads packages.
}

\examples{
# Each file written as R, rather than as an escaped string
webr_repl_project(list(
  "main.R"  = { source("utils.R"); summarise(mtcars) },
  "utils.R" = { summarise <- function(d) summary(d) }
))

# Strings still work, and the two forms mix
files <- list(
  "main.R" = "source('utils.R')\nresult <- analyze_data(mtcars)",
  "utils.R" = "analyze_data <- function(data) { summary(data) }",
  "README.md" = "# My Analysis\nThis project analyzes the mtcars dataset."
)
webr_repl_project(files, autorun_files = "main.R")

# Autorun every R file in the project
webr_repl_project(files, autorun_files = "all")

# File paths input
project_dir <- tempfile()
dir.create(project_dir)
main <- file.path(project_dir, "main.R")
utils <- file.path(project_dir, "utils.R")
writeLines("source('utils.R')", main)
writeLines("# utils", utils)
webr_repl_project(c(main, utils))
}
\seealso{
\code{\link[=webr_repl_link]{webr_repl_link()}} for the single-file case.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shinylive-link.R
\name{shinylive_project}
\alias{shinylive_project}
\title{Create a Shinylive sharelink for multi-file projects}
\usage{
shinylive_project(
  input,
  engine,
  mode = "editor",
  header = TRUE,
  base_url = NULL
)
}
\arguments{
\item{input}{Input for multiple files. Can be:
\itemize{
\item Named list: \code{list("app.R" = code1, "utils.R" = code2)}
\item Vector of file paths: \code{c("app.R", "utils.R", "data.csv")}
}}

\item{engine}{Engine to use: "r" for R Shiny or "python" for Python Shiny}

\item{mode}{Shinylive display mode (default \code{"editor"}). \code{"editor"} shows an
editable code panel beside the running app; \code{"app"} shows only the running app.}

\item{header}{Logical, whether to show the Shinylive header bar. It applies only
when \code{mode = "app"} and is ignored in the default \code{"editor"} mode. Defaults to \code{TRUE}.}

\item{base_url}{Custom Shinylive base URL. If NULL (default), links point at https://shinylive.io.}
}
\value{
shinylive_project object containing the Shinylive URL and metadata
}
\description{
Creates Shinylive projects for either R or Python from named lists or
file path vectors.
}
\examples{
# Named list input
files <- list(
  "app.R" = "library(shiny)\nshinyApp(fluidPage(), function(i, o) {})",
  "utils.R" = "# Utility functions"
)
shinylive_project(files, engine = "r", mode = "editor")

# File paths input
project_dir <- tempfile()
dir.create(project_dir)
app <- file.path(project_dir, "app.R")
utils <- file.path(project_dir, "utils.R")
writeLines("library(shiny)", app)
writeLines("# utils", utils)
shinylive_project(c(app, utils), engine = "r")
}
\seealso{
\code{\link[=shinylive_r_link]{shinylive_r_link()}} and \code{\link[=shinylive_py_link]{shinylive_py_link()}} for single apps;
\code{\link[=decode_shinylive_link]{decode_shinylive_link()}} and \code{\link[=preview_shinylive_link]{preview_shinylive_link()}} to read a link back.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/webr-repl-links.R
\name{webr_repl_exercise}
\alias{webr_repl_exercise}
\title{Create paired exercise and solution webR REPL links}
\usage{
webr_repl_exercise(
  exercise_text,
  solution_text,
  exercise_name,
  base_path = "/home/web_user/",
  version = "latest",
  base_url = NULL
)
}
\arguments{
\item{exercise_text}{Character string containing the exercise code with placeholders or TODOs}

\item{solution_text}{Character string containing the complete solution code}

\item{exercise_name}{Base name for the exercise (will create \code{"name_exercise.R"} and \code{"name_solution.R"})}

\item{base_path}{Base directory path for files (default: \code{"/home/web_user/"})}

\item{version}{webR version to use ("latest" or specific version >= "v0.5.4")}

\item{base_url}{webR application URL. If NULL, uses global option or builds from version}
}
\value{
webr_exercise object holding the paired \code{exercise} and \code{solution} links
}
\description{
Generates a pair of webR links for educational purposes: one for student exercises
(without autorun) and one for solutions (with autorun enabled).
}
\examples{
exercise_code <- "
# Exercise: Calculate mean of mtcars$mpg
# TODO: Complete the line below
mean_mpg <- 0
print(mean_mpg)
"

solution_code <- "
# Solution: Calculate mean of mtcars$mpg
mean_mpg <- mean(mtcars$mpg)
print(mean_mpg)
"

links <- webr_repl_exercise(exercise_code, solution_code, "basic_stats")
links$exercise
links$solution

# Custom path and version
webr_repl_exercise(exercise_code, solution_code, "stats",
                   base_path = "/exercises/", version = "v0.5.4")

}
\seealso{
\code{\link[=webr_repl_link]{webr_repl_link()}}, which this builds on;
\code{vignette("teaching", package = "livelink")} for using links in a course.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/decode-webr-repl-link.R
\name{decode_webr_link}
\alias{decode_webr_link}
\title{Decode webR REPL link(s) to extract files to local directory}
\usage{
decode_webr_link(
  url,
  output_dir = file.path(tempdir(), "webr_files"),
  overwrite = FALSE,
  create_subdir = TRUE,
  name_dirs = TRUE
)
}
\arguments{
\item{url}{Character string or vector containing webR REPL URL(s)}

\item{output_dir}{Character string specifying the output directory path.
Defaults to a \code{webr_files} directory inside the session temporary
directory; pass an explicit path to extract somewhere permanent.}

\item{overwrite}{Logical. Whether to overwrite existing files (default: FALSE)}

\item{create_subdir}{Logical. If \code{TRUE} (default), each decoded link is
extracted into its own subdirectory under \code{output_dir} rather than directly
into it. For a single URL the subdirectory is named \verb{webr_<hash>}, where
\verb{<hash>} is a short fingerprint of the URL. For multiple URLs, see
\code{name_dirs}. Set \code{FALSE} to extract straight into \code{output_dir}.}

\item{name_dirs}{Logical. For multiple URLs, controls how the per-link
subdirectories are named: \code{TRUE} (default) numbers them \code{script_01},
\code{script_02}, ...; \code{FALSE} names each one \verb{webr_<hash>} from the URL
fingerprint. Ignored for a single URL, and ignored when
\code{create_subdir = FALSE} (all files then extract into \code{output_dir}).}
}
\value{
For a single URL, a \code{webr_decoded} object. For multiple URLs, a
\code{webr_decoded_batch} object.
}
\description{
Decodes webR REPL sharelinks to extract the embedded files and save them
to a local directory. This is the reverse operation of creating webR links.
Handles both single URLs and multiple URLs automatically.
}
\examples{
# Round-trip: build a link, then decode it back to files
url <- as.character(webr_repl_link("plot(1:10)"))

result <- decode_webr_link(url)
print(result)

# Extract to a directory of your choosing
out <- file.path(tempdir(), "my_code")
decode_webr_link(url, output_dir = out, create_subdir = FALSE, overwrite = TRUE)
list.files(out)

# Several links at once
urls <- c(url, as.character(webr_repl_link("hist(rnorm(100))")))
decode_webr_link(urls, output_dir = file.path(tempdir(), "my_scripts"))
}
\seealso{
\code{\link[=preview_webr_link]{preview_webr_link()}} to inspect a link without writing files, and
\code{\link[=webr_repl_link]{webr_repl_link()}}, \code{\link[=webr_repl_project]{webr_repl_project()}}, and \code{\link[=webr_repl_exercise]{webr_repl_exercise()}}, which
create the links this function decodes.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/format-methods.R
\name{format.livelink}
\alias{format.livelink}
\alias{format.webr_link}
\alias{format.webr_project}
\alias{format.webr_exercise}
\alias{format.webr_directory}
\alias{format.webr_decoded}
\alias{format.webr_decoded_batch}
\alias{format.webr_preview}
\alias{format.shinylive_link}
\alias{format.shinylive_project}
\alias{format.shinylive_directory}
\alias{format.shinylive_decoded}
\alias{format.shinylive_decoded_batch}
\alias{format.shinylive_preview}
\title{Format a livelink object as a character vector}
\usage{
\method{format}{webr_link}(x, ...)

\method{format}{webr_project}(x, ...)

\method{format}{webr_exercise}(x, ...)

\method{format}{webr_directory}(x, ...)

\method{format}{webr_decoded}(x, ...)

\method{format}{webr_decoded_batch}(x, ...)

\method{format}{webr_preview}(x, ...)

\method{format}{shinylive_link}(x, ...)

\method{format}{shinylive_project}(x, ...)

\method{format}{shinylive_directory}(x, ...)

\method{format}{shinylive_decoded}(x, ...)

\method{format}{shinylive_decoded_batch}(x, ...)

\method{format}{shinylive_preview}(x, ...)
}
\arguments{
\item{x}{A livelink object (a link, project, exercise, directory, decoded
result, batch, or preview).}

\item{...}{Passed to the object's \code{\link[=print]{print()}} method, so preview-specific
options such as \code{show_content} work here too.}
}
\value{
A character vector of formatted lines.
}
\description{
Returns an object's printed representation as a character vector, one element
per line, so it can be captured, logged, pasted into a report, or otherwise
reused. \code{\link[=print]{print()}} renders the very same content to the console; \code{format()}
hands it back to you instead.
}
\examples{
link <- webr_repl_link("plot(1:10)")
format(link)
writeLines(format(link))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/classes.R
\name{print.webr_preview}
\alias{print.webr_preview}
\title{Print method for webr_preview objects}
\usage{
\method{print}{webr_preview}(x, show_content = FALSE, max_content_length = 500, ...)
}
\arguments{
\item{x}{webr_preview object}

\item{show_content}{Logical. Whether to print the contents of each file (default: FALSE)}

\item{max_content_length}{Maximum number of characters of content to show per file (default: 500)}

\item{...}{Additional arguments (ignored)}
}
\value{
Invisibly returns the object
}
\description{
Print method for webr_preview objects
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shinylive-link.R
\name{shinylive_directory}
\alias{shinylive_directory}
\title{Create Shinylive sharelinks from a directory of Shiny apps}
\usage{
shinylive_directory(
  directory_path,
  engine,
  mode = "editor",
  header = TRUE,
  app_file = NULL,
  base_url = NULL
)
}
\arguments{
\item{directory_path}{Character string specifying the path to the directory containing Shiny app directories}

\item{engine}{Engine to use: "r" for R Shiny or "python" for Python Shiny}

\item{mode}{Shinylive display mode (default \code{"editor"}). \code{"editor"} shows an
editable code panel beside the running app; \code{"app"} shows only the running app.}

\item{header}{Logical, whether to show the Shinylive header bar. It applies only
when \code{mode = "app"} and is ignored in the default \code{"editor"} mode. Defaults to \code{TRUE}.}

\item{app_file}{Main app filename to look for (default: "app.R" for R, "app.py" for Python)}

\item{base_url}{Custom Shinylive base URL. If NULL (default), links point at https://shinylive.io.}
}
\value{
shinylive_directory object containing URLs and metadata for all found apps
}
\description{
Batch processes directories containing Shiny applications to create individual Shinylive links.
Each subdirectory is treated as a separate Shiny app project.

Only text files with extensions .R, .py, .txt, .md, .csv, .json, .yaml, or
.yml are embedded in a link. Other files (for example images or binary data)
are skipped with a warning.
}
\examples{
# Each app lives in its own subdirectory:
#   shiny_apps/
#     app1/app.R
#     app2/app.R
shiny_apps <- tempfile()
dir.create(file.path(shiny_apps, "app1"), recursive = TRUE)
dir.create(file.path(shiny_apps, "app2"), recursive = TRUE)
writeLines("library(shiny)", file.path(shiny_apps, "app1", "app.R"))
writeLines("library(shiny)", file.path(shiny_apps, "app2", "app.R"))

links <- shinylive_directory(shiny_apps, engine = "r", mode = "editor")
print(links)

# Extract just the URLs
repl_urls(links)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as-data-frame.R
\name{as.data.frame.livelink}
\alias{as.data.frame.livelink}
\alias{as.data.frame.webr_directory}
\alias{as.data.frame.shinylive_directory}
\alias{as.data.frame.webr_decoded_batch}
\alias{as.data.frame.shinylive_decoded_batch}
\title{Turn a livelink container into a data frame}
\usage{
\method{as.data.frame}{webr_directory}(x, row.names = NULL, optional = FALSE, ...)

\method{as.data.frame}{shinylive_directory}(x, row.names = NULL, optional = FALSE, ...)

\method{as.data.frame}{webr_decoded_batch}(x, row.names = NULL, optional = FALSE, ...)

\method{as.data.frame}{shinylive_decoded_batch}(x, row.names = NULL, optional = FALSE, ...)
}
\arguments{
\item{x}{A \code{webr_directory}, \code{shinylive_directory}, \code{webr_decoded_batch}, or
\code{shinylive_decoded_batch} object.}

\item{row.names}{A character vector of row names, or \code{NULL}.}

\item{optional}{Ignored; present for compatibility with the generic.}

\item{...}{Ignored.}
}
\value{
A data frame. For a directory, one row per generated link with columns
\code{filename} and \code{url}. For a decoded batch, one row per successfully decoded
URL with columns \code{name}, \code{url}, \code{total_files}, \code{total_size}, and
\code{output_dir} (URLs that failed to decode carry no result and are omitted; the
object's \code{total_urls} and \code{successful_urls} fields hold the counts).
}
\description{
The container classes coerce to a tidy data frame, so a folder of links or a
batch of decoded results can be tabulated, filtered, joined, or written to
CSV with the tools you already use. For a tibble, wrap the result:
\code{tibble::as_tibble(as.data.frame(x))}.
}
\examples{
dir <- tempfile()
dir.create(dir)
writeLines("plot(1:10)",       file.path(dir, "one.R"))
writeLines("hist(rnorm(100))", file.path(dir, "two.R"))

links <- webr_repl_directory(dir)
as.data.frame(links)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/webr-repl-directory-links.R
\name{webr_repl_directory}
\alias{webr_repl_directory}
\title{Create webR REPL sharelinks from a directory of R files}
\usage{
webr_repl_directory(
  directory_path,
  autorun = FALSE,
  single_link = FALSE,
  pattern = "\\\\.R$",
  base_path = "/home/web_user/",
  panels = NULL,
  version = "latest",
  base_url = NULL
)
}
\arguments{
\item{directory_path}{Character string specifying the path to the directory containing R files}

\item{autorun}{Logical. Whether to enable autorun for all generated links (default: FALSE).
With \code{single_link = TRUE}, this runs every R file in the bundle on arrival.}

\item{single_link}{Logical. If \code{FALSE} (default), each matched file becomes its
own link and the result is a \code{webr_directory}. If \code{TRUE}, all matched files
are packed into one link and the result is a single \code{webr_project}.}

\item{pattern}{Regular expression matched against file names in
\code{directory_path}. Defaults to \code{"\\\\.R$"}, i.e. files ending in \code{.R}.}

\item{base_path}{Base directory path for files in webR (default: "/home/web_user/")}

\item{panels}{Character vector or string specifying which webR interface panels to show.
Valid panels: "plot", "files", "terminal", "editor". Can be c("plot", "files") or "plot-files".
If NULL (default), shows all panels.}

\item{version}{webR version to use ("latest" or specific version >= "v0.5.4")}

\item{base_url}{webR application URL. If NULL, uses global option or builds from version}
}
\value{
By default a \code{webr_directory} object, whose \code{urls} element is a named character
vector mapping each filename to its webR sharelink. With \code{single_link = TRUE}, a
single \code{webr_project} object bundling every matched file into one link.
}
\description{
Batch processes all R files in a directory. By default each file becomes its
own webR sharelink; with \code{single_link = TRUE} the whole directory is bundled
into one link instead, exactly as \code{\link[=webr_repl_project]{webr_repl_project()}} would. Useful for
converting collections of scripts, examples, or course materials.
}
\examples{
# A directory of R scripts
examples <- tempfile()
dir.create(examples)
writeLines("plot(1:10)", file.path(examples, "plot.R"))
writeLines("hist(rnorm(100))", file.path(examples, "hist.R"))

links <- webr_repl_directory(examples, autorun = TRUE)
print(links)

# Bundle the whole directory into one link instead
webr_repl_directory(examples, single_link = TRUE, panels = c("editor", "plot"))

# Show only the editor and terminal panels
webr_repl_directory(examples, panels = c("editor", "terminal"))

# Match a subset of files
webr_repl_directory(examples, pattern = "^plot")

# The URLs, named by file
repl_urls(links)

}
\seealso{
\code{\link[=webr_repl_project]{webr_repl_project()}}, which bundles a named list or a vector of file
paths into one link.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/webr-repl-links.R
\name{webr_repl_link}
\alias{webr_repl_link}
\title{Create a webR REPL sharelink from R code}
\usage{
webr_repl_link(
  input = NULL,
  filename = "script.R",
  path = NULL,
  autorun = FALSE,
  panels = NULL,
  version = "latest",
  base_url = NULL
)
}
\arguments{
\item{input}{Code input. Can be:
\itemize{
\item R expression (no quotes needed): \code{webr_repl_link({ plot(1:10) })}
\item Character string: R code to execute
\item File path: Path to R file to read
\item NULL: Read from clipboard (requires clipr package)
}}

\item{filename}{Name for the file (default: \code{"script.R"})}

\item{path}{Full path where the file will be placed in webR. If NULL (default),
the file is placed at \code{"/home/web_user/{filename}"}.}

\item{autorun}{Logical. Whether to auto-execute the code when link is opened (default: \code{FALSE}).
Only R files (\code{.R}) can be auto-executed.}

\item{panels}{Character vector or string specifying which webR interface panels to show.
Valid panels: \code{"plot"}, \code{"files"}, \code{"terminal"}, \code{"editor"}. Can be \code{c("plot", "files")} or \code{"plot-files"}.
If NULL (default), shows all panels.}

\item{version}{webR version to use (\code{"latest"} or specific version >= "v0.5.4")}

\item{base_url}{webR application URL. If NULL, uses global option or builds from version}
}
\value{
webr_link object containing the webR sharelink and metadata
}
\description{
Generates a shareable URL for R code that can be executed in the webR environment.
Supports expressions, file paths, character strings, and clipboard input.
}
\section{Comments in expression input}{

Expression input recovers your source from R's \emph{source references}, which R only
attaches when \code{keep.source} is enabled. Comments therefore survive in an
interactive session, but are dropped when the calling code is parsed without
source references -- notably inside a knitted 'Quarto' or 'R Markdown' document,
because 'knitr' evaluates chunks through \code{evaluate::evaluate()}, which discards
them. No \code{keep.source} setting recovers them there.

If you need comments preserved, pass the code as a string or a file path, or write
it as a chunk in the document -- see \link{livelink-knitr} and
\code{vignette("links-in-documents", package = "livelink")}.
}

\examples{
# Expression input (no quotes needed!)
webr_repl_link({
  plot(1:10)
  summary(mtcars)
})

# Traditional string input
webr_repl_link("plot(1:10)")

# Choose which panels the REPL shows
webr_repl_link({ hist(rnorm(100)) }, panels = c("plot", "editor"))

# Run the code as soon as the link opens
webr_repl_link("plot(1:10)", autorun = TRUE)

# File path input
script <- tempfile(fileext = ".R")
writeLines("plot(1:10)", script)
webr_repl_link(script)

# Read the code from the clipboard
if (interactive()) {
  webr_repl_link()
}
}
\seealso{
\code{\link[=webr_repl_project]{webr_repl_project()}} for multi-file projects and
\code{\link[=webr_repl_exercise]{webr_repl_exercise()}} for exercise and solution pairs; \link{livelink-knitr} to
give a document chunk its own link; \code{vignette("getting-started", package = "livelink")} for an introduction.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shinylive-link.R
\name{shinylive_py_link}
\alias{shinylive_py_link}
\title{Create a Shinylive sharelink for Python Shiny apps}
\usage{
shinylive_py_link(
  input = NULL,
  mode = "editor",
  header = TRUE,
  base_url = NULL
)
}
\arguments{
\item{input}{App input. Can be:
\itemize{
\item Character string: Python code for the app
\item File path: Path to app.py file
\item Vector of file paths: Multiple files for the app
\item Named list: \code{list("app.py" = code1, "utils.py" = code2)}
\item NULL: Read from clipboard
}}

\item{mode}{Shinylive display mode (default \code{"editor"}). \code{"editor"} shows an
editable code panel beside the running app; \code{"app"} shows only the running app.}

\item{header}{Logical, whether to show the Shinylive header bar. It applies only
when \code{mode = "app"} and is ignored in the default \code{"editor"} mode. Defaults to \code{TRUE}.}

\item{base_url}{Custom Shinylive base URL. If NULL (default), links point at https://shinylive.io.}
}
\value{
shinylive_link object containing the Shinylive URL and metadata
}
\description{
Generates a shareable URL for Python Shiny applications that can run in the browser
using Shinylive. Supports character strings, file paths, named lists, and clipboard input.
}
\examples{
# String input
app_code <- "
from shiny import App, render, ui
app_ui = ui.page_fluid(ui.h2('Hello World'))
def server(input, output, session): pass
app = App(app_ui, server)
"
shinylive_py_link(app_code)

# Multiple files as a named list
shinylive_py_link(list(
  "app.py" = app_code,
  "utils.py" = "def helper(): return 42"
))

# File path input
app_dir <- tempfile()
dir.create(app_dir)
app_path <- file.path(app_dir, "app.py")
writeLines("from shiny import App, ui", app_path)
shinylive_py_link(app_path)

# Read the app from the clipboard
if (interactive()) {
  shinylive_py_link()
}
}
\seealso{
\code{\link[=shinylive_project]{shinylive_project()}} for multi-file apps; \code{\link[=decode_shinylive_link]{decode_shinylive_link()}}
and \code{\link[=preview_shinylive_link]{preview_shinylive_link()}} to read a link back.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/knitr-engine.R
\name{livelink-knitr}
\alias{livelink-knitr}
\alias{use_livelink_hook}
\alias{use_livelink_engine}
\title{Turn document chunks into shareable links}
\usage{
use_livelink_hook()

use_livelink_engine()
}
\value{
Called for their side effect. Invisibly \code{TRUE} if registration
happened, \code{FALSE} if 'knitr' is not installed.
}
\description{
livelink plugs into 'knitr' two ways, and which you want depends on one
question: \strong{should the code also run in your document?}

\strong{A chunk hook}, on an ordinary \code{r} chunk. The chunk runs as usual -- its
output, plots and all, appear in the rendered page -- and a link is added
underneath. Use this for code you want your reader to \emph{see the result of} and
\emph{also} be able to open and play with.

\if{html}{\out{<div class="sourceCode">}}\preformatted{```\{r\}
#| livelink: true
#| autorun: true
# Load the data
data(mtcars)
plot(mtcars$mpg, mtcars$wt)
```
}\if{html}{\out{</div>}}

\strong{An engine}, as \verb{```\{livelink\}}. The chunk is displayed but \strong{not}
run: only the link is produced. Use this for code your session cannot or should
not execute -- a Shiny app, something needing a package you have not installed,
or anything slow.

\if{html}{\out{<div class="sourceCode">}}\preformatted{```\{livelink\}
#| engine.target: shinylive-r
library(shiny)
shinyApp(fluidPage(), function(input, output) \{\})
```
}\if{html}{\out{</div>}}

There is deliberately no \code{{shinylive-r}} or \code{{shinylive-py}} engine. 'knitr'
cannot dispatch a chunk whose engine name contains a hyphen (its chunk syntax
forbids it), and in Quarto such a cell is handed to the Shinylive extension
rather than to 'knitr'. Name Shinylive through \code{engine.target} instead.
}
\details{
Reach for either of these rather than expression input
(\code{webr_repl_link({ ... })}) inside a knitted document. 'knitr' evaluates
chunks through \code{evaluate::evaluate()}, which discards source references, and
comments live in those -- so comments inside a \code{{ }} expression are silently
dropped from the link when the document renders, and no \code{keep.source} setting
brings them back. Both the hook and the engine are handed the chunk's verbatim
source, so nothing is lost.

Both are registered automatically when livelink is loaded, provided 'knitr' is
installed. Call these yourself only if you have reset \code{knitr::knit_hooks} or
\code{knitr::knit_engines}.
}
\section{Chunk options}{
\describe{
\item{\code{livelink}}{Hook only. \code{true} for a webR link, or name the target
directly: \code{"webr"}, \code{"shinylive-r"}, \code{"shinylive-py"}.}
\item{\code{engine.target}}{Engine only. \code{"webr"} (default), \code{"shinylive-r"}, or
\code{"shinylive-py"}.}
\item{\code{autorun}}{Logical. Run the code as soon as the link opens. webR only.}
\item{\code{panels}}{Character vector of webR panels, e.g. \code{c("editor", "plot")}.}
\item{\code{mode}}{Shinylive only. Display mode, \code{"editor"} (default) or \code{"app"}.}
\item{\code{filename}}{webR only. Name for the script file in the webR virtual
filesystem (default \code{"script.R"}); must end in \code{.R} for \code{autorun} to work.}
\item{\code{link.text}}{Text for the hyperlink. Defaults to \code{"Open in webR"} or
\code{"Open in Shinylive"}.}
\item{\code{link.only}}{Engine only. If \code{TRUE}, emit the link without the source.}
}
}

\section{Setting options once}{
These are ordinary 'knitr' chunk options, so \code{opts_chunk} sets them for a whole
document, and a single chunk opts out with \code{livelink: false}:

\if{html}{\out{<div class="sourceCode">}}\preformatted{knitr::opts_chunk$set(livelink = TRUE, autorun = TRUE)
}\if{html}{\out{</div>}}
}

\section{\code{echo} does not gate the link}{
It is natural to assume the code must be visible for a link to be made. It need
not be. \code{echo} controls whether the \strong{source is shown in your page}; the link is
built from the chunk's source, which 'knitr' hands over either way. So
\code{echo: false} gives a working link whose code the reader simply cannot see.

\code{eval: false} is the other half: the chunk is displayed but not run, which makes
an \code{r} chunk behave rather like the engine.
}

\examples{
\dontshow{if (requireNamespace("knitr", quietly = TRUE)) withAutoprint(\{ # examplesIf}
# Both are registered on load; call directly only after resetting knitr's hooks.
use_livelink_hook()
use_livelink_engine()
\dontshow{\}) # examplesIf}
}
\seealso{
\code{vignette("links-in-documents", package = "livelink")} for the whole
picture, and \code{\link[=webr_repl_link]{webr_repl_link()}} for why a braced expression loses comments in
a knitted document.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/options.R
\name{set_webr_base_url}
\alias{set_webr_base_url}
\title{Set global base URL for webR links}
\usage{
set_webr_base_url(base_url = NULL)
}
\arguments{
\item{base_url}{Custom base URL to use for all webR links}
}
\value{
Invisibly returns the \code{base_url} value (or \code{NULL} if resetting
to default).
}
\description{
Overrides the base URL used when building webR REPL links, for instance to
point at a self-hosted or pinned webR deployment. The value is stored in the
\code{livelink.base_url} option and applies to webR links only; Shinylive links are
unaffected. Once set, a custom base URL takes precedence over the \code{version}
argument passed to the link builders.
}
\examples{
# Remember the current setting so it can be restored afterwards
old <- getOption("livelink.base_url")

# Set custom base URL
set_webr_base_url("https://my-custom-webr.com/")

# Reset to default (removes custom setting)
set_webr_base_url(NULL)

# Restore the previous setting
set_webr_base_url(old)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/knit-print.R
\name{knit_print.livelink}
\alias{knit_print.livelink}
\alias{knit_print.webr_link}
\alias{knit_print.webr_project}
\alias{knit_print.webr_exercise}
\alias{knit_print.webr_directory}
\alias{knit_print.shinylive_link}
\alias{knit_print.shinylive_project}
\alias{knit_print.shinylive_directory}
\title{Render livelink objects as links in knitted documents}
\usage{
\method{knit_print}{webr_link}(x, ...)

\method{knit_print}{webr_project}(x, ...)

\method{knit_print}{webr_exercise}(x, ...)

\method{knit_print}{webr_directory}(x, ...)

\method{knit_print}{shinylive_link}(x, ...)

\method{knit_print}{shinylive_project}(x, ...)

\method{knit_print}{shinylive_directory}(x, ...)
}
\arguments{
\item{x}{A livelink object (a link, project, exercise, or directory).}

\item{...}{Ignored.}
}
\value{
A \code{knit_asis} object (via \code{\link[knitr:asis_output]{knitr::asis_output()}}).
}
\description{
'knitr' calls \code{knit_print()} on the last value of a chunk. Without these
methods a link object would fall back to \code{\link[=print]{print()}}, dumping cli console output
(a header, box glyphs, metadata) into the rendered page. These methods instead
emit a clickable Markdown link, which is almost always what you want when a
link object is the visible result of a chunk.

A single link becomes \verb{[Open in webR](url)} (or Shinylive); a project the
same. A directory or an exercise, which carry several named URLs, become a
bulleted list, one titled link per entry.

These fire only inside 'knitr'. Call \code{\link[=print]{print()}} explicitly for the full cli
description, or \code{\link[=as.character]{as.character()}} for the bare URL.
}
\seealso{
\link{livelink-knitr} for the chunk hook and engine, \code{\link[=format.livelink]{format.livelink()}}
and \code{\link[=as.character.webr_link]{as.character.webr_link()}} for other renderings, and
\code{vignette("links-in-documents", package = "livelink")}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/decode-shinylive-link.R
\name{decode_shinylive_link}
\alias{decode_shinylive_link}
\title{Decode Shinylive link(s) to extract files to local directory}
\usage{
decode_shinylive_link(
  url,
  output_dir = file.path(tempdir(), "shinylive_files"),
  overwrite = FALSE,
  create_subdir = TRUE,
  name_dirs = TRUE
)
}
\arguments{
\item{url}{Character string or vector containing Shinylive URL(s)}

\item{output_dir}{Character string specifying the output directory path.
Defaults to a \code{shinylive_files} directory inside the session temporary
directory; pass an explicit path to extract somewhere permanent.}

\item{overwrite}{Logical. Whether to overwrite existing files (default: FALSE)}

\item{create_subdir}{Logical. If \code{TRUE} (default), each decoded link is
extracted into its own subdirectory under \code{output_dir} rather than directly
into it. For a single URL the subdirectory is named \verb{shinylive_<hash>}, where
\verb{<hash>} is a short fingerprint of the URL. For multiple URLs, see
\code{name_dirs}. Set \code{FALSE} to extract straight into \code{output_dir}; with multiple
URLs this means identically named files (such as \code{app.R}) collide, so a later
app overwrites an earlier one when \code{overwrite = TRUE}, or is skipped when
\code{overwrite = FALSE}.}

\item{name_dirs}{Logical. For multiple URLs, controls how the per-link
subdirectories are named: \code{TRUE} (default) numbers them \code{app_01}, \code{app_02},
...; \code{FALSE} names each one \verb{shinylive_<hash>} from the URL fingerprint.
Ignored for a single URL, and ignored when \code{create_subdir = FALSE}.}
}
\value{
For a single URL, a \code{shinylive_decoded} object. For multiple URLs, a
\code{shinylive_decoded_batch} object.
}
\description{
Decodes Shinylive sharelinks to extract the embedded files and save them
to a local directory. This is the reverse operation of creating Shinylive links.
Handles both single URLs and multiple URLs automatically.
}
\examples{
# Round-trip: build a link, then decode it back to files
url <- as.character(shinylive_r_link("library(shiny)"))

result <- decode_shinylive_link(url)
print(result)

# Extract to a directory of your choosing
out <- file.path(tempdir(), "my_app")
decode_shinylive_link(url, output_dir = out, create_subdir = FALSE, overwrite = TRUE)
list.files(out)

# Both engines at once
urls <- c(url, as.character(shinylive_py_link("from shiny import App")))
decode_shinylive_link(urls, output_dir = file.path(tempdir(), "my_apps"))

}
\seealso{
\code{\link[=preview_shinylive_link]{preview_shinylive_link()}} to inspect a link without writing files;
\code{\link[=shinylive_r_link]{shinylive_r_link()}} and \code{\link[=shinylive_py_link]{shinylive_py_link()}} to create links.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shinylive-link.R
\name{shinylive_r_link}
\alias{shinylive_r_link}
\title{Create a Shinylive sharelink for R Shiny apps}
\usage{
shinylive_r_link(input = NULL, mode = "editor", header = TRUE, base_url = NULL)
}
\arguments{
\item{input}{App input. Can be:
\itemize{
\item R expression (no quotes needed): \code{shinylive_r_link({ shinyApp(ui, server) })}
\item Character string: R code for the app
\item File path: Path to app.R file
\item Vector of file paths: Multiple files for the app
\item Named list: \code{list("app.R" = code1, "utils.R" = code2)}
\item NULL: Read from clipboard
}}

\item{mode}{Shinylive display mode (default \code{"editor"}). \code{"editor"} shows an
editable code panel beside the running app; \code{"app"} shows only the running app.}

\item{header}{Logical, whether to show the Shinylive header bar. It applies only
when \code{mode = "app"} and is ignored in the default \code{"editor"} mode. Defaults to \code{TRUE}.}

\item{base_url}{Custom Shinylive base URL. If NULL (default), links point at https://shinylive.io.}
}
\value{
shinylive_link object containing the Shinylive URL and metadata
}
\description{
Generates a shareable URL for R Shiny applications that can run in the browser
using Shinylive. Supports expressions, character strings, file paths, named lists,
and clipboard input.
}
\section{Comments in expression input}{

Comments inside a \code{{ }} expression are recovered from R's source references, so
they survive interactively but are dropped inside a knitted 'Quarto' or
'R Markdown' document. Pass a string or a file path, or use the \code{livelink} chunk
engine, if you need them preserved. See \code{\link[=webr_repl_link]{webr_repl_link()}} for the details.
}

\examples{
# Expression input (no quotes needed!)
shinylive_r_link({
  ui <- fluidPage(titlePanel("Hello World"))
  server <- function(input, output) {}
  shinyApp(ui, server)
})

# Multiple files as a named list
shinylive_r_link(list(
  "app.R" = "library(shiny)\nshinyApp(fluidPage(), function(i, o) {})",
  "utils.R" = "helper <- function() 42"
))

# File path input
app_dir <- tempfile()
dir.create(app_dir)
app_path <- file.path(app_dir, "app.R")
writeLines("library(shiny)", app_path)
shinylive_r_link(app_path)

# Read the app from the clipboard
if (interactive()) {
  shinylive_r_link()
}
}
\seealso{
\code{\link[=shinylive_project]{shinylive_project()}} for multi-file apps; \code{\link[=decode_shinylive_link]{decode_shinylive_link()}}
and \code{\link[=preview_shinylive_link]{preview_shinylive_link()}} to read a link back; \link{livelink-knitr} to give
a document chunk its own link;
\code{vignette("webr-and-shinylive", package = "livelink")} for the guide.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/classes.R
\name{print.shinylive_directory}
\alias{print.shinylive_directory}
\title{Print method for shinylive_directory objects}
\usage{
\method{print}{shinylive_directory}(x, ...)
}
\arguments{
\item{x}{shinylive_directory object}

\item{...}{Additional arguments (ignored)}
}
\value{
Invisibly returns the object
}
\description{
Print method for shinylive_directory objects
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extract-link.R
\name{as.character.webr_link}
\alias{as.character.webr_link}
\alias{as.character.shinylive_link}
\alias{as.character.webr_project}
\alias{as.character.webr_exercise}
\alias{as.character.webr_directory}
\alias{as.character.webr_decoded}
\alias{as.character.webr_decoded_batch}
\alias{as.character.webr_preview}
\alias{as.character.shinylive_project}
\alias{as.character.shinylive_directory}
\alias{as.character.shinylive_decoded}
\alias{as.character.shinylive_decoded_batch}
\alias{as.character.shinylive_preview}
\title{Extract URLs as character vector}
\usage{
\method{as.character}{webr_link}(x, ...)

\method{as.character}{shinylive_link}(x, ...)

\method{as.character}{webr_project}(x, ...)

\method{as.character}{webr_exercise}(x, ...)

\method{as.character}{webr_directory}(x, ...)

\method{as.character}{webr_decoded}(x, ...)

\method{as.character}{webr_decoded_batch}(x, ...)

\method{as.character}{webr_preview}(x, ...)

\method{as.character}{shinylive_project}(x, ...)

\method{as.character}{shinylive_directory}(x, ...)

\method{as.character}{shinylive_decoded}(x, ...)

\method{as.character}{shinylive_decoded_batch}(x, ...)

\method{as.character}{shinylive_preview}(x, ...)
}
\arguments{
\item{x}{Link object}

\item{...}{Additional arguments}
}
\value{
A character vector of URLs. Most objects yield a single URL; exercise objects
return a length-2 named vector (\code{exercise}, \code{solution}); directory and batch
objects return one URL per file.
}
\description{
Extract URLs as character vector
}
\seealso{
\code{\link[=repl_urls]{repl_urls()}} for the same extraction as a named generic you can call
explicitly.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/decode-shinylive-link.R
\name{preview_shinylive_link}
\alias{preview_shinylive_link}
\title{Preview Shinylive link contents without writing files to disk}
\usage{
preview_shinylive_link(url)
}
\arguments{
\item{url}{Character string containing the Shinylive URL}
}
\value{
shinylive_preview object with file information and metadata
}
\description{
Decodes a Shinylive URL and returns information about the embedded files
without actually saving them to disk. Use print method options to control display.
}
\examples{
url <- as.character(shinylive_r_link("library(shiny)"))

# Inspect a link without writing anything to disk
preview <- preview_shinylive_link(url)

# Default print (no content)
print(preview)

# Show file contents
print(preview, show_content = TRUE)

# Show file contents with custom length limit
print(preview, show_content = TRUE, max_content_length = 200)

# Access the preview data
preview$files_data
preview$total_files

}
\seealso{
\code{\link[=decode_shinylive_link]{decode_shinylive_link()}} to extract files to disk;
\code{\link[=shinylive_r_link]{shinylive_r_link()}} and \code{\link[=shinylive_py_link]{shinylive_py_link()}} to create links.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/decode-webr-repl-link.R
\name{preview_webr_link}
\alias{preview_webr_link}
\title{Preview webR REPL link contents without writing files to disk}
\usage{
preview_webr_link(url)
}
\arguments{
\item{url}{Character string containing the webR URL}
}
\value{
A \code{webr_preview} object (a list) with elements including \code{files_data}
(the embedded files), \code{total_files}, \code{total_size}, \code{mode}, \code{version},
\code{flags}, and \code{autorun_files}. Its \code{print()} method accepts \code{show_content}
and \code{max_content_length} to display file bodies.
}
\description{
Decodes a webR URL and returns information about the embedded files
without actually saving them to disk. Use print method options to control display.
}
\examples{
url <- as.character(webr_repl_link("plot(1:10)"))

# Inspect a link without writing anything to disk
preview <- preview_webr_link(url)

# Default print (no content)
print(preview)

# Show file contents
print(preview, show_content = TRUE)

# Show file contents with custom length limit
print(preview, show_content = TRUE, max_content_length = 200)

# Access the preview data
preview$files_data
preview$total_files

}
\seealso{
\code{\link[=decode_webr_link]{decode_webr_link()}} to extract the embedded files to disk once you
are happy with the preview.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/classes.R
\name{print.webr_decoded}
\alias{print.webr_decoded}
\title{Print method for webr_decoded objects}
\usage{
\method{print}{webr_decoded}(x, ...)
}
\arguments{
\item{x}{webr_decoded object}

\item{...}{Additional arguments (ignored)}
}
\value{
Invisibly returns the object
}
\description{
Print method for webr_decoded objects
}
